/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim, Princeton University			   *
 *   Author: Vladimir Kim (kim_vladimir@yahoo.com, vk@princeton.edu)       *
 *   Research Supervisor: Tom Funkhouser (funk@cs.princeton.edu)           *
 *                                  ****                                   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "CVLib.h"
#include "DrawingRoutines.h"
#include <Qt/qmap.h>
#include "del_interface.hpp"
#include "Numerical.h"
#include "TimeProfiler.h"
#include "PDSampling.h"
using namespace tpp;

SortableEntry::SortableEntry(double val, int id, void * ptr)
{
	m_fValue = val;
	m_iID = id;
	m_pData = ptr;
}

bool SortingFunction(SortableEntry entry1, SortableEntry entry2)
{
	return entry1.m_fValue < entry2.m_fValue;
}


R3Mesh * CVLibRoutines::m_pTempMeshPointer;

void CVLibRoutines::FindFeaturePnts(QImage & img, int &numPnts, std::vector<double> & trg, 
									double minDist, bool addMorePnts)
{
	IplImage * cvImgGray = cvCreateImage(cvSize(img.width(), img.height()), IPL_DEPTH_8U, 1);
	IplImage * cvImgEig = cvCreateImage(cvSize(img.width(), img.height()), IPL_DEPTH_32F, 1);
	IplImage * cvImgTmp = cvCreateImage(cvSize(img.width(), img.height()), IPL_DEPTH_32F, 1);	
	CvScalar s;
	for (int i=0; i<img.width(); i++)
	{
		for (int j=0; j<img.height(); j++)
		{
			s.val[0] = qGray(img.pixel(i,j));
			assert(s.val[0] <= 255 && s.val[0]>=0);
			cvSet2D(cvImgGray, j, i, s);
		}
	}
	CvPoint2D32f * corners = new CvPoint2D32f[numPnts];
	int trgNumPnts = numPnts;
	numPnts = 0;
	for (double step=1.; step>0 && numPnts<trgNumPnts; step-=0.05)
	{
		numPnts = trgNumPnts;
		cvGoodFeaturesToTrack(cvImgGray, cvImgEig, cvImgTmp, corners, &numPnts, step, minDist, NULL, 
							  qMax(minDist/2., 3.), 0);
	}
		
	if (numPnts<trgNumPnts && addMorePnts)
	{
		std::cout<<"[WARNING] More points than features: "<<numPnts<<" / "<<trgNumPnts;
		double scaleX = img.width()/2.;
		double scaleY = img.height()/2.;		
		double shift = -1;
//		std::cout<<"minDist="<<minDist<<std::endl;
		DartThrowing morePnts(minDist/qMin(scaleX, scaleY), true, trgNumPnts, 10);

		//DartThrowing morePnts(sqrt(trgNumPnts / .7)/2, true, 1000, 10);
		for (int i=0; i<numPnts; i++)
			morePnts.addPoint(Vec2(corners[i].x/scaleX+shift, corners[i].y/scaleY+shift));
		morePnts.complete();
		for (int i=0; i<(int)morePnts.points.size(); i++)
		{
			trg.push_back((morePnts.points[i].x-shift) * scaleX);
			trg.push_back((morePnts.points[i].y-shift) * scaleY);
		}			
		std::cout<<". Extended to="<<morePnts.points.size()<<" / "<<trgNumPnts<<std::endl;
	}	
	else
	{
		for (int i=0; i<numPnts; i++)
		{
			trg.push_back(corners[i].x);
			trg.push_back(corners[i].y);
		}		
	}
	
	delete [] corners;
}

void CVLibRoutines::FindFeatureMesh(QImage img, int &numPnts, std::vector<double> & vertices, 
									std::vector<int> & triangles, PointsToMeshMethod triangluationMethod,
									bool addMorePnts)
{
	QImage unmodifiedSrc = img;
	bool verbous = true;
	double minDist = qMax(qMin(img.width()/(2.*(double)sqrt(numPnts)), img.height()/(2.*(double)sqrt(numPnts))), 1.);
	bool pointsGiven = vertices.size() > 0;
	if (!pointsGiven)	// find feature ponts (not given)
		FindFeaturePnts(img, numPnts, vertices, minDist, addMorePnts);
	
	std::cout<<"Feature pnts="<<vertices.size()/2<<std::endl;
	
	// Add boundary vertices
	std::vector<int> specialVertices;	
	if (!pointsGiven)
	{
		ProjectToBoundaryVertices(vertices, img.width(), img.height(), minDist);
		//AddBoundaryVertices(vertices, specialVertices, img.width(), img.height(), minDist);
	}
	std::cout<<"Feature pnts + Bnd Pnts="<<vertices.size()/2<<std::endl;	
	// (temp) render feature points
	DrawPoints(img, vertices);	
	img.save("MeshFeatures.png");
		
	if (triangluationMethod==TRIANGULATE_DELAUNAY)
	{
		GetDelaunayTriangulation(vertices, triangles);
	}
	else if (triangluationMethod==TRIANGULATE_NAIVE)
	{
		// Create (very) rough meshing - Don't forget triangle norm should be pointing up!
		triangles.push_back(specialVertices[0]);	
		triangles.push_back(specialVertices[1]);	
		triangles.push_back(specialVertices[3]);
		triangles.push_back(specialVertices[0]);
		triangles.push_back(specialVertices[3]);
		triangles.push_back(specialVertices[2]);
		
		// Add unvisited vertices
		for (int i=0; i<(int)vertices.size()/2; i++)
		{
			if (verbous)
			{
				std::cout<<"Adding "<<i<<"/"<<(vertices.size()/2)<<std::endl;
				QImage tempImg(img);	
				DrawMesh(tempImg, vertices, triangles);
				DrawPoints(tempImg, vertices);
				DrawingRoutines::DrawCircle(tempImg, vertices[i*2+0]/img.width(), vertices[i*2+1]/img.height(), 
											1./(double)qMax(img.width(), img.height()), 0, qRgb(0, 195, 0), qRgb(0,0,0));
				tempImg.save("image"+QString::number(i)+".png");
			}
			
			if (i!=specialVertices[0] && i!=specialVertices[1] && i!=specialVertices[2] && i!=specialVertices[3])
				AddVertexToTriangles(i, triangles, vertices, minDist/2, img.width(), img.height());
		}
	}	
	
	if (verbous)
	{
		// (temp) render mesh
		std::cout<<"Triangles="<<triangles.size()<<std::endl;
		DrawMesh(img, vertices, triangles);
		DrawPoints(img, vertices);
		img.save("MeshTriangles.png");
	}
	if (!pointsGiven)	
		SortVerticesByImportance(unmodifiedSrc, vertices, triangles, SORTING_VERTICES_DECIMATION);
}

void CVLibRoutines::GetDelaunayTriangulation(std::vector<double> & vertices, std::vector<int> & triangles)
{
	Delaunay::Point tempP;
	std::vector< Delaunay::Point > v;
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		tempP[0] = vertices[i+0];			
		tempP[1] = vertices[i+1];			
		v.push_back(tempP);
	}
	
	Delaunay delobject(v);
	delobject.Triangulate();
	
	for(Delaunay::fIterator fit  = delobject.fbegin(); fit != delobject.fend(); ++fit)
	{			
		triangles.push_back(delobject.Org(fit));
		triangles.push_back(delobject.Dest(fit));
		triangles.push_back(delobject.Apex(fit));			
	}	
}

void CVLibRoutines::GetRidOfRedundant(std::vector<double> & vertices, std::vector<double> & edgeVerticesX, 
									  std::vector<double> & edgeVerticesY, double width, double height, double minDist)
{
	minDist = minDist*minDist;
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		bool removeVertex = false;
		double vx = vertices[i+0];
		double vy = vertices[i+1];
		for (int j=0; j<(int)edgeVerticesX.size() && !removeVertex; j++)
		{
			double bndX = edgeVerticesX[j];
			double bndY = 0;
			removeVertex = (vx-bndX)*(vx-bndX) + (vy-bndY)*(vy-bndY) < minDist;
			
			bndX = edgeVerticesX[j];
			bndY = height;
			removeVertex = removeVertex || ((vx-bndX)*(vx-bndX) + (vy-bndY)*(vy-bndY) < minDist);
		}
		
		for (int j=0; j<(int)edgeVerticesY.size() && !removeVertex; j++)
		{
			double bndX = 0;
			double bndY = edgeVerticesY[j];
			removeVertex = (vx-bndX)*(vx-bndX) + (vy-bndY)*(vy-bndY) < minDist;
			
			bndX = width;
			bndY = edgeVerticesY[j];
			removeVertex = removeVertex || ((vx-bndX)*(vx-bndX) + (vy-bndY)*(vy-bndY) < minDist);
		}
		if (removeVertex)
		{
			vertices.erase(vertices.begin()+i, vertices.begin()+(i+2));
			i-=2;
		}
	}
}

void CVLibRoutines::ProjectToBoundaryVertices(std::vector<double> & vertices, double width, 
											  double height, double minDist)
{
	bool corner00 = false;
	bool corner01 = false;
	bool corner11 = false;
	bool corner10 = false;	
	minDist = qMin(4*minDist, qMin(width/25, height/25));

	std::vector<bool> left;
	std::vector<bool> right;
	std::vector<bool> top;
	std::vector<bool> bottom;
	for (int i=0; i<ceil(width+1); i++)
	{
		top.push_back(false);
		bottom.push_back(false);		
	}
	for (int i=0; i<ceil(height+1); i++)
	{
		left.push_back(false);
		right.push_back(false);		
	}
		
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		double & vx = vertices[i+0];
		double & vy = vertices[i+1];
		assert(vx>=0 && vx<=width && vy>=0 && vy<=height);
		bool removeVertex = false;
		if (width - vx < minDist)
		{
			vx = width;
			if (right[ceil(vy)])
				removeVertex = true;
			else
				right[ceil(vy)] = true;
		}
		if (vx < minDist)
		{
			vx = 0;
			if (left[ceil(vy)])
				removeVertex = true;
			else
				left[ceil(vy)] = true;			
		}
		if (height - vy < minDist)
		{
			vy = height;
			if (bottom[ceil(vx)])
				removeVertex = true;
			else
				bottom[ceil(vx)] = true;			
		}
		
		if (vy < minDist)
		{
			vy = 0;
			if (top[ceil(vx)])
				removeVertex = true;
			else
				top[ceil(vx)] = true;						
		}
		
		if (removeVertex)
		{
			vertices.erase(vertices.begin()+i, vertices.begin()+i+2);
			i-=2;
		}
		else
		{
			corner00 = vx==0 && vy==0;
			corner01 = vx==0 && vy==height;
			corner10 = vx==width && vy==0;
			corner11 = vx==width && vy==height;		
		}
	}
	if (!corner00)
	{
		vertices.push_back(0);		vertices.push_back(0);	
	}
	if (!corner01)
	{
		vertices.push_back(0);		vertices.push_back(height);	
	}
	if (!corner10)
	{
		vertices.push_back(width);		vertices.push_back(0);	
	}
	if (!corner11)
	{
		vertices.push_back(width);		vertices.push_back(height);	
	}
}

void CVLibRoutines::AddBoundaryVertices(std::vector<double> & vertices, std::vector<int> & specialVertexIDs, double width, double height, double minDist)
{
	std::vector<double> sortedBoundaryX;
	std::vector<double> sortedBoundaryY;	
	sortedBoundaryX.push_back(0);
	sortedBoundaryY.push_back(0);	
	sortedBoundaryX.push_back(width);
	sortedBoundaryY.push_back(height);	
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		sortedBoundaryX.push_back(vertices[i+0]);
		sortedBoundaryY.push_back(vertices[i+1]);
	}
	std::sort(sortedBoundaryX.begin(), sortedBoundaryX.end());
	std::sort(sortedBoundaryY.begin(), sortedBoundaryY.end());	
	int inspectForDelete = 1;	
	while (inspectForDelete<(int)sortedBoundaryX.size()-1)	//do not inspect first and last boundary vertex
	{
		if (sortedBoundaryX[inspectForDelete+1]-sortedBoundaryX[inspectForDelete-1] < minDist
			|| sortedBoundaryX[inspectForDelete+1]-sortedBoundaryX[inspectForDelete] < minDist/2
			|| sortedBoundaryX[inspectForDelete]-sortedBoundaryX[inspectForDelete-1] < minDist/2)
			sortedBoundaryX.erase(sortedBoundaryX.begin()+inspectForDelete);
		else
			inspectForDelete++;
	}
	inspectForDelete=1;
	while (inspectForDelete<(int)sortedBoundaryY.size()-1)	//do not inspect first and last boundary vertex
	{
		if (sortedBoundaryY[inspectForDelete+1]-sortedBoundaryY[inspectForDelete-1] < minDist
			|| sortedBoundaryY[inspectForDelete+1]-sortedBoundaryY[inspectForDelete] < minDist/2
			|| sortedBoundaryY[inspectForDelete]-sortedBoundaryY[inspectForDelete-1] < minDist/2)
			sortedBoundaryY.erase(sortedBoundaryY.begin()+inspectForDelete);
		else
			inspectForDelete++;
	}
	GetRidOfRedundant(vertices, sortedBoundaryX, sortedBoundaryY, width, height, minDist);		
	
	specialVertexIDs.push_back(vertices.size()/2);		//1st corner [0, 0]
	specialVertexIDs.push_back(vertices.size()/2+1);	//2nd corner [0, H]
//	firstTri2 = vertices.size()/2 + 2;	- nearest to corner
	
	for (int i=0; i<(int)sortedBoundaryX.size(); i++)
	{
		vertices.push_back(sortedBoundaryX[i]);
		vertices.push_back(0);
		vertices.push_back(sortedBoundaryX[i]);
		vertices.push_back(height);		
	}
	specialVertexIDs.push_back(vertices.size()/2-2);		//3st corner [W, 0]	
	specialVertexIDs.push_back(vertices.size()/2-1);		//4st corner [W, H]		
	
//	firstTri3 = vertices.size()/2;	//nearest to 1st corner
	for (int i=1; i<(int)sortedBoundaryY.size()-1; i++)
	{
		vertices.push_back(0);
		vertices.push_back(sortedBoundaryY[i]);
		vertices.push_back(width);
		vertices.push_back(sortedBoundaryY[i]);
	}
	
	std::vector<bool> vertexVisited;
	std::vector<QMap<int, double> > vertexToEdgeMap;
	for (int i=0; i<(int)vertices.size()/2; i++)	
	{
		vertexVisited.push_back(false);
		vertexToEdgeMap.push_back(QMap<int, double>());
	}		
}

void CVLibRoutines::AddVertexToTriangles(int vertex, std::vector<int> & triangles, std::vector<double> & vertices, 
										 double minDist, double width, double height)
{

	//Don't forget triangle norm should be pointing up!

	double vx = vertices[vertex*2+0];
	double vy = vertices[vertex*2+1];	

	int inTriangle = -1;
	int edgeV1 = -1;
	int edgeV2 = -1;
	int inV=-1, adjV=-1;	
	double edgeDist;
	//find corresponding triangle and minimal edge
	for (int i=0; i<(int)triangles.size() && inTriangle<0; i+=3)
	{
		int v0 = triangles[i+0];		int v1 = triangles[i+1];		int v2 = triangles[i+2];
		
		double v0x = vertices[v0*2+0];			double v0y = vertices[v0*2+1];
		double v1x = vertices[v1*2+0];			double v1y = vertices[v1*2+1];
		double v2x = vertices[v2*2+0];			double v2y = vertices[v2*2+1];		
		if (CheckInTriangle(v0x, v0y, v1x, v1y, v2x, v2y, vx, vy))
		{
			inTriangle = i/3;
			edgeV1 = v0;	edgeV2 = v1;	inV = v2;
			edgeDist = DistanceToEdge(v0x, v0y, v1x, v1y, vx, vy);
			
			double dist = DistanceToEdge(v0x, v0y, v2x, v2y, vx, vy);
			if (dist < edgeDist)
			{
				edgeDist = dist;
				edgeV1 = v0;	edgeV2 = v2;	inV = v1;
			}
			
			dist = DistanceToEdge(v1x, v1y, v2x, v2y, vx, vy);
			if (dist < edgeDist)
			{
				edgeDist = dist;
				edgeV1 = v1;	edgeV2 = v2;	inV = v0;
			}			
		}
	}
	if (!(edgeV1>=0 && edgeV2>=0 && inTriangle>=0))
	{
		std::cout<<"e1="<<edgeV1<<" e2="<<edgeV2<<" tri="<<inTriangle<<std::endl;
	}
	assert(edgeV1>=0 && edgeV2>=0 && inTriangle>=0);
	bool boundary = (vertices[2*edgeV1+0]==0 && vertices[2*edgeV2+0]==0) || (vertices[2*edgeV1+0]==width && vertices[2*edgeV2+0]==width);
	boundary = boundary || (vertices[2*edgeV1+1]==0 && vertices[2*edgeV2+1]==0) || (vertices[2*edgeV1+1]==height && vertices[2*edgeV2+1]==height);
	
	//case 1: (not boundary) 
	if (!boundary)
	{
		// find 2nd triangle adjacent to the minimal edge
		int adjTriangle = -1;
		for (int i=0; i<(int)triangles.size() && adjTriangle<0; i+=3)
		{
			int v0 = triangles[i+0];		int v1 = triangles[i+1];		int v2 = triangles[i+2];
			if (i/3==inTriangle)
				continue;
			if ((v0==edgeV1 && v1==edgeV2) || (v0==edgeV2 && v1==edgeV1))
			{
				adjTriangle = i/3;
				adjV = v2;
			}
			else if ((v0==edgeV1 && v2==edgeV2) || (v0==edgeV2 && v2==edgeV1))
			{
				adjTriangle = i/3;
				adjV = v1;
			}
			else if ((v2==edgeV1 && v1==edgeV2) || (v2==edgeV2 && v1==edgeV1))
			{
				adjTriangle = i/3;
				adjV = v0;
			}
		}
		assert(adjTriangle>=0);
		
		// destroy 2 current triangles
		if (inTriangle > adjTriangle)
		{
			triangles.erase(triangles.begin()+inTriangle*3, triangles.begin()+(inTriangle*3+3));
			triangles.erase(triangles.begin()+adjTriangle*3, triangles.begin()+(adjTriangle*3+3));
		}
		else
		{
			triangles.erase(triangles.begin()+adjTriangle*3, triangles.begin()+(adjTriangle*3+3));			
			triangles.erase(triangles.begin()+inTriangle*3, triangles.begin()+(inTriangle*3+3));			
		}
		
		// generate 4 new triangles
		assert(inV>=0 && adjV>=0);
		triangles.push_back(edgeV1);		triangles.push_back(inV);		triangles.push_back(vertex);
		triangles.push_back(inV);			triangles.push_back(edgeV2);	triangles.push_back(vertex);
		triangles.push_back(edgeV2);		triangles.push_back(adjV);		triangles.push_back(vertex);
		triangles.push_back(adjV);			triangles.push_back(edgeV1);	triangles.push_back(vertex);		
	}

	//case 2a: (boundary, vertex is far)
	if (boundary && edgeDist >= minDist)
	{
		// destroy triangle
		triangles.erase(triangles.begin()+inTriangle*3, triangles.begin()+(inTriangle*3+3));

		// generate 3 new triangles
		triangles.push_back(edgeV1);		triangles.push_back(inV);		triangles.push_back(vertex);		
		triangles.push_back(inV);			triangles.push_back(edgeV2);	triangles.push_back(vertex);		
		triangles.push_back(edgeV2);		triangles.push_back(edgeV1);	triangles.push_back(vertex);				
	}

	//case 2b: (boundary, vertex is close)
	if (boundary && edgeDist < minDist)
	{
		// destroy triangle
		triangles.erase(triangles.begin()+inTriangle*3, triangles.begin()+(inTriangle*3+3));
		
		// project vertex to boundary		
		if (vertices[edgeV1*2+0]==0 && vertices[edgeV2*2+0]==0)
			vertices[vertex*2 + 0] = 0;
		else if (vertices[edgeV1*2+1]==0 && vertices[edgeV2*2+1]==0)
			vertices[vertex*2 + 1] = 0;			
		else if (vertices[edgeV1*2+0]==width && vertices[edgeV2*2+0]==width)
			vertices[vertex*2 + 0] = width;			
		else if (vertices[edgeV1*2+1]==height && vertices[edgeV2*2+1]==height)
			vertices[vertex*2 + 1] = height;
		else
			assert(false);
		
		// generate 2 new triangles
		triangles.push_back(edgeV1);		triangles.push_back(inV);		triangles.push_back(vertex);		
		triangles.push_back(inV);			triangles.push_back(edgeV2);	triangles.push_back(vertex);		
	}
}

bool CVLibRoutines::CheckInTriangle(double v0x, double v0y, double v1x, double v1y, 
									double v2x, double v2y, double vx, double vy)
{
	// Compute vectors        
	double dx0 = v1x-v0x;	double dy0 = v1y-v0y;
	double dx1 = v2x-v0x;	double dy1 = v2y-v0y;
	double dx2 = vx-v0x;	double dy2 = vy-v0y;
	
	// Compute dot products
	double dot00 = dx0*dx0 + dy0*dy0;	double dot01 = dx0*dx1 + dy0*dy1;	double dot02 = dx0*dx2 + dy0*dy2;
	double dot11 = dx1*dx1 + dy1*dy1;	double dot12 = dx1*dx2 + dy1*dy2;
	
	// Compute barycentric coordinates
	double denom = (dot00 * dot11 - dot01 * dot01);
	double u = (dot11 * dot02 - dot01 * dot12) / denom;
	double v = (dot00 * dot12 - dot01 * dot02) / denom;
	
	// Check if point is in triangle
//	std::cout<<"uv=["<<u<<", "<<v<<"] in?="<<((u >= 0) && (v >= 0) && (u + v <= 1))<<std::endl;
	return (u >= 0) && (v >= 0) && (u + v <= 1);
}

double CVLibRoutines::DistanceToEdge(double x1, double y1, double x2, double y2, double px, double py)
{
	//NOTE: ported from MATLAB/HumanPose/RasterizePieceLin.m
	double A, B, C, D, dot, len_sq, param, xx, yy;
	A = px - x1;				B = py - y1;
	C = x2 - x1;				D = y2 - y1;
	dot = A * C + B * D;
	len_sq = C * C + D * D;
	param = dot / len_sq;
	
	if (param < 0)
	{
		xx = x1;					yy = y1;
	}
	else if (param > 1)
	{
		xx = x2;					yy = y2;
	}
	else
	{
		xx = x1 + param * C;		yy = y1 + param * D;
	}
	
	return sqrt((px-xx) * (px-xx) + (py-yy)*(py-yy));
}

void CVLibRoutines::DrawPoints(QImage & img, std::vector<double> & vertices)
{
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		DrawingRoutines::DrawCircle(img, vertices[i]/img.width(), vertices[i+1]/img.height(), 
									2./(double)qMax(img.width(), img.height()), 0, qRgb(0, 0, 255), qRgb(0,0,0));
	}	
}

void CVLibRoutines::DrawMesh(QImage & img, std::vector<double> & vertices, std::vector<int> & triangles)
{
	for (int i=0; i<(int)triangles.size(); i+=3)
	{
		int v1 = triangles[i+0];
		int v2 = triangles[i+1];
		int v3 = triangles[i+2];
		DrawingRoutines::DrawLine(img, vertices[v1*2+0]/img.width(), vertices[v1*2+1]/img.height(), 
								  vertices[v2*2+0]/img.width(), vertices[v2*2+1]/img.height(),
								  1/(double)img.width(), qRgb(255, 0, 0));
		DrawingRoutines::DrawLine(img, vertices[v1*2+0]/img.width(), vertices[v1*2+1]/img.height(), 
								  vertices[v3*2+0]/img.width(), vertices[v3*2+1]/img.height(),
								  1/(double)img.width(), qRgb(255, 0, 0));
		DrawingRoutines::DrawLine(img, vertices[v2*2+0]/img.width(), vertices[v2*2+1]/img.height(), 
								  vertices[v3*2+0]/img.width(), vertices[v3*2+1]/img.height(),
								  1/(double)img.width(), qRgb(255, 0, 0));
	}	
}

void CVLibRoutines::SortVerticesByImportance(QImage & img, std::vector<double> & vertices, std::vector<int> & triangles, SortingMeshVerticesMethod sortingMethod)
{
	if (sortingMethod==SORTING_VERTICES_DECIMATION)
	{
		std::vector<int> order;
		SortVerticesByDecimation(img, vertices, triangles, order);
//		ImageBasedDecimation(img, vertices, triangles, targetNumVertices);
	}
}

void CVLibRoutines::SortVerticesByDecimation(QImage & img, std::vector<double> & vertices, 
											 std::vector<int> & triangles, std::vector<int> & vertexOrder)
{
//	TimeProfiler profiler;
//	profiler.startedProcess("SortingByDecimation");
	bool debug = false;
	QMap<R3MeshVertex *, int> vertexToID;
	R3Mesh * mesh = ConvertToR3Mesh(vertices, triangles);
	for (int i=0; i<mesh->NVertices(); i++)
		vertexToID[mesh->Vertex(i)] = i;
	
	m_pTempMeshPointer = mesh;	
	
	//precompute edge errors
	for (int i=0; i<mesh->NEdges(); i++)
	{
		R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 0)));
		R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 1)));
		mesh->SetEdgeValue(mesh->Edge(i), mesh->EdgeLength(mesh->Edge(i)));
	}
	
	while (true)
	{
		//find min edge
		R3MeshEdge * minEdge = NULL;
		double currVal;
		for (int i=0; i<mesh->NEdges(); i++)
		{
			double val = mesh->EdgeValue(mesh->Edge(i));
			R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 0)));
			R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 1)));
			if (minEdge==NULL || val < currVal)		// if minimum
			{
				currVal = val;
				minEdge = mesh->Edge(i);
			}
		}
		
		if (minEdge==NULL || currVal==SYM_INFINITY)
			break;
		
		//decimate min edge
		R3Point collapseTo = mesh->EdgeMidpoint(minEdge);
		
		R3MeshVertex * removedVertex = mesh->VertexOnEdge(minEdge, 1);
		if (vertexToID[removedVertex]>=0)
		{
			vertexOrder.insert(vertexOrder.begin(), vertexToID[removedVertex]);
			vertexToID[removedVertex] = -1;
		}
		
		
		R3MeshVertex * newVertex = mesh->CollapseEdge(minEdge, collapseTo);		
		
		//update edge list
		if (newVertex==NULL)	//can't delete
		{
			mesh->SetEdgeValue(minEdge, SYM_INFINITY);	
		}
		else
		{
			for (int i=0; i<mesh->VertexValence(newVertex); i++)
			{
				R3MeshEdge * e = mesh->EdgeOnVertex(newVertex, i);
				R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(e, 0)));
				R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(e, 1)));
				mesh->SetEdgeValue(e, mesh->EdgeLength(e));
			}
			
			if (debug)
			{
				std::vector<double> tempVert;
				std::vector<int> tempTri;
				ConvertFromR3Mesh(mesh, tempVert, tempTri);
				QImage tempImg(img);
				DrawMesh(tempImg, tempVert, tempTri);
				DrawPoints(tempImg, tempVert);
				tempImg.save("tempImg_f_"+QString::number(mesh->NFaces())+".png");
			}
		}
	}	
	
	for (int i=0; i<mesh->NVertices(); i++)
	{
		if (vertexToID[mesh->Vertex(i)]>=0)
		{
			vertexOrder.insert(vertexOrder.begin(), vertexToID[mesh->Vertex(i)]);
			vertexToID[mesh->Vertex(i)] = -1;
		}
	}
	
	assert(vertexOrder.size()==vertices.size()/2);
	
	std::vector<double> verticesInNewOrder;
	std::vector<int> vertexOrderInverse;
	for (int i=0; i<(int)vertexOrder.size(); i++)
		vertexOrderInverse.push_back(-1);
	for (int i=0; i<(int)vertexOrder.size(); i++)
	{
		verticesInNewOrder.push_back(vertices[2*vertexOrder[i]+0]);
		verticesInNewOrder.push_back(vertices[2*vertexOrder[i]+1]);		
		vertexOrderInverse[vertexOrder[i]] = i;
	}
	vertices = verticesInNewOrder;
	
	for (int i=0; i<(int)triangles.size(); i++)
		triangles[i] = vertexOrderInverse[triangles[i]];
//	vertexOrder.clear();
//	vertexOrder = vertexOrderInverse;
//	profiler.finishedProcess("SortingByDecimation", true);	
}

void CVLibRoutines::ImageBasedDecimation(QImage & img, std::vector<double> & vertices, std::vector<int> & triangles, int targetNumVertices)
{
	R3Mesh * mesh = ConvertToR3Mesh(vertices, triangles);
	m_pTempMeshPointer = mesh;	

	//precompute edge errors
	for (int i=0; i<mesh->NEdges(); i++)
	{
		R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 0)));
		R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 1)));
		mesh->SetEdgeValue(mesh->Edge(i), GetEdgeWeight(img, p1.X(), p1.Y(), p2.X(), p2.Y()));
	}
	
	int edges = mesh->NEdges();
	while (true)
	{
		edges--;
		//find min edge
		R3MeshEdge * minEdge = NULL;
		double currVal;
		for (int i=0; i<mesh->NEdges(); i++)
		{
			double val = mesh->EdgeValue(mesh->Edge(i));
			R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 0)));
			R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(mesh->Edge(i), 1)));
			if ((minEdge==NULL || val < currVal)		// if minimum
				&& qAbs(p1.X() - p2.X())<img.width()
				&& qAbs(p1.Y() - p2.Y())<img.height())	//not on opposite corners
			{
				currVal = val;
				minEdge = mesh->Edge(i);
			}
		}
		
		R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(minEdge, 0)));
		R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(minEdge, 1)));
		if (minEdge==NULL || currVal==SYM_INFINITY)
			break;
		
		//decimate min edge
		R3Point collapseTo = mesh->EdgeMidpoint(minEdge);
//		if (p1.X()==0 || p1.X()==img.width())
//			collapseTo.X(p1.X());
//		if (p2.X()==0 || p2.X()==img.width())
//			collapseTo.X(p2.X());
//		if (p1.Y()==0 || p1.Y()==img.height())
//			collapseTo.Y(p1.Y());
//		if (p2.Y()==0 || p2.Y()==img.height())
//			collapseTo.Y(p2.Y());
			
		R3MeshVertex * newVertex = mesh->CollapseEdge(minEdge, collapseTo);		
		
		//update edge list
		if (newVertex==NULL)	//can't delete
		{
			mesh->SetEdgeValue(minEdge, SYM_INFINITY);	
		}
		else
		{
			for (int i=0; i<mesh->VertexValence(newVertex); i++)
			{
				R3MeshEdge * e = mesh->EdgeOnVertex(newVertex, i);
				R3Point p1(mesh->VertexPosition(mesh->VertexOnEdge(e, 0)));
				R3Point p2(mesh->VertexPosition(mesh->VertexOnEdge(e, 1)));
				mesh->SetEdgeValue(e, GetEdgeWeight(img, p1.X(), p1.Y(), p2.X(), p2.Y()));
			}
			
			std::vector<double> tempVert;
			std::vector<int> tempTri;
			ConvertFromR3Mesh(mesh, tempVert, tempTri);
			QImage tempImg(img);
			DrawMesh(tempImg, tempVert, tempTri);
			DrawPoints(tempImg, tempVert);
			tempImg.save("tempImg_f_"+QString::number(mesh->NFaces())+".png");
		}
	}
	
}

double CVLibRoutines::GetEdgeWeight(QImage & img, double v1x, double v1y, double v2x, double v2y)
{
	if (v1x>=img.width())
		v1x = img.width()-1;
	if (v1y>=img.height())
		v1y = img.height()-1;
	if (v2x>=img.width())
		v2x = img.width()-1;
	if (v2y>=img.height())
		v2y = img.height()-1;
	
	QRgb c1 = img.pixel(v1x, v1y);
	QRgb c2 = img.pixel(v2x, v2y);	
	double dr = (double)(qRed(c1) - qRed(c2)) / 255.;
	double dg = (double)(qGreen(c1) - qGreen(c2)) / 255.;
	double db = (double)(qBlue(c1) - qBlue(c2)) / 255.;	
	return dr*dr + dg*dg + db*db;
}

R3Mesh * CVLibRoutines::ConvertToR3Mesh(std::vector<double> & vertices, std::vector<int> & triangles)
{
	std::vector<R3MeshVertex *> gapsVertices;
	R3Mesh * newMesh = new R3Mesh();
	R3Point newPnt;
	newPnt.Z(0);
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		newPnt.X(vertices[i+0]);
		newPnt.Y(vertices[i+1]);
		gapsVertices.push_back(newMesh->CreateVertex(newPnt));
	}
	
	for (int i=0; i<(int)triangles.size(); i+=3)
		newMesh->CreateFace(gapsVertices[triangles[i+0]], gapsVertices[triangles[i+1]], gapsVertices[triangles[i+2]]);
	
	return newMesh;
}

void CVLibRoutines::ConvertFromR3Mesh(R3Mesh * srcMesh, std::vector<double> & vertices, std::vector<int> & triangles)
{
	for (int i=0; i<srcMesh->NVertices(); i++)
	{
		vertices.push_back(srcMesh->VertexPosition(srcMesh->Vertex(i)).X());
		vertices.push_back(srcMesh->VertexPosition(srcMesh->Vertex(i)).Y());
	}
	
	for (int i=0; i<srcMesh->NFaces(); i++)
	{
		R3MeshFace * face = srcMesh->Face(i);
		triangles.push_back(srcMesh->VertexID(srcMesh->VertexOnFace(face, 0)));
		triangles.push_back(srcMesh->VertexID(srcMesh->VertexOnFace(face, 1)));
		triangles.push_back(srcMesh->VertexID(srcMesh->VertexOnFace(face, 2)));		
	}
}


//void CVLibRoutines::GetOutwardNormal(double v1x, double v1y, double v2x, double v2y, double v3x, double v3y, double * nx, double *ny)
//{
//	double mx = 0.5*(v1x + v2x);
//	double my = 0.5*(v1y + v2y);
//	*nx = v2y-v1y;			// norm = [dy, -dx]
//	*ny = -(v2x-v1x);
//	double norm = sqrt(*nx * *nx + *ny * *ny);
//	double dx = v3x-mx;
//	double dy = v3y-my;	
//	if (dx * *nx + dy * *ny>0)		//orient outward
//		norm *=-1;
//	*nx/=norm;
//	*ny/=norm;
//}

//int CVLibRoutines::PickNextVertexGivenEdge(int v1, int v2, double nx, double ny, std::vector<double> &vertices)
//{
//	double mx = 0.5*(vertices[2*v1+0] + vertices[2*v2+0]);
//	double my = 0.5*(vertices[2*v1+1] + vertices[2*v2+1]);	
//	int minDistVertex = -1;
//	double minDist = -1;
//	for (int i=0; i<(int)vertices.size()/2; i++)
//	{
//		if (i!=v1 && i!=v2)
//		{
//			double dx = vertices[2*i+0] - mx;
//			double dy = vertices[2*i+1] - my;
//			if (dx*nx + dy*ny > 0)
//			{
//				double dist = dx*dx + dy*dy;
//				if (minDist < 0 || dist < minDist)
//				{
//					minDist = dist;
//					minDistVertex = i;
//				}
//			}
//		}
//	}
//	return minDistVertex;
//}

/*
 // Add vertices to the mesh - wave algorithm
 std::vector<int> edgesToConsider;
 std::vector<double> edgeNormals;
 
 //start at the corner
 edgesToConsider.push_back(firstTri2);	edgesToConsider.push_back(firstTri3);
 double v1x = vertices[firstTri2*2+0];	double v1y = vertices[firstTri2*2+1];
 double v2x = vertices[firstTri3*2+0];	double v2y = vertices[firstTri3*2+1];	
 double v3x = vertices[firstTri1*2+0];	double v3y = vertices[firstTri1*2+1];		
 double nx, ny;
 GetOutwardNormal(v1x, v1y, v2x, v2y, v3x, v3y, &nx, &ny);
 edgeNormals.push_back(nx);				edgeNormals.push_back(ny);
 vertexToEdgeMap[firstTri1].insert(firstTri2, 1);		vertexToEdgeMap[firstTri2].insert(firstTri1, 1);
 vertexToEdgeMap[firstTri1].insert(firstTri3, 1);		vertexToEdgeMap[firstTri3].insert(firstTri1, 1);
 vertexToEdgeMap[firstTri3].insert(firstTri2, 1);		vertexToEdgeMap[firstTri3].insert(firstTri2, 1);	
 triangles.push_back(firstTri1);		triangles.push_back(firstTri2);		triangles.push_back(firstTri3);	
 vertexVisited[firstTri1] = true;	vertexVisited[firstTri2] = true;	vertexVisited[firstTri3] = true;
 
 std::cout<<"Start with: "<<firstTri1<<", "<<firstTri2<<", "<<firstTri3<<std::endl;
 while (edgesToConsider.size()>0)
 {
 int v1 = edgesToConsider[0];
 int v2 = edgesToConsider[1];
 
 // get next vertex
 int v3 = PickNextVertexGivenEdge(v1, v2, edgeNormals[0], edgeNormals[1], vertices);
 std::cout<<"Adding Vertex "<<v3<<std::endl;
 if (v3<0)
 {
 edgesToConsider.erase(edgesToConsider.begin(), edgesToConsider.begin()+2);
 edgeNormals.erase(edgeNormals.begin(), edgeNormals.begin()+2);			
 continue;
 }
 //		assert(v3>=0);
 //construct another triangle
 triangles.push_back(v1);			triangles.push_back(v2);		triangles.push_back(v3);
 
 //if 2 new edges are not boundary and were not examined before, add them to the list
 if (!vertexToEdgeMap[v1].contains(v3))	//if edge was not examined
 {
 vertexToEdgeMap[v1].insert(v3, 1);		vertexToEdgeMap[v3].insert(v1, 1);	//add to the list of examined
 bool boundaryEdge = (vertices[v1*2+0]==0 && vertices[v3*2+0]==0) || (vertices[v1*2+1]==0 && vertices[v3*2+1]==0);
 if (!boundaryEdge)	//if edge is not boundary - add to the 'to consider' list
 {
 edgesToConsider.push_back(v1);		edgesToConsider.push_back(v3);
 double nx, ny;
 GetOutwardNormal(vertices[v1*2+0], vertices[v1*2+1], vertices[v3*2+0], vertices[v3*2+0], 
 vertices[v2*2+0], vertices[v2*2+0], &nx, &ny);
 edgeNormals.push_back(nx);			edgeNormals.push_back(ny);
 }
 }
 
 if (!vertexToEdgeMap[v2].contains(v3))	//if edge was not examined
 {
 vertexToEdgeMap[v2].insert(v3, 1);		vertexToEdgeMap[v3].insert(v2, 1);	//add to the list of examined
 bool boundaryEdge = (vertices[v2*2+0]==0 && vertices[v3*2+0]==0) || (vertices[v2*2+1]==0 && vertices[v3*2+1]==0);
 if (!boundaryEdge)	//if edge is not boundary - add to the 'to consider' list
 {
 edgesToConsider.push_back(v2);			edgesToConsider.push_back(v3);
 double nx, ny;
 GetOutwardNormal(vertices[v2*2+0], vertices[v2*2+1], vertices[v3*2+0], vertices[v3*2+0], 
 vertices[v1*2+0], vertices[v1*2+0], &nx, &ny);		
 edgeNormals.push_back(nx);				edgeNormals.push_back(ny);				
 }
 }		
 
 
 edgesToConsider.erase(edgesToConsider.begin(), edgesToConsider.begin()+2);
 edgeNormals.erase(edgeNormals.begin(), edgeNormals.begin()+2);
 }
 */