/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim, Princeton University			   *
 *   Author: Vladimir Kim (kim_vladimir@yahoo.com, vk@princeton.edu)       *
 *   Research Supervisor: Tom Funkhouser (funk@cs.princeton.edu)           *
 *                                  ****                                   *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "DrawingRoutines.h"
#include <math.h>
#include <assert.h>

void DrawingRoutines::DrawMaskOver(QImage & image, QImage & mask, QRgb emptyColor)
{
	assert(mask.width()==image.width() && mask.height()==image.height());
	assert(qRed(emptyColor)>=0 && qRed(emptyColor) <=255
		   && qGreen(emptyColor)>=0 && qGreen(emptyColor) <=255
		   && qBlue(emptyColor)>=0 && qBlue(emptyColor) <=255);
	
	for (int i=0; i<image.width(); i++)
	{
		for (int j=0; j<image.height(); j++)
		{
			if (mask.pixel(i,j)!=emptyColor)
				image.setPixel(i,j, mask.pixel(i,j));
		}
	}
}

void DrawingRoutines::DrawCircle(QImage & target, float x, float y, float radius, float width,
					   QRgb inside_color, QRgb outside_color, DrawDirections direct)
{
	int minX = x * target.width()  - radius * target.width()  - width * target.width()  - 1;
	int minY = y * target.height() - radius * target.height() - width * target.height() - 1; 
	int maxX = x * target.width()  + radius * target.width()  + width * target.width()  + 1;
	int maxY = y * target.height() + radius * target.height() + width * target.height() + 1; 
	for (int i = minX; i <= maxX; i++)
	{
		for (int j=minY; j<=maxY; j++)
		{
			double dx = ((float)i + 0.5)/target.width()  - x;
			double dy = ((float)j + 0.5)/target.height() - y;
			
			if (dx*dx + dy*dy < radius*radius)
				DrawPixel(target, i, j, inside_color, direct);
			else if (dx*dx + dy*dy < (radius + width)*(radius+width))
				DrawPixel(target, i, j, outside_color, direct);
		}
	}
}

void DrawingRoutines::DrawEllipse(QImage & target, float x, float y, float radiusX, float radiusY,
						QRgb color, DrawDirections direct)
{	
	double relativeCenterX = x * target.width();
	double relativeCenterY = y * target.height();	
	double relRadiusX = radiusX * target.width();
	double relRadiusY = radiusY * target.height();	
	for (int pixelX = (int)((x-radiusX)*target.width()-1); pixelX<=(int)((x+radiusX)*target.width()+1); pixelX++)
		for (int pixelY = (int)((y-radiusY)*target.height()-1); pixelY<=(int)((y+radiusY)*target.height()+1); pixelY++)
			if (pow((pixelX-relativeCenterX)/relRadiusX, 2) + pow((pixelY-relativeCenterY)/relRadiusY, 2)<=1)
				DrawPixel(target, pixelX, pixelY, color, direct);
	
}

void DrawingRoutines::DrawRectangle(QImage & target, float x0, float y0, float w, float h, 
						  QRgb color, DrawDirections direct)
{
	for (int pixelX = (int)((x0)*target.width()); pixelX<=(int)((x0+w)*target.width()); pixelX++)
		for (int pixelY = (int)((y0)*target.height()); pixelY<=(int)((y0+h)*target.height()); pixelY++)
			DrawPixel(target, pixelX, pixelY, color, direct);
	
}

void DrawingRoutines::DrawLine(QImage & target, float x1, float y1, float x2, float y2,
					 float width, QRgb color, DrawDirections direct)
{
	bool stepAlongX = qAbs(y2-y1) < qAbs(x2-x1);

	if (stepAlongX)
	{
		if (x2<x1)
		{
			double temp = x2;		x2 = x1;		x1 = temp;
			temp = y2;				y2 = y1;		y1 = temp;			
		}
		
		double pixel = 1. / target.width();
		double step = (y2-y1) / (x2-x1);
		double y = y1;
		for (double x = x1; x <= x2; x += pixel)
		{
			if (width==1)
				DrawPixel(target, (float)x, (float)y, color, direct);
			else 
				DrawCircle(target, (float)x, (float)y, width, 0, color, qRgb(0,0,0), direct);

			y += step * pixel;
		}
	}
	else
	{
		if (y2<y1)
		{
			double temp = x2;		x2 = x1;		x1 = temp;
			temp = y2;				y2 = y1;		y1 = temp;			
		}
		
		double pixel = 1. / target.height();
		double step = (x2-x1) / (y2-y1);
		double x = x1;
		for (double y = y1; y <= y2; y += pixel)
		{
			if (width==1)
				DrawPixel(target, (float)x, (float)y, color, direct);
			else 
				DrawCircle(target, (float)x, (float)y, width, 0, color, qRgb(0,0,0), direct);
			
			x += step * pixel;
		}		
	}
	
}

//void DrawingRoutines::DrawArrow(QImage & target, float x1, float y1, float x2, float y2,
//					  float width, QRgb color, DrawDirections direct)
//{
//	
//}

void DrawingRoutines::DrawPixel(QImage & target, float x, float y, QRgb color, DrawDirections direct )
{

	if (x<0 || x >= 1 || y< 0 || y>=1)
		return;

	assert(qRed(color)>=0 && qRed(color) <=255
		   && qGreen(color)>=0 && qGreen(color) <=255
		   && qBlue(color)>=0 && qBlue(color) <=255);
	
	int i = x * target.width();
	int j = y * target.height();

	if (direct == DRAW_OVER)
	{
		target.setPixel(i,j, color);
	}
	else if (direct== ADD_INTENSITY)
	{
		QRgb newColor = target.pixel(x,y);
		assert(qRed(newColor)>=0 && qRed(newColor) <=255
			   && qGreen(newColor)>=0 && qGreen(newColor) <=255
			   && qBlue(newColor)>=0 && qBlue(newColor) <=255);
		
		int newRed = qRed(newColor) + qRed(color);		int newGreen = qGreen(newColor) + qGreen(color);		int newBlue = qBlue(newColor) + qBlue(color);
		int gray = (11*newRed + 16*newGreen + 6*newBlue)/32;
		if (gray > 255) { newRed = newRed * 255 / gray;			newGreen = newGreen * 255 / gray;		newBlue = newBlue * 255 / gray; }
		newColor = qRgb(newRed, newGreen, newBlue);
		target.setPixel(x, y, newColor);
	}
	else
	{
		std::cout<<"DrawingRoutines: NOT IMPLEMENTED FUNCTIONALITY"<<std::endl;
		exit(0);
	}
}

void DrawingRoutines::DrawPixel(QImage & target, int x, int y, QRgb color, DrawDirections direct )
{
//	std::cout<<"Pixel="<<x<<", "<<y<<std::endl;
//	std::cout<<"targetDim="<<target.width()<<" x "<<target.height()<<std::endl;
	if (x<0 || x >= target.width() || y< 0 || y>=target.height())
		return;
	assert(qRed(color)>=0 && qRed(color) <=255
		   && qGreen(color)>=0 && qGreen(color) <=255
		   && qBlue(color)>=0 && qBlue(color) <=255);
	
	if (direct == DRAW_OVER)
	{
//		std::cout<<"Drawing over: x="<<x<<", "<<y<<std::endl;
		target.setPixel(x,y, color);
//		std::cout<<"????"<<std::endl;
	}
	else if (direct== ADD_INTENSITY)	//checks if luminance (e.g. gray) is not > 255
	{
		QRgb newColor = target.pixel(x,y);
		assert(qRed(newColor)>=0 && qRed(newColor) <=255
			   && qGreen(newColor)>=0 && qGreen(newColor) <=255
			   && qBlue(newColor)>=0 && qBlue(newColor) <=255);
		
		int newRed = qRed(newColor) + qRed(color);		int newGreen = qGreen(newColor) + qGreen(color);		int newBlue = qBlue(newColor) + qBlue(color);
		int gray = (11*newRed + 16*newGreen + 6*newBlue)/32;
		if (gray > 255) { newRed = newRed * 255 / gray;			newGreen = newGreen * 255 / gray;		newBlue = newBlue * 255 / gray; }
		newColor = qRgb(newRed, newGreen, newBlue);
		target.setPixel(x, y, newColor);
	}
	else if (direct == TAKE_MIN)
	{
		QRgb newColor = target.pixel(x,y);
		assert(qRed(newColor)>=0 && qRed(newColor) <=255
			   && qGreen(newColor)>=0 && qGreen(newColor) <=255
			   && qBlue(newColor)>=0 && qBlue(newColor) <=255);
		
		int newRed = qMin(qRed(newColor), qRed(color));
		int newGreen = qMin(qGreen(newColor), qGreen(color));
		int newBlue = qMin(qBlue(newColor), qBlue(color));		
		newColor = qRgb(newRed, newGreen, newBlue);		
		target.setPixel(x, y, newColor);		
	}
	else
	{
		std::cout<<"DrawingRoutines: NOT IMPLEMENTED FUNCTIONALITY"<<std::endl;
		exit(0);
	}	
}


void DrawingRoutines::ClearImage(QImage & target, QRgb color)
{
	assert(qRed(color)>=0 && qRed(color) <=255
		   && qGreen(color)>=0 && qGreen(color) <=255
		   && qBlue(color)>=0 && qBlue(color) <=255);
	
	for (int i=0; i<target.width(); i++)
		for (int j = 0; j<target.height(); j++)
			target.setPixel(i, j, color);
}




