#include "LinearAlgebra.h"
////////// COMPLEX VECTOR	/////////
EntryType SymVectorComplex::GetEntryType()
{
	return SYM_ENTRY_COMPLEX;
}

SymVectorComplex::SymVectorComplex(int m)
{
	this->m = m;
	m_pData = gsl_vector_complex_alloc(m);
	m_bFreeData = true;
}

SymVectorComplex::~SymVectorComplex()
{
	if (m_bFreeData)
		gsl_vector_complex_free (m_pData);
}

SymComplex & SymVectorComplex::operator()(int i)
{
	return *(gsl_vector_complex_ptr (m_pData, i));
}

////////// REAL VECTOR	/////////

EntryType SymVectorReal::GetEntryType()
{
	return SYM_ENTRY_REAL;
}

SymVectorReal::SymVectorReal(gsl_vector * vec, bool shallowCopyOnly)
{
	m = vec->size;	
	if (shallowCopyOnly)
	{
		m_bFreeData = false;
		m_pData = vec;
	}
	else
	{
		m_bFreeData = true;
		m_pData = gsl_vector_alloc(m);
		CopyValuesFrom(vec);
	}
}

SymVectorReal::SymVectorReal(int m, double initVal)
{
	m_bFreeData = true;
	this->m = m;
	m_pData = gsl_vector_alloc(m);
	gsl_vector_set_all (m_pData, initVal);
}

SymVectorReal::SymVectorReal(int m)
{
	m_bFreeData = true;
	this->m = m;
	m_pData = gsl_vector_alloc(m);
	gsl_vector_set_zero (m_pData);
}

SymVectorReal::SymVectorReal(QString inFile)
{
	FILE * file = fopen(inFile.toStdString().c_str(), "rb");
	assert(file!=NULL);
	fscanf(file, "%d", &m);
	assert(m>0);
	m_bFreeData = true;
	m_pData = gsl_vector_alloc(m);	
	LoadSymVector(file);
}

SymVectorReal::SymVectorReal(std::vector<double> & inVector)
{
	m = inVector.size();
	m_bFreeData = true;
	m_pData = gsl_vector_alloc(m);	
	for (int i=0; i<(int)inVector.size(); i++)
		(*this)(i) = inVector[i];
}

void SymVectorReal::SaveSymVector(QString outFile)
{
	FILE * file = fopen(outFile.toStdString().c_str(), "wb");
	fprintf(file, "%d", m);
	gsl_vector_fwrite(file, m_pData);
	fclose(file);
}

void SymVectorReal::LoadSymVector(FILE * file)
{
	gsl_vector_fread(file, m_pData);	
}

std::vector<double> SymVectorReal::ToVector()
{
	std::vector<double> retvec;
	for (int i=0; i<m; i++)
		retvec.push_back((*this)(i));
	return retvec;
}

void SymVectorReal::CopyValuesFrom(gsl_vector * vec)
{
	gsl_vector_memcpy(m_pData, vec);
}


SymVectorReal & SymVectorReal::operator=(SymVectorReal & vec)
{
	gsl_vector_memcpy(m_pData, vec.m_pData);
	return *this;
}

void SymVectorReal::Add(SymVectorReal & v2)
{
	assert(v2.m==m);
	for (int i=0; i<m; i++)
		(*this)(i) += v2(i);
}

void SymVectorReal::SetAll(double val)
{
	for (int i=0; i<m; i++)
		(*this)(i) = val;	
}

void SymVectorReal::PrintVector(double absMin)
{
	std::cout<<"[ ";
	for (int i=0; i<m; i++)
	{
		if (qAbs((*this)(i)) < absMin)
			std::cout<<0<<" ";
		else
			std::cout<<(*this)(i)<<" ";			

	}
	std::cout<<"]"<<std::flush;
}

SymVectorReal::~SymVectorReal()
{
	if (m_bFreeData)
		gsl_vector_free (m_pData);
}

double & SymVectorReal::operator()(int i)
{
	return *(gsl_vector_ptr (m_pData, i));
}

double SymVectorReal::SumElements(double p)
{
	double currSum=0;
	for (int i=0; i<m; i++)
		currSum += pow((*this)(i), p);
	return currSum;
}

void SymVectorReal::Multiply(double val)
{
	for (int i=0; i<m; i++)
		(*this)(i) *= val;
}

////////// COMPLEX MATRIX	/////////
SymMatrix2DComplex::SymMatrix2DComplex(int m, int n)
: SymMatrix2D(m, n)
{
	m_pData = gsl_matrix_complex_alloc (m, n);
	gsl_matrix_complex_set_zero(m_pData);
	m_bFreeData = true;
}

SymMatrix2DComplex::~SymMatrix2DComplex()
{
	if (m_bFreeData)
		gsl_matrix_complex_free(m_pData);
}

void SymMatrix2DComplex::PrintMatrix(double absMax)
{
	for (int i=0; i<m; i++)
	{
		std::cout<<"[\t";		
		for (int j=0; j<n; j++)
		{
			SymComplex val = (*this)(i,j);
			if (sqrt(val.dat[0]*val.dat[0]+val.dat[1]*val.dat[1])<absMax)
				std::cout<<"0*\t";
			else
			{
				if (val.dat[1]>0)
					std::cout<<val.dat[0]<<"+"<<val.dat[1]<<"i"<<"\t";
				else if (val.dat[1]<0)
					std::cout<<val.dat[0]<<val.dat[1]<<"i"<<"\t";
				else
					std::cout<<val.dat[0]<<"\t";
			}
				
		}
		std::cout<<"]\n";
	}
	std::cout<<std::endl;
}

double SymMatrix2DComplex::GetMax(bool absVal)
{
	double maxVal = (*this)(0, 0).dat[0];
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			if (absVal)
			{
				double dx = (*this)(i, j).dat[0];
				double dy = (*this)(i, j).dat[1];
				maxVal = qMax(maxVal, sqrt(dx*dx+dy*dy));
			}
			else
				maxVal = qMax(maxVal, (*this)(i, j).dat[0]);
				
	return maxVal;
}

double SymMatrix2DComplex::GetMin(bool absVal)
{
	double minVal = sqrt((*this)(0, 0).dat[0]*(*this)(0, 0).dat[0] + (*this)(0, 0).dat[1]*(*this)(0, 0).dat[1]);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			if (absVal)
			{
				double dx = (*this)(i, j).dat[0];
				double dy = (*this)(i, j).dat[1];
				minVal = qMin(minVal, sqrt(dx*dx+dy*dy));
			}
			else
				minVal = qMin(minVal, (*this)(i, j).dat[0]);
	
	return minVal;	
}

QImage SymMatrix2DComplex::ToImage(bool )
{
	QImage img(m, n, QImage::Format_RGB32);
	double maxVal = GetMax(true);
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			int c = (qAbs((*this)(i,j).dat[0]) / maxVal) * 255;
			assert(c<=255 && c>=0);
			img.setPixel(i,j, qRgb(c,c,c));
		}
	}
	return img;
}

void SymMatrix2DComplex::Clear()
{
	gsl_matrix_complex_free(m_pData);
}

int SymMatrix2DComplex::GetNumDims()
{
	return 2;
}

EntryType SymMatrix2DComplex::GetEntryType()
{
	return SYM_ENTRY_COMPLEX;
}

void SymMatrix2DComplex::WriteRow(int row, gsl_complex_packed_array data)
{
	for (int i=0; i<n; i++)
		gsl_matrix_complex_set(m_pData, row, i, gsl_complex_rect(data[2*i + 0], data[2*i + 1]));
}

void SymMatrix2DComplex::WriteCol(int col, gsl_complex_packed_array data)
{
	for (int i=0; i<m; i++)
		gsl_matrix_complex_set(m_pData, i, col, gsl_complex_rect(data[2*i + 0], data[2*i + 1]));	
}

void SymMatrix2DComplex::ReadRow(int row, gsl_complex_packed_array data)
{
	for (int j=0; j<n; j++)
	{
		SymComplex val = (*this)(row, j);
		data[2*j+0] = val.dat[0];
		data[2*j+1] = val.dat[1];
	}		
}

void SymMatrix2DComplex::ReadCol(int col, gsl_complex_packed_array data)
{
	for (int j=0; j<m; j++)
	{
		SymComplex val = (*this)(j, col);
		data[2*j+0] = val.dat[0];
		data[2*j+1] = val.dat[1];
	}	
}


SymMatrix2DComplex SymMatrix2DComplex::operator*(SymMatrix2DComplex & m2)
{
	SymMatrix2DComplex C(m, m2.n);
	gsl_blas_zgemm (CblasNoTrans, CblasNoTrans, gsl_complex_rect (1, 0), m_pData, m2.m_pData, gsl_complex_rect (0, 0), C.m_pData);	
	return C;
}
		
void SymMatrix2DComplex::PointwiseMultiply(SymMatrix2DComplex & m2, SymMatrix2DComplex & trg)
{
	gsl_matrix_complex_set_zero(trg.m_pData);
	gsl_matrix_complex_add(trg.m_pData, m2.m_pData);
	gsl_matrix_complex_mul_elements(trg.m_pData, m_pData);
}

void SymMatrix2DComplex::Multiply(SymMatrix2DComplex & m2, SymMatrix2DComplex & trg)
{
	gsl_matrix_complex_set_zero(trg.m_pData);
	gsl_blas_zgemm (CblasNoTrans, CblasNoTrans, gsl_complex_rect (1, 0), m_pData, m2.m_pData, gsl_complex_rect (0, 0), trg.m_pData);	
}

void SymMatrix2DComplex::DivideBy(double val)
{
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			SymComplex & trgVal = (*this)(i,j);
			trgVal.dat[0] /= val;
			trgVal.dat[1] /= val;			
		}
	}
}

void SymMatrix2DComplex::Transpose()
{
	assert(m==n);	//if matrix is not square, use Transpose(trg) instead
	gsl_matrix_complex_transpose (this->m_pData);
}

void SymMatrix2DComplex::Transpose(SymMatrix2DComplex & trg)
{
	assert(m==trg.n && n==trg.m);
	gsl_matrix_complex_transpose_memcpy(trg.m_pData, m_pData);
}
		
SymComplex & SymMatrix2DComplex::operator()(int i, int j)
{
	return *(gsl_matrix_complex_ptr (m_pData, i, j));
}

SymMatrix2DComplex & SymMatrix2DComplex::operator=(SymMatrix2DComplex & src)
{
	assert(m==src.m && n==src.n);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			gsl_matrix_complex_set(m_pData, i, j, gsl_matrix_complex_get(src.m_pData, i, j));
	return *this;	
}

float * SymMatrix2DComplex::GetRealDataArray()
{
	float * array = new float[m*n];
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			array[j*m+i] = GSL_REAL(gsl_matrix_complex_get(m_pData, i, j));
		}
	}
	return array;
}

void SymMatrix2DComplex::Conj(SymMatrix2DComplex & trg)
{
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			gsl_matrix_complex_set(trg.m_pData, i, j, gsl_complex_conjugate(gsl_matrix_complex_get(m_pData, i, j)));
		}
	}	
}

////////// REAL MATRIX	/////////
SymMatrix2DReal::SymMatrix2DReal(int m, int n)
: SymMatrix2D(m,n)
{
	m_pData = gsl_matrix_alloc (m, n);
	gsl_matrix_set_zero(m_pData);
	m_pTempFilterGrid = NULL;
	m_pTempFilterOpenCVImg32F = NULL;
	m_bFreeData = true;
}

SymMatrix2DReal::SymMatrix2DReal(int m, int n, double initValue)
: SymMatrix2D(m,n)
{
	m_pData = gsl_matrix_alloc (m, n);
	gsl_matrix_set_all(m_pData, initValue);
	m_pTempFilterGrid = NULL;
	m_pTempFilterOpenCVImg32F = NULL;
	m_bFreeData = true;
}


SymMatrix2DReal::SymMatrix2DReal(SymMatrix2DReal & srcMatrix)
: SymMatrix2D(srcMatrix.m, srcMatrix.n)
{
	m_bFreeData = true;	
	m_pTempFilterGrid = NULL;
	m_pTempFilterOpenCVImg32F = NULL;
	m_pData = gsl_matrix_alloc (srcMatrix.m, srcMatrix.n);	
	for (int i=0; i<srcMatrix.m; i++)
		for (int j=0; j<srcMatrix.n; j++)	
			(*this)(i,j) = srcMatrix(i,j);
}

SymMatrix2DReal::SymMatrix2DReal(QImage & img)
: SymMatrix2D(img.width(), img.height())
{
	m_bFreeData = true;	
	m_pTempFilterGrid = NULL;
	m_pTempFilterOpenCVImg32F = NULL;
	m_pData = gsl_matrix_alloc (img.width(), img.height());	
	FromImage(img);
}

void SymMatrix2DReal::FromImage(QImage & img)
{
	assert(m==img.width() && n==img.height());
	for (int i=0; i<img.width(); i++)
		for (int j=0; j<img.height(); j++)	
			(*this)(i,j) = (double)qGray(img.pixel(i,j))/255.;	
}

SymMatrix2DReal::~SymMatrix2DReal()
{
	if (m_bFreeData)
		gsl_matrix_free (m_pData);
}

void SymMatrix2DReal::Clear()
{
	gsl_matrix_free(m_pData);
	if (m_pTempFilterGrid!=NULL)
		delete m_pTempFilterGrid;
	
	if 	(m_pTempFilterOpenCVImg32F!=NULL)
	{
		delete m_pTempFilterOpenCVImg32F;
		delete m_pTempProcessOpenCVImg32F;
		delete m_pTempFilterOpenCVImg8U;
		delete m_pTempProcessOpenCVImg8U;	
	}
}

int SymMatrix2DReal::GetNumDims()
{
	return 2;
}

EntryType SymMatrix2DReal::GetEntryType()
{
	return SYM_ENTRY_REAL;
}

void SymMatrix2DReal::PrintMatrix(double absMax)
{
	for (int i=0; i<m; i++)
	{
		std::cout<<"[\t";		
		for (int j=0; j<n; j++)
		{
			if (qAbs((*this)(i,j))<absMax)
				std::cout<<"0*\t";
			else
				std::cout<<(*this)(i,j)<<"\t";
		}
		std::cout<<"]\n";
	}
	std::cout<<std::endl;
}

double SymMatrix2DReal::GetMax(bool absVal)
{
	double maxVal = absVal ? qAbs((*this)(0, 0)) : (*this)(0, 0);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			if (absVal)
				maxVal = qMax(maxVal, qAbs((*this)(i, j)));
			else
				maxVal = qMax(maxVal, (*this)(i, j));
	return maxVal;
}

double SymMatrix2DReal::GetMin(bool absVal)
{
	double maxVal = absVal ? qAbs((*this)(0, 0)) : (*this)(0, 0);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			if (absVal)
				maxVal = qMin(maxVal, qAbs((*this)(i, j)));
			else
				maxVal = qMin(maxVal, (*this)(i, j));
	return maxVal;	
}

QImage SymMatrix2DReal::ToImage(bool shiftByMin)
{
//	shiftByMin = false;
	QImage img(m, n, QImage::Format_RGB32);
	double maxVal = GetMax(true);
//	std::cout<<"GetMax="<<GetMax(true)<<std::endl;
	double minVal = GetMin(false);
//	double maxColor = (*this)(0,0), minColor=(*this)(0,0);
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			int c;
			if (maxVal==minVal)
				c = 255;
			else if (minVal < 0 || shiftByMin)
				c = ((((*this)(i,j))-minVal) / (maxVal-minVal)) * 255;
			else
				c = ((qAbs((*this)(i,j))) / (maxVal)) * 255;
			
			//assert(c<=255 && c>=0);
			img.setPixel(i,j, qRgb(c,c,c));
		}
	}
	
	return img;
}

QImage SymMatrix2DReal::ClampToImage(double minVal, double maxVal)
{
	QImage img(m, n, QImage::Format_RGB32);
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			double val = (*this)(i,j);
			if (val < minVal)
				val = minVal;
			else if (val>maxVal)
				val = maxVal;
			int c = ((val-minVal) / (maxVal-minVal)) * 255;
			assert(c<=255 && c>=0);
			img.setPixel(i,j, qRgb(c,c,c));
		}
	}
	return img;	
}

void SymMatrix2DReal::WriteRow(int row, gsl_complex_packed_array data)
{
	for (int i=0; i<n; i++)
		gsl_matrix_set(m_pData, row, i, data[2*i + 0]);	
}

void SymMatrix2DReal::WriteCol(int col, gsl_complex_packed_array data)
{
	for (int i=0; i<m; i++)
		gsl_matrix_set(m_pData, i, col, data[2*i + 0]);		
}

SymMatrix2DReal & SymMatrix2DReal::operator=(SymMatrix2DComplex & src)
{
	assert(m==src.m && n==src.n);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
		{
			double real = gsl_matrix_complex_get(src.m_pData, i, j).dat[0];
			double imaginary = gsl_matrix_complex_get(src.m_pData, i, j).dat[1];
			gsl_matrix_set(m_pData, i, j, sqrt(real*real+imaginary*imaginary));
		}
	return *this;
}

SymMatrix2DReal & SymMatrix2DReal::operator=(SymMatrix2DReal & src)
{
	assert(m==src.m && n==src.n);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			gsl_matrix_set(m_pData, i, j, gsl_matrix_get(src.m_pData, i, j));
	return *this;	
}

void SymMatrix2DReal::ReadRow(int row, gsl_complex_packed_array data)
{
	double * rowData = gsl_matrix_ptr(m_pData, row, 0);
	gsl_fft_real_unpack (rowData, data, 1, n);
}

void SymMatrix2DReal::ReadCol(int col, gsl_complex_packed_array data)
{
	for (int j=0; j<m; j++)
	{
		data[2*j+0] = (*this)(j, col);
		data[2*j+1] = 0;
	}		
}

void SymMatrix2DReal::Subtract(double value)
{
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*this)(i,j) -= value;
}

void SymMatrix2DReal::Subtract(SymMatrix & otherMatrix)
{
	assert(otherMatrix.GetNumDims()==2 && otherMatrix.GetEntryType()==SYM_ENTRY_REAL);
	Subtract((SymMatrix2DReal &)otherMatrix);
}

void SymMatrix2DReal::Subtract(SymMatrix2DReal & otherMatrix)	
{
	assert(otherMatrix.m==m && otherMatrix.n==n);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*this)(i,j) -= otherMatrix(i,j);
//	std::cout<<"\tAfterSubtr: max="<<GetMax(false)<<", min="<<GetMin(false)<<std::endl;
}

void SymMatrix2DReal::ConstMinusValue(double c)
{
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*this)(i,j) = c - (*this)(i,j);
}

int SymMatrix2DReal::GetNumElements()
{
	return m*n;
}

void SymMatrix2DReal::Add(SymMatrix2DReal & otherMatrix)
{
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*this)(i,j) += otherMatrix(i,j);
}

SymMatrix2DReal SymMatrix2DReal::operator*(SymMatrix2DReal & m2)
{
	SymMatrix2DReal C(m, m2.n);
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1, m_pData, m2.m_pData, 0, C.m_pData);	
	return C;
}

void SymMatrix2DReal::PointwiseMultiply(SymMatrix2DReal & m2, SymMatrix2DReal & trg)
{
	gsl_matrix_set_zero(trg.m_pData);
	gsl_matrix_add(trg.m_pData, m2.m_pData);
	gsl_matrix_mul_elements(trg.m_pData, m_pData);
}

void SymMatrix2DReal::Multiply(SymMatrix2DReal & m2, SymMatrix2DReal & trg)
{
	gsl_matrix_set_zero(trg.m_pData);	
	gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1, m_pData, m2.m_pData, 1, trg.m_pData);	
}

void SymMatrix2DReal::Multiply(SymVectorReal & v, SymVectorReal & vTrg)
{
	gsl_vector_set_zero(vTrg.m_pData);
	gsl_blas_dgemv (CblasNoTrans, 1, m_pData, v.m_pData, 1, vTrg.m_pData);	
}

void SymMatrix2DReal::Transpose(SymMatrix2DReal & trg)
{
	assert(m==trg.n && n==trg.m);
	gsl_matrix_transpose_memcpy(trg.m_pData, m_pData);
}


void SymMatrix2DReal::Inverse(InverseMethod method, SymMatrix2DReal * workA)
{
	bool deleteWorkA = workA==NULL;
	if (workA==NULL)
		workA = new SymMatrix2DReal(m, n);
	if (workA!=this)
		*workA = *this;
		
	if (method==INV_LU)
	{
		assert(m==n);	//only valid for square matrices
		int signum;
		gsl_permutation * p = gsl_permutation_alloc(m);
		gsl_linalg_LU_decomp(workA->m_pData, p, &signum);
		gsl_linalg_LU_invert(workA->m_pData, p, m_pData);
		delete p;
	}
	else
	{
		assert(false);	//not implemented method
	}
	
	if (deleteWorkA)
		delete workA;
}

/**
 * Outputs: THIS = U * S * V' (note this is thin SVD)
 * @param U			must have same dimension as matrix (this): M x N
 * @param V			N x N orthogonal square matrix
 * @param S			N x N diagonal matrix represented by vector of length N
 * @param work		(optional) workspace vector of length N for Regular and Golub-Reinsch methods
 * @param workX		(optional) workspace N x N matrix for Golub-Reinsch method
 */
void SymMatrix2DReal::ThinSVD(SymMatrix2DReal * U, SymMatrix2DReal * V, SymVectorReal * S, 
							  SVDMethod method, SymVectorReal * work, SymMatrix2DReal * workX)
{
	assert(m==U->m && n==U->n);
	if (U!=this)
		*U = *this;
	
	if (method==SVD_REGULAR)
	{
		bool deleteWork=(work==NULL);
		if (work==NULL)
			work = new SymVectorReal(U->n);
		gsl_linalg_SV_decomp (U->m_pData, V->m_pData, S->m_pData, work->m_pData);
		if (deleteWork)
			delete work;
	}
	else if (method==SVD_GOLUB_REINSCH)
	{
		bool deleteWork=(work==NULL);
		bool deleteWorkX=(workX==NULL);
		if (work==NULL)
			work = new SymVectorReal(U->n);
		if (workX==NULL)
			workX = new SymMatrix2DReal(U->n, U->n);
		
		gsl_linalg_SV_decomp_mod (U->m_pData, workX->m_pData, V->m_pData, S->m_pData, work->m_pData);
		
		if (deleteWork)
			delete work;
		if (deleteWorkX)
			delete workX;
	}
	else if (method==SVD_JACOBI)
	{
		gsl_linalg_SV_decomp_jacobi (U->m_pData, V->m_pData, S->m_pData);
	}
	else
	{
		assert(false);
	}
}

void SymMatrix2DReal::DecomposeCholesky()
{
	gsl_linalg_cholesky_decomp(m_pData);
}

void SymMatrix2DReal::SolveUsingCholesky(SymVectorReal & b, SymVectorReal &x)
{
	gsl_linalg_cholesky_solve (m_pData, b.m_pData, x.m_pData);
}


void SymMatrix2DReal::PointwisePower(double p, bool absVal)
{
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			if (absVal)
				gsl_matrix_set(m_pData, i,j, pow(qAbs(gsl_matrix_get(m_pData, i, j)), p));				
			else
				gsl_matrix_set(m_pData, i,j, pow(gsl_matrix_get(m_pData, i, j), p));
}

double SymMatrix2DReal::SumElements(double p)
{
	double sum = 0;
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			sum += pow(qAbs(gsl_matrix_get(m_pData, i, j)), p);
	return sum;
}

double SymMatrix2DReal::Mean()
{
	double sum = 0;
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			sum += gsl_matrix_get(m_pData, i, j);
	return sum / (m*n);	
}

void SymMatrix2DReal::DivideBy(double value)
{
	assert(value!=0);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			gsl_matrix_set(m_pData, i,j, gsl_matrix_get(m_pData, i, j)/value);	
}


double & SymMatrix2DReal::operator()(int i, int j)
{
	return *(gsl_matrix_ptr (m_pData, i, j));
}

float * SymMatrix2DReal::GetRealDataArray()
{
//	return m_pData->data;
	float * array = new float[m*n];
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			array[j*m+i] = gsl_matrix_get(m_pData, i, j);
	return array;
}

double SymMatrix2DReal::CalculateDifference(SymMatrix2DReal & m2, double p)
{
	assert(m==m2.m && n==m2.n);		//dimension must agree
	double diff = 0;
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			diff += pow(qAbs((*this)(i,j)-m2(i,j)), p);
	
	return diff;
}

double SymMatrix2DReal::SampleBilinear(double x, double y)
{
    //[m, n, colors] = size(f);
	SymMatrix2DReal & f = (SymMatrix2DReal &)*this;
	int x1 = (int)floor(x);
	int x2 = (int)ceil(x);
	int y1 = (int)floor(y);
	int y2 = (int)ceil(y);
	double I = 0;
	
    if (y1>=0 && x1>=0 && x2<f.m && y2<f.n)
	{
        if (x2>x1 && y2>y1)
		{
            I = f(x1,y1)*(x2-x)*(y2-y) + f(x2,y1)*(x-x1)*(y2-y);
			I = I + f(x1,y2)*(x2-x)*(y-y1) + f(x2,y2)*(x-x1)*(y-y1);
		}
		else if (y1==y2 && x1==x2)
		{
			I = f(x1, y1);
		}
		else if (y1==y2)
		{
			double t = (x - x1) / (x2 - x1);
			I = t * f(x2, y1) + (1-t) * f(x1, y1);
		}
		else if (x1==x2)
		{
			double t = (y - y1) / (y2 - y1);
			I = t * f(x1, y2) + (1-t) * f(x1, y1);
		}
		
	}
	else if (y2>=0 && x2>=0 && x2<f.m && y2<f.n)
	{
		I = f(x2, y2);
	}
	else if (y1>=0 && x1>=0 && x1<f.m && y1<f.n)
	{
		I = f(x1, y1);
	}
	return I;	
}

void SymMatrix2DReal::CopySquare(SymMatrix2D * trg, int minSrcX, int minSrcY, int minTrgX, int minTrgY, int widthX, int widthY)
{
	assert(minSrcX>=0 && minSrcY>=0 && minSrcX+widthX<=m && minSrcY+widthY<=n);
	assert(minTrgX>=0 && minTrgY>=0 && minTrgX+widthX<=trg->m && minTrgY+widthY<=trg->n);
	assert(trg->GetEntryType()==SYM_ENTRY_REAL);
	for (int i=0; i<widthX; i++)
	{
		for (int j=0; j<widthY; j++)
		{
			((SymMatrix2DReal&)*trg)(minTrgX+i, minTrgY+j) = (*this)(minSrcX+i, minSrcY+j);
		}
	}
}

//filtering
void SymMatrix2DReal::FillGrid()
{
	if (m_pTempFilterGrid==NULL)
		m_pTempFilterGrid = new R2Grid(m, n);
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*m_pTempFilterGrid)(i,j) = (*this)(i,j);
}

void SymMatrix2DReal::CopyFromGrid()
{
	for (int i=0; i<m; i++)
		for (int j=0; j<n; j++)
			(*this)(i,j) = (*m_pTempFilterGrid)(i,j);	
}

void SymMatrix2DReal::GaussBlur(double std_dev)
{
	FillGrid();
	m_pTempFilterGrid->Blur(std_dev);
	CopyFromGrid();
}

void SymMatrix2DReal::InitOpenCV()
{
	if (m_pTempFilterOpenCVImg32F==NULL)
	{
		m_pTempFilterOpenCVImg32F = cvCreateImage(cvSize(m, n), IPL_DEPTH_32F, 1);
		m_pTempProcessOpenCVImg32F = cvCreateImage(cvSize(m, n), IPL_DEPTH_32F, 1);
		m_pTempFilterOpenCVImg8U = cvCreateImage(cvSize(m, n), IPL_DEPTH_8U, 1);
		m_pTempProcessOpenCVImg8U = cvCreateImage(cvSize(m, n), IPL_DEPTH_8U, 1);
	}
}

void SymMatrix2DReal::FillOpenCV(IplImage * trg, double scale)
{
//	std::cout<<"\tCreating array"<<std::endl;
	assert(trg!=NULL);
	CvScalar s;	
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
//			std::cout<<"\tFilling: "<<i<<", "<<j<<std::endl;
			s.val[0] = (*this)(i,j) * scale;
			cvSet2D(trg, j, i, s);
		}
	}
}

void SymMatrix2DReal::CopyFromOpenCV(IplImage * src, double scale)
{
	assert(src!=NULL);
	CvScalar s;
	for (int i=0; i<m; i++)
	{
		for (int j=0; j<n; j++)
		{
			s = cvGet2D(src, j, i);
			(*this)(i,j) = s.val[0]*scale;
		}
	}	
}

void SymMatrix2DReal::CannyFilterOpenCV(double threshold1, double threshold2, int window)
{
	InitOpenCV();
	FillOpenCV(m_pTempFilterOpenCVImg8U, 255.);
	std::cout<<"Canny Thresh="<<threshold1<<", "<<threshold2<<std::endl;
	cvCanny(m_pTempFilterOpenCVImg8U, m_pTempProcessOpenCVImg8U, 
			threshold1, threshold2, window);
	CopyFromOpenCV(m_pTempProcessOpenCVImg8U, 1./255.);
}

void SymMatrix2DReal::SobelFilterOpenCV(int window)
{
	InitOpenCV();	
	FillOpenCV(m_pTempFilterOpenCVImg32F);
	cvSobel( m_pTempFilterOpenCVImg32F, m_pTempProcessOpenCVImg32F, 1, 0, window);
	CopyFromOpenCV(m_pTempProcessOpenCVImg32F);
	PointwisePower(1, true);	
	DivideBy(GetMax(true));
}

void SymMatrix2DReal::BlurFilterOpenCV(double stdDev)
{
	InitOpenCV();	
	FillOpenCV(m_pTempFilterOpenCVImg32F);	
	int filter_radius = (int) (3 * stdDev + 0.5);	//copied from GAPS
	if (filter_radius%2==0)
		filter_radius++;
	cvSmooth(m_pTempFilterOpenCVImg32F, m_pTempProcessOpenCVImg32F, CV_GAUSSIAN, filter_radius, 0, stdDev);
	CopyFromOpenCV(m_pTempProcessOpenCVImg32F);	
}

