#include "DeformImageWarpLinearMesh.h"
#include "DrawingRoutines.h"
#include <iostream>
#include <time.h>
#include "CVLib.h"

ImageWarpLinearMesh::ImageWarpLinearMesh(QImage & img, int  maxWidth, int maxHeight, 
										 std::vector<ImgVertex> &vertices, std::vector<ImgFace> &faces,
										 bool inverse, bool fixBoundary, double paramStep, 
										 double minParamStep, ImagePenalization penalization, 
										 ImageHierarchyDescriptor * hierarchyDescr, double zoomWindow)
: ImageWarp(img)
{
	std::cout<<"Creating warp linear mesh"<<std::endl;
	m_fZoomWindow = zoomWindow;
	m_iMaxWidth = maxWidth;
	m_iMaxHeight = maxHeight;	
	InitializeBuffers(img);
	m_bRetrievingImage = false;
	m_bPreserveScale = true;
	m_bShowWireframe = false;
	m_HierarchyDescriptor = hierarchyDescr;
	m_SeedingType = IMAGE_SEED_MANUAL;
	m_NumCtrlPnts[0] = (int)vertices.size()/2;
	m_NumCtrlPnts[1] = 0;	
	m_fInitialParamStep = paramStep;
	m_fMinParamStep = minParamStep;
	m_Penalization = penalization;	
	m_bFixBoundary = fixBoundary;
	m_bInverseMapping = inverse;
		
	InitializeMesh(img, vertices, faces, m_bInverseMapping);

	//initialize hierarchy / fill parameters
	assert(hierarchyDescr==NULL || hierarchyDescr->m_HierarchyType!=OVERLAP_CHILDREN);	//TODO: support bilinear hierarchy
	int numParams = vertices.size()/2;
	std::cout<<"hierarchy"<<std::endl;
	CreateHierarchy(&numParams, hierarchyDescr, paramStep, minParamStep);	
	std::cout<<"hierarchy - done"<<std::endl;
}

ImageWarpLinearMesh::ImageWarpLinearMesh(QImage & img, int  maxWidth, int maxHeight, int numPoints, 
										 bool inverse, bool fixBoundary, double paramStep, 
										 double minParamStep, ImagePenalization penalization, 
										 ImageHierarchyDescriptor * hierarchyDescr, double zoomWindow,
										 bool addMorePnts)
: ImageWarp(img)
{
	std::cout<<"Creating warp linear mesh - feature pnts"<<std::endl;	
	m_fZoomWindow = zoomWindow;	
	m_iMaxWidth = maxWidth;
	m_iMaxHeight = maxHeight;	
	InitializeBuffers(img);
	m_bRetrievingImage = false;
	m_bPreserveScale = true;
	m_bShowWireframe = false;
	m_HierarchyDescriptor = hierarchyDescr;
	m_SeedingType = IMAGE_SEED_FEATURE_PNTS;
	m_NumCtrlPnts[0] = numPoints;
	m_NumCtrlPnts[1] = 0;	
	m_fInitialParamStep = paramStep;
	m_fMinParamStep = minParamStep;
	m_Penalization = penalization;	
	m_bFixBoundary = fixBoundary;
	m_bInverseMapping = inverse;
	
	std::vector<double> vertices;
	std::vector<int> faces;
	std::cout<<"finding feat pnts"<<std::endl;
	CVLibRoutines::FindFeatureMesh(m_Source, numPoints, vertices, faces, CVLibRoutines::TRIANGULATE_DELAUNAY, addMorePnts);
	std::cout<<"finding feat pnts - DONE"<<std::endl;
	
	//initialize vertices
	std::vector<ImgVertex> imgVertices;
	for (int i=0; i<(int)vertices.size(); i+=2)
	{
		ImgVertex v;
		v.x = vertices[i+0];		v.y = vertices[i+1];		
		v.tu = vertices[i+0];		v.tv = vertices[i+1];		
		
		v.m_BoundaryVertexFlags = ImgVertex::BND_VERTEX_NOT_FIXED;
		if (v.x==0 || v.x==img.width())
		{
			v.m_BoundaryVertexFlags |= ImgVertex::BND_VERTEX_FIXED_X;
			v.m_fBoundaryX = v.x;
		}
		if (v.y==0 || v.y==img.height())
		{
			v.m_BoundaryVertexFlags |= ImgVertex::BND_VERTEX_FIXED_Y;				
			v.m_fBoundaryY = v.y;
		}
		
		v.m_iParamID = i/2;
		imgVertices.push_back(v);
	}
	
	//initialize faces
	std::vector<ImgFace> imgFaces;
	for (int i=0; i<(int)faces.size(); i+=3)
	{
		ImgFace f;
		f.v1 = faces[i+0];		f.v2 = faces[i+1];		f.v3 = faces[i+2];			f.v4 = -1;
		imgFaces.push_back(f);
	}
	
	InitializeMesh(img, imgVertices, imgFaces, m_bInverseMapping);
	
	//initialize hierarchy / fill parameters
	std::cout<<"feat pnts hierarchy"<<std::endl;	
	assert(hierarchyDescr==NULL || hierarchyDescr->m_HierarchyType!=OVERLAP_CHILDREN);	//TODO: support bilinear hierarchy
	int numParams = vertices.size();
	CreateHierarchy(&numParams, hierarchyDescr, paramStep, minParamStep);
	std::cout<<"feat pnts hierarchy - DONE!"<<std::endl;	
}


ImageWarpLinearMesh::ImageWarpLinearMesh(QImage & img, int  maxWidth, int maxHeight, 
										 int numRegularCtrlPnts[2], bool inverse, bool fixBoundary, 
										 double paramStep, double minParamStep, ImagePenalization penalization, 
										 ImageHierarchyDescriptor * hierarchyDescr, FaceType faceType,
										 double zoomWindow)
: ImageWarp(img)
{	
	m_fZoomWindow = zoomWindow;	
	m_iMaxWidth = maxWidth;
	m_iMaxHeight = maxHeight;	
	InitializeBuffers(img);
	m_bRetrievingImage = false;
	m_bPreserveScale = true;
	m_bShowWireframe = false;
	m_FaceType = faceType;
	m_HierarchyDescriptor = hierarchyDescr;
	m_SeedingType = IMAGE_SEED_REGULAR_LATTICE;
	m_fInitialParamStep = paramStep;
	m_fMinParamStep = minParamStep;
	
	m_bFixBoundary = fixBoundary;
	m_bInverseMapping = inverse;
	int numParams = (numRegularCtrlPnts[0]) * (numRegularCtrlPnts[1]) * 2;

	m_Penalization = penalization;

	m_NumCtrlPnts[0] = numRegularCtrlPnts[0];	m_NumCtrlPnts[1] = numRegularCtrlPnts[1];
	
	std::vector<ImgVertex> vertices;
	double px, py, tu, tv;
	ImgVertex v;	
	
	//Initialize vertices
	std::cout<<"Initializing vertices"<<std::endl;	
	for (int i=0; i<numRegularCtrlPnts[0]; i++)
	{
		for (int j=0; j<numRegularCtrlPnts[1]; j++)		
		{
			tu = (double)i / (double)(numRegularCtrlPnts[0]-1);
			px = tu * (double)img.width();
			tv = (double)j / (double)(numRegularCtrlPnts[1]-1);
			py = tv * (double)img.height();
			v.tu = px;	 v.tv = py;
			v.x  = px;	 v.y  = py;
			v.m_BoundaryVertexFlags = ImgVertex::BND_VERTEX_NOT_FIXED;
			if (v.x==0 || v.x==img.width())
			{
				v.m_BoundaryVertexFlags |= ImgVertex::BND_VERTEX_FIXED_X;
				v.m_fBoundaryX = v.x;
			}
			if (v.y==0 || v.y==img.height())
			{
				v.m_BoundaryVertexFlags |= ImgVertex::BND_VERTEX_FIXED_Y;				
				v.m_fBoundaryY = v.y;
			}

			//assume that all vertex coords are in the beginning of params
			v.m_iParamID = (i * (numRegularCtrlPnts[1]) + j);

			vertices.push_back(v);
		}
	}
	//Initialize faces
	std::vector<ImgFace> faces;
	ImgFace face1, face2;		
	for (int i=0; i<numRegularCtrlPnts[0]-1; i++)
	{
		for (int j=0; j<numRegularCtrlPnts[1]-1; j++)
		{
			int v1 = i * (numRegularCtrlPnts[1]) + j;
			int v2 = (i+1) * (numRegularCtrlPnts[1]) + j;
			int v3 = (i+1) * (numRegularCtrlPnts[1]) + (j+1);
			int v4 = i * (numRegularCtrlPnts[1]) + (j+1);				
			if (faceType==FACE_TRIANGLE1 || (faceType==FACE_TRIANGLE3 && (i+j)%2==0))
			{
				face2.v1 = v1;				face2.v2 = v2;			face2.v3 = v4;			face2.v4 = -1;
				faces.push_back(face2);
				face1.v1 = v2;				face1.v2 = v3;			face1.v3 = v4;			face1.v4 = -1;				
			}
			else if (faceType==FACE_TRIANGLE2 || (faceType==FACE_TRIANGLE3 && (i+j)%2==1))
			{
				face2.v1 = v1;				face2.v2 = v2;			face2.v3 = v3;			face2.v4 = -1;
				faces.push_back(face2);
				face1.v1 = v3;				face1.v2 = v4;			face1.v3 = v1;			face1.v4 = -1;				
			}
			else if (faceType==FACE_QUAD)
			{
				face1.v1 = v1;				face1.v2 = v2;			face1.v3 = v3;			face1.v4 = v4;
			}
			faces.push_back(face1);			
		}
	}
	
	//generate mesh
	InitializeMesh(img, vertices, faces, inverse);
	
	//initialize hierarchy / fill parameters
	CreateHierarchy(&numParams, hierarchyDescr, paramStep, minParamStep);	
}

DeformationModel * ImageWarpLinearMesh::CloneSelfUsingBestParams(SymVectorReal & bestParams)
{
	bool showWireframe = m_bShowWireframe;
	m_bShowWireframe = false;
	DeformAtCurrentHierarchyLevel(bestParams);
	QImage newImg = GetDeformedImage();
	
	//TODO: allow other seeding types and hierarchies
	assert(m_HierarchyDescriptor!=NULL && m_HierarchyDescriptor->m_HierarchyType==AS_RIGID_AS_POSSIBLE);		
	ImageWarpLinearMesh * retModel;	
	if (m_SeedingType==IMAGE_SEED_REGULAR_LATTICE)
	{
		retModel = new ImageWarpLinearMesh(newImg, m_iMaxWidth, m_iMaxHeight, m_NumCtrlPnts, m_bInverseMapping, m_bFixBoundary, m_fInitialParamStep, 
								   m_fMinParamStep, m_Penalization, m_HierarchyDescriptor, m_FaceType);
		retModel->m_bShowWireframe = showWireframe;
	}
	else if (m_SeedingType==IMAGE_SEED_FEATURE_PNTS)
	{		
		retModel = new ImageWarpLinearMesh(newImg, m_iMaxWidth, m_iMaxHeight, m_NumCtrlPnts[0], m_bInverseMapping, m_bFixBoundary, m_fInitialParamStep, 
										   m_fMinParamStep, m_Penalization, m_HierarchyDescriptor);
		retModel->m_bShowWireframe = showWireframe;		
	}
	else
	{
		assert(false);
	}
	return retModel;
}

void ImageWarpLinearMesh::ChildUpdateAtHierarchySwitch(int currLevel, int newLevel)
{
	assert(currLevel>=-1 && newLevel>=-1);
	if (m_HierarchyDescriptor->m_HierarchyType==AS_RIGID_AS_POSSIBLE)
	{
		*m_pAllLastParams = *m_pFinalParams;
		if (m_bPolarCoords)	
			ConvertCart2Polar(*m_pAllLastParams, *m_pAllLastParams);
	}
	else if (m_HierarchyDescriptor->m_HierarchyType==OVERLAP_CHILDREN)
	{
//		*m_pAllLastParams = *m_pFinalParams;
//		InitializeParameterHierearchy(m_pHierarchyTreeRoots, *m_pAllLastParams, HIERARCHY_TRAVERSE_BREADTH, 
//									  WEIGHTS_BILINEAR, PARENTSLOC_KEEP_AS_IT_IS);		
	}
}

void ImageWarpLinearMesh::CreateHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, 
										  double paramStep, double minParamStep)
{
	if (hierarchyDescr!=NULL)
	{
//		hierarchyDescr->m_iInitNumPntsX = numRegularCtrlPnts[0];
//		hierarchyDescr->m_iInitNumPntsY = numRegularCtrlPnts[1];	//should be initialized in params
//		assert(hierarchyDescr->m_iInitNumPntsX>0 && hierarchyDescr->m_iInitNumPntsY>0)
		std::cout<<"Creating hierarchy: "<<std::endl;
		
		m_pHierarchyTreeRoots.clear();
		if (hierarchyDescr->m_HierarchyType==CONST_CHILDREN || hierarchyDescr->m_HierarchyType==OVERLAP_CHILDREN)
			CreateLinearWeightsHierarchy(numNodes, hierarchyDescr, paramStep, minParamStep);	// * see below
		else if (hierarchyDescr->m_HierarchyType==AS_RIGID_AS_POSSIBLE)
			CreateAsRigidAsPossibleHierarchy(numNodes, hierarchyDescr, paramStep, minParamStep);	// * see below
		
		m_pFinalParams = new SymVectorReal(m_pAllLastParams->m);
		*m_pFinalParams = *m_pAllLastParams;
		
		if (hierarchyDescr->m_HierarchyType==OVERLAP_CHILDREN)
		{
			assert(hierarchyDescr->m_iAssocX==4 && hierarchyDescr->m_iAssocY==4 
				   && hierarchyDescr->m_iOverlapX==2 && hierarchyDescr->m_iOverlapY==2);
			//for now only bilinear method is supported and it can only handle 4x4 neighborhoods for each pnt
			assert(m_pHierarchyTreeRoots.size()>0);
			InitializeParameterHierearchy(m_pHierarchyTreeRoots, *m_pAllLastParams, HIERARCHY_TRAVERSE_BREADTH, 
										  WEIGHTS_BILINEAR, PARENTSLOC_KEEP_AS_IT_IS);
		}
		else if (hierarchyDescr->m_HierarchyType==AS_RIGID_AS_POSSIBLE)
		{
			InitializeParameterHierearchy(m_pHierarchyTreeRoots, *m_pAllLastParams, HIERARCHY_TRAVERSE_NONE, 
										  WEIGHTS_BILINEAR, PARENTSLOC_KEEP_AS_IT_IS);
		}
	}
	else	//just fill parameters if no hierarchy (i.e. flat hierarchy)
	{
		m_pAllMinParamStep = new SymVectorReal(*numNodes, minParamStep);
		m_pAllParamStep = new SymVectorReal(*numNodes, paramStep);
		m_pAllLastParams = new SymVectorReal(*numNodes);		
		
		m_ParamPrintNodes.clear();
		std::vector<int> nodes;		nodes.push_back(0);		nodes.push_back(1);
		m_ParamPrintNodes.push_back(HierarchyPrintNode("", nodes, 0));
		for (int i=0; i<(int)m_Vertices.size(); i++)
		{
			((SymVectorReal&)*m_pAllLastParams)(2 * m_Vertices[i].m_iParamID+0) = m_Vertices[i].x;
			((SymVectorReal&)*m_pAllLastParams)(2 * m_Vertices[i].m_iParamID+1) = m_Vertices[i].y;			
		}		
	}
	m_pInitParams = m_pAllLastParams;
	m_pParameterSteps = m_pAllParamStep;
	m_pMinimalParamSteps = m_pAllMinParamStep;
		
	// * Note: CreateHierarchy function MUST fill:
	//	 m_ParamPrintNodes, m_pAllParamStep, m_pAllMinParamStep, m_pAllLastParams	
	//		if using breadth-first traversal (i.e. linear parent/child weights) fill:
	//			m_pHierarchyTreeRoots
	//		if not using breadth-first traversal fill:
	//			m_iLayerToFirstElement, m_iLayerToMaxElements, m_iMaxHierarchyLevels, m_pHierarchyBreadthFirstList 
	// You als MAY fill:
	//		m_PrecomputedSingleLayers - e.g. level descriptors
	//You can assume that triangulation has been initialized
}

void ImageWarpLinearMesh::CreateLinearWeightsHierarchy(int * numNodes, ImageHierarchyDescriptor * hierarchyDescr, 
													   double paramStep, double minParamStep)
{	
	*numNodes = 0;
	double paramStepX = paramStep;
	double paramStepY = paramStep;
	double minParamStepX = minParamStep;
	double minParamStepY = minParamStep;
	int assocX = hierarchyDescr->m_iAssocX;
	int assocY = hierarchyDescr->m_iAssocY;
	int overlapX = hierarchyDescr->m_HierarchyType==CONST_CHILDREN ? 0 : hierarchyDescr->m_iOverlapX;
	int overlapY = hierarchyDescr->m_HierarchyType==CONST_CHILDREN ? 0 : hierarchyDescr->m_iOverlapY;	
	int stepX = assocX - overlapX;
	int stepY = assocY - overlapY;
	assert(stepX!=0 && stepY!=0);
	int currNumX = hierarchyDescr->m_iInitNumPntsX;
	int currNumY = hierarchyDescr->m_iInitNumPntsY;
	*numNodes = 2*currNumX * currNumY;
	
	std::vector<DeformationHierarchyNode *> childrenField;
	std::vector<DeformationHierarchyNode *> parentField;
	std::vector<double> stepSizes;
	std::vector<double> minStepSizes;	
	
	int depth=0;

	m_ParamPrintNodes.clear();
	std::vector<int> nodes;		nodes.push_back(0);		nodes.push_back(1);
	m_ParamPrintNodes.push_back(HierarchyPrintNode("", nodes, depth));
	
	for (int i=0; i<currNumX; i++)	//initially fill with current field at the finest resolution
	{
		for (int j=0; j<currNumY; j++)
		{
			DeformationHierarchyNode * child = new DeformationHierarchyNode();
			child->m_ParamIDs.push_back(2*(i * currNumY + j)+0);
			child->m_ParamIDs.push_back(2*(i * currNumY + j)+1);			
			
			if (hierarchyDescr->m_HierarchyType==CONST_CHILDREN)
				child->m_RelationshipToParent = DeformationHierarchyNode::HIERARCH_REL_ADD;
			else if (hierarchyDescr->m_HierarchyType==OVERLAP_CHILDREN)
				child->m_RelationshipToParent = DeformationHierarchyNode::HIERARCH_REL_OFFSET_FROM_AVERAGE;
			childrenField.push_back(child);

			stepSizes.push_back(paramStepX);
			stepSizes.push_back(paramStepY);
			minStepSizes.push_back((double)m_Source.width());
			minStepSizes.push_back((double)m_Source.height());
		}
	}

	int childrenFieldDimX = currNumX;
	int childrenFieldDimY = currNumY;

	std::vector<int> allParentDimsX;
	std::vector<int> allParentDimsY;	
	std::vector<int> allParentsInitParamMap;
	allParentDimsX.push_back(currNumX);
	allParentDimsY.push_back(currNumY);	
	allParentsInitParamMap.push_back(0);		
	
	currNumX = ceil(((double)(currNumX - overlapX))/((double)stepX))+2;
	currNumY = ceil(((double)(currNumY - overlapY))/((double)stepY))+2;	
	
	std::cout<<"Created children: "<<childrenFieldDimX<<" x "<<childrenFieldDimY<<std::endl;
	while ((currNumX * currNumY) >= hierarchyDescr->m_iMinPnts		//recursively fill until latest field is too small
		   || (childrenFieldDimX<=currNumX && childrenFieldDimY<=currNumY) )			//or until parent's don't get smaller
	{
		std::cout<<"Creating new parent ring: "<<currNumX<<" x "<<currNumY<<std::endl;
		depth++;		
		paramStepX *= stepX;
		paramStepY *= stepY;

		allParentDimsX.push_back(currNumX);
		allParentDimsY.push_back(currNumY);	
		allParentsInitParamMap.push_back(*numNodes);		
		
		std::vector<int> nodes;		nodes.push_back(stepSizes.size());		nodes.push_back(stepSizes.size()+1);
		m_ParamPrintNodes.push_back(HierarchyPrintNode("", nodes, depth));
		
			//generate new field (level up) of parents
		parentField.clear();			
		for (int i=0; i<currNumX; i++)		
		{
			for (int j=0; j<currNumY; j++)
			{
				DeformationHierarchyNode * newNode = new DeformationHierarchyNode();
				newNode->m_ParamIDs.push_back(*numNodes + 2*(i * currNumY + j)+0);
				newNode->m_ParamIDs.push_back(*numNodes + 2*(i * currNumY + j)+1);				
				if (hierarchyDescr->m_HierarchyType==CONST_CHILDREN)
					newNode->m_RelationshipToParent = DeformationHierarchyNode::HIERARCH_REL_ADD;
				else if (hierarchyDescr->m_HierarchyType==OVERLAP_CHILDREN)
					newNode->m_RelationshipToParent = DeformationHierarchyNode::HIERARCH_REL_OFFSET_FROM_AVERAGE;
				parentField.push_back(newNode);
				
					//fill parent's children
				std::vector<int> xChildIds = GetChildrenIndices(0, ((double)i-1)/(double)(currNumX-1), ((double)i+1)/(double)(currNumX-1), 1, childrenFieldDimX);
				std::vector<int> yChildIds = GetChildrenIndices(0, ((double)j-1)/(double)(currNumY-1), ((double)j+1)/(double)(currNumY-1), 1, childrenFieldDimY);
				
				for (int childX = 0; childX<(int)xChildIds.size(); childX++)
					for (int childY = 0; childY<(int)yChildIds.size(); childY++)
						newNode->m_apChildren.push_back(childrenField[xChildIds[childX] * childrenFieldDimY + yChildIds[childY]]);
				
				
				//fill step size
				stepSizes.push_back(paramStepX);
				stepSizes.push_back(paramStepY);	
				minStepSizes.push_back(minParamStepX);
				minStepSizes.push_back(minParamStepY);
			}
		}
		
		childrenField.clear();		childrenField = parentField;
		*numNodes += 2*currNumX * currNumY;		
		childrenFieldDimX = currNumX;
		childrenFieldDimY = currNumY;		
		if (currNumX==3 && currNumY==3)
		{
			currNumX = 0;
			currNumY = 0;
		}
		else
		{
			currNumX = ceil(((double)currNumX - overlapX)/((double)assocX))+2;
			currNumY = ceil(((double)currNumY - overlapY)/((double)assocY))+2;
		}

		std::cout<<"Current # nodes = "<<*numNodes<<std::endl;
	}
	
	m_pHierarchyTreeRoots.clear();
	for (int i=0; i<(int)parentField.size(); i++)
		m_pHierarchyTreeRoots.push_back(parentField[i]);
	
	m_pAllParamStep = new SymVectorReal(*numNodes);
	m_pAllMinParamStep = new SymVectorReal(*numNodes);
	m_pAllLastParams = new SymVectorReal(*numNodes);
	
	for (int i=0; i<(int)stepSizes.size(); i++)
	{
		(*m_pAllParamStep)(i) = stepSizes[i];
		(*m_pAllMinParamStep)(i) = minStepSizes[i];
	}
	
	//Fill parents' values
	for (int l=0; l<(int)allParentDimsX.size(); l++)
	{
		for (int x=0; x<allParentDimsX[l]; x++)
		{
			for (int y=0; y<allParentDimsY[l]; y++)
			{
				int paramID = allParentsInitParamMap[l] + 2*(x * allParentDimsY[l] + y);
				(*m_pAllLastParams)(paramID+0) = ((double)x * (double)m_Source.width())/((double)allParentDimsX[l]-1);
				(*m_pAllLastParams)(paramID+1) = ((double)y * (double)m_Source.height())/((double)allParentDimsY[l]-1);
			}
		}
	}		
}

std::vector<int> ImageWarpLinearMesh::GetChildrenIndices(double minX, double x1, double x2, double maxX, int numX)
{
	assert(x1<x2 && x2>minX && x1<maxX);
	x1 -= minX;
	x2 -= minX;
	maxX -= minX;
	bool minBrd = x1<=0;
	
	if (x1<0)
		x1 = 0;
	if (x2>maxX)
		x2 = maxX;
	double dx = maxX / (numX-1);
	double minIntX = x1/dx;
	double maxIntX = x2/dx;
	
	std::vector<int> indices;
		
	for (int i=ceil(minIntX); i<=floor(maxIntX); i++)
	{
		if (minBrd || i*dx!=x1)		//additional check, since otherwise could add 9 neighbors!
			indices.push_back(i);
	}
	return indices;
}

void ImageWarpLinearMesh::InitializeBuffers(QImage & img)
{
	std::cout<<"init buffer"<<std::endl;
	m_HighResImg = img;
	if (m_iMaxWidth<img.width() || m_iMaxHeight<img.height())
	{
#ifdef USE_QT_FRAMEBUFFERS
		m_pHighResBuffer = InitializeBuffer(m_HighResTextureIds, false, img);
#else
		m_iHighResBuffer = InitializeBuffer(false, img);
#endif
		img = img.scaled(m_iMaxWidth, m_iMaxHeight, Qt::KeepAspectRatio, Qt::SmoothTransformation);
		
		m_Source = img;
		m_Deformed = img;
		delete m_pSourceMatrix;
		delete m_pDeformedMatrix;
		m_pSourceMatrix = new SymMatrix2DReal(img.width(), img.height());
		m_pDeformedMatrix =  new SymMatrix2DReal(img.width(), img.height());
		for (int i=0; i<img.width(); i++)
		{
			for (int j=0; j<img.height(); j++)
			{
				QRgb color = img.pixel(i,j);			
				double & val = (*m_pSourceMatrix)(i,j);
				val = 0.3*qRed(color)/255 + 0.59*qGreen(color)/255 + 0.11*qBlue(color)/255;
			}
		}		
	}
	else
	{
#ifdef USE_QT_FRAMEBUFFERS
		m_pHighResBuffer = NULL;
#else
		m_iHighResBuffer.m_aColorBuffer = NULL;
#endif
	}
	std::cout<<"init buffer - 1"<<std::endl;		
#ifdef USE_QT_FRAMEBUFFERS
	m_pFramebuffer = InitializeBuffer(m_TextureIds, true, img);
#else
	m_iFramebuffer = InitializeBuffer(true, img);
#endif
	std::cout<<"init buffer - DONE"<<std::endl;	
}

#ifdef USE_QT_FRAMEBUFFERS
QGLPixelBuffer * ImageWarpLinearMesh::InitializeBuffer(GLuint * textureIDs, bool initGrayscale, QImage & img)
{
	QGLFormat format;
	format.setSampleBuffers(true);
	format.setSamples(4);
	std::cout<<"qgl buffer"<<std::endl;
	QGLPixelBuffer * newBuffer = new QGLPixelBuffer(img.width(), img.height(), format);
	std::cout<<"qgl buffer - DONE"<<std::endl;
	newBuffer->makeCurrent();	
	
	//generate 2D openGL textures: color
	glGenTextures(initGrayscale ? 2 : 1, textureIDs);
	glBindTexture(GL_TEXTURE_2D, textureIDs[0]);
	img = img.convertToFormat (QImage::Format_RGB32); 
	glTexImage2D(GL_TEXTURE_2D, 0, 3, img.width(), img.height(), 
				 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, img.bits());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
	
	//generate 2D openGL textures: grayscale	
	if (initGrayscale)
	{
		float * data = m_pSourceMatrix->GetRealDataArray();
		glBindTexture(GL_TEXTURE_2D, textureIDs[1]);
		glTexImage2D(GL_TEXTURE_2D, 0, 1, img.width(), img.height(), 
					 0, GL_LUMINANCE, GL_FLOAT, data);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
		delete [] data;	
	}
	
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_FLAT);	//GL_SMOOTH
	glClearColor(1.0f, 1.0f, 1.0f, 1.);				
//	glClearColor(0.0f, 0.0f, 0.0f, 1.);					
	glClearDepth(1.0f);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glViewport(0,0, img.width(), img.height());
	gluOrtho2D(0, img.width(), 0, img.height());
	
	glScalef(m_fZoomWindow, -m_fZoomWindow, 1);
	glTranslatef(-(img.width()*(m_fZoomWindow-1.)/2.), -img.height() + (img.height()*(m_fZoomWindow-1.)/2.), 0);
	glMatrixMode(GL_MODELVIEW);		
	return newBuffer;
}
#else
BufferID ImageWarpLinearMesh::InitializeBuffer(bool initGrayscale, QImage & img)
{
	img.save("tempimg.png");
	std::cout<<"OpenGLBuffer: "<<img.width()<<" x "<<img.height()<<std::endl;
	BufferID newBuffer;
	newBuffer.m_iWidth = img.width();
	newBuffer.m_iHeight = img.height();
	std::cout<<"OpenGLBuffer - init"<<std::endl;
	// generate openGL buffers
	glGenRenderbuffersEXT(1, &newBuffer.m_iFrameBufferObject);
	
	std::cout<<"OpenGLBuffer: 2"<<std::endl;
	
	// attach depth / color to the buffer object
	glBindFramebufferEXT(GL_FRAMEBUFFER, newBuffer.m_iFrameBufferObject);
	
	glGenRenderbuffersEXT(1, &newBuffer.m_iDepthBuffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, newBuffer.m_iDepthBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, 
							 newBuffer.m_iWidth, newBuffer.m_iHeight);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, 
								 GL_RENDERBUFFER_EXT, newBuffer.m_iDepthBuffer);
	
	glGenRenderbuffersEXT(1, &newBuffer.m_iColorBuffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, newBuffer.m_iColorBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA, 
							 newBuffer.m_iWidth, newBuffer.m_iHeight);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
								 GL_RENDERBUFFER_EXT, newBuffer.m_iColorBuffer);


	std::cout<<"OpenGLBuffer: 3"<<std::endl;	
	newBuffer.m_aColorBuffer = new unsigned char[newBuffer.m_iWidth*newBuffer.m_iHeight*4];
	for (int i=0; i<newBuffer.m_iWidth*newBuffer.m_iHeight*4; i++)
		newBuffer.m_aColorBuffer[i] = 0;

	std::cout<<"OpenGLBuffer: 4"<<std::endl;		
	//generate 2D openGL textures: color
	GLuint textures[2];
	textures[0] = 0;
	textures[1] = 0;	
	glGenTextures(initGrayscale ? 2 : 1, textures);
	newBuffer.m_iTextureBuffer = textures[0];
	newBuffer.m_iTextureBufferGrayscale = textures[1];
	glBindTexture(GL_TEXTURE_2D, newBuffer.m_iTextureBuffer);
	img = img.convertToFormat (QImage::Format_RGB32); 
	glTexImage2D(GL_TEXTURE_2D, 0, 3, img.width(), img.height(), 
				 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, img.bits());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
	
	glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
							  GL_TEXTURE_2D, newBuffer.m_iTextureBuffer, 0 );

	std::cout<<"OpenGLBuffer: 5"<<std::endl;		
	//generate 2D openGL textures: grayscale	
	if (initGrayscale)
	{
		float * data = m_pSourceMatrix->GetRealDataArray();
		glBindTexture(GL_TEXTURE_2D, newBuffer.m_iTextureBufferGrayscale);
		glTexImage2D(GL_TEXTURE_2D, 0, 1, img.width(), img.height(), 
					 0, GL_LUMINANCE, GL_FLOAT, data);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
		delete [] data;	
		glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
								  GL_TEXTURE_2D, newBuffer.m_iTextureBufferGrayscale, 0 );		
	}
	
	std::cout<<"OpenGLBuffer: 7"<<std::endl;		
	assert(glCheckFramebufferStatusEXT(GL_FRAMEBUFFER)==GL_FRAMEBUFFER_COMPLETE);
	
	// setup camera
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_FLAT);	//GL_SMOOTH
	glClearColor(1.0f, 1.0f, 1.0f, 1.);				
	//	glClearColor(0.0f, 0.0f, 0.0f, 1.);					
	glClearDepth(1.0f);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glViewport(0,0, img.width(), img.height());
	gluOrtho2D(0, img.width(), 0, img.height());
	
	glScalef(m_fZoomWindow, -m_fZoomWindow, 1);
	glTranslatef(-(img.width()*(m_fZoomWindow-1.)/2.), -img.height() + (img.height()*(m_fZoomWindow-1.)/2.), 0);
	glMatrixMode(GL_MODELVIEW);		
	std::cout<<"OpenGLBuffer: 8"<<std::endl;		
	return newBuffer;
}
#endif

void ImageWarpLinearMesh::InitializeMesh(QImage & img, std::vector<ImgVertex> &vertices, std::vector<ImgFace> & faces, bool inverse)
{
	assert(img.width()>0 && img.height()>0);
	m_bInverseMapping = inverse;	
	
	//generate 2D mesh
	for (int i=0; i<(int)vertices.size(); i++)	//vertices
		m_Vertices.push_back(vertices[i]);
	
	for (int i=0; i<(int)faces.size(); i++)	//faces
	{
		ImgFace face = faces[i];

		//fix face to have normal pointing upward
		if (CheckTriangle(face.v1, face.v2, face.v3))
		{
			int temp = face.v1;			face.v1 = face.v3;			face.v3 = temp;
			assert(!CheckTriangle(face.v1, face.v2, face.v3));
		}

		m_Faces.push_back(face);
		m_Vertices[face.v1].m_AdjacentFaces.push_back(m_Faces.size()-1);
		m_Vertices[face.v2].m_AdjacentFaces.push_back(m_Faces.size()-1);
		m_Vertices[face.v3].m_AdjacentFaces.push_back(m_Faces.size()-1);		
		if (face.v4!=-1)
		{
			m_Vertices[face.v4].m_AdjacentFaces.push_back(m_Faces.size()-1);
			m_Vertices[face.v4].AddAdjacentVertex(face.v1);	m_Vertices[face.v4].AddAdjacentVertex(face.v2);	m_Vertices[face.v4].AddAdjacentVertex(face.v3);
		}
		
		m_Vertices[face.v1].AddAdjacentVertex(face.v2);	m_Vertices[face.v1].AddAdjacentVertex(face.v3);	m_Vertices[face.v1].AddAdjacentVertex(face.v4);
		m_Vertices[face.v2].AddAdjacentVertex(face.v1);	m_Vertices[face.v2].AddAdjacentVertex(face.v3);	m_Vertices[face.v2].AddAdjacentVertex(face.v4);
		m_Vertices[face.v3].AddAdjacentVertex(face.v1);	m_Vertices[face.v3].AddAdjacentVertex(face.v2);	m_Vertices[face.v3].AddAdjacentVertex(face.v4);		
	}
}

ImageWarpLinearMesh::~ImageWarpLinearMesh()
{
}

void ImageWarpLinearMesh::GenerateHierarchicalParameterVectorChildImplement(SymVectorReal & params, SymVectorReal & trg)
{
	//Perform as-rigid-as-possible deformation to get final vertices. 
	//Params - same dimensions as trg, params - suggested vertex locations, 
	//trg - final vertex locations. Might need intermediate vector, where
	//vertices sorted.
	
	//NOTE: this function should be fast, since it is called very frequently
	GenHierarchicalParamsAsRigidAsPossible(params, trg);	
}

SymMatrix & ImageWarpLinearMesh::Deform(SymVectorReal & params, double * outPenalty)
{
	if (outPenalty!=NULL)
		*outPenalty = 0;
	
	//Deform mesh according to params
	for (int i=0; i<(int)m_Vertices.size(); i++)	//note not all params correspond to some vertex (due to hierarchy)
	{
		ImgVertex & v = m_Vertices[i];
		int paramID = v.m_iParamID;
		if (paramID < 0)
			continue;	//e.g. vertex is fixed
		
		if ((v.m_BoundaryVertexFlags & ImgVertex::BND_VERTEX_FIXED_X)==0 || !m_bFixBoundary)
			v.x = params(paramID*2+0);				
		else
			v.x = v.m_fBoundaryX;
		
		if ((v.m_BoundaryVertexFlags & ImgVertex::BND_VERTEX_FIXED_Y)==0 || !m_bFixBoundary)
			v.y = params(paramID*2+1);
		else
			v.y = v.m_fBoundaryY;
		
		if (outPenalty!=NULL && (v.m_BoundaryVertexFlags & ImgVertex::BND_VERTEX_FIXED_X)!=0
			&& (m_Penalization.m_iFields & ImagePenalization::STAY_ON_CANVAS)
			&& v.x > 0 && v.x < m_Source.width())
			*outPenalty = SYM_INFINITY;
		
		if (outPenalty!=NULL && (v.m_BoundaryVertexFlags & ImgVertex::BND_VERTEX_FIXED_Y)!=0
			&& (m_Penalization.m_iFields & ImagePenalization::STAY_ON_CANVAS)
			&& v.y > 0 && v.y < m_Source.height())
			*outPenalty = SYM_INFINITY;		
	}
	
	//Render textured mesh (gray-scale)
#ifdef USE_QT_FRAMEBUFFERS
	m_pFramebuffer->makeCurrent();
	glClear(GL_COLOR_BUFFER_BIT);
	glBindTexture(GL_TEXTURE_2D, m_TextureIds[1]);
	Draw(false);
	QImage deformedGray = m_pFramebuffer->toImage();
#else
	QImage deformedGray = DrawAndReadBuffer(false, true);
#endif
	
	for (int i=0; i<m_Source.width(); i++)
	{
		for (int j=0; j<m_Source.height(); j++)
		{
			if (m_bInvertColors)
				((SymMatrix2DReal&)*m_pDeformedMatrix)(i,j) = 1.-((double)qGray(deformedGray.pixel(i,j)))/255.;
			else
				((SymMatrix2DReal&)*m_pDeformedMatrix)(i,j) = ((double)qGray(deformedGray.pixel(i,j)))/255.;
		}
	}
	
	if (outPenalty!=0 && *outPenalty!=SYM_INFINITY)
		*outPenalty = CalculateDeformationPenalty();
	return *m_pDeformedMatrix;	
}

#ifndef USE_QT_FRAMEBUFFERS
void ImageWarpLinearMesh::LoadBuffer(bool highResBuffer, bool grayScale)
{
	BufferID & bufferID = highResBuffer ? m_iHighResBuffer : m_iFramebuffer;
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, bufferID.m_iFrameBufferObject);
    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight );
	glClear(GL_COLOR_BUFFER_BIT);

	if (grayScale)
	{
		assert(!highResBuffer);
		glBindTexture(GL_TEXTURE_2D, bufferID.m_iTextureBufferGrayscale);
	}
	else
		glBindTexture(GL_TEXTURE_2D, bufferID.m_iTextureBuffer);
}

QImage ImageWarpLinearMesh::ReadBuffer(bool highResBuffer, bool grayScale)
{
	BufferID & bufferID = highResBuffer ? m_iHighResBuffer : m_iFramebuffer;

	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
	GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(bufferID.m_iWidth >= viewport[2]);
    assert(bufferID.m_iHeight >= viewport[3]);
	
    glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight, 
                 GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
	
	//QImage deformedGray = m_pFramebuffer->toImage();
	return QImage(bufferID.m_aColorBuffer, 
				  bufferID.m_iWidth, bufferID.m_iHeight, 
				  QImage::Format_ARGB32 );

}


QImage ImageWarpLinearMesh::DrawAndReadBuffer(bool highResBuffer,
											  bool grayScale)
{
	LoadBuffer(highResBuffer, grayScale);
//	BufferID & bufferID = highResBuffer ? m_iHighResBuffer : m_iFramebuffer;
//	//	m_pFramebuffer->makeCurrent();
//	//	glClear(GL_COLOR_BUFFER_BIT);
//	//	glBindTexture(GL_TEXTURE_2D, m_TextureIds[1]);
//	//	Draw(false);
//	//	QImage deformedGray = m_pFramebuffer->toImage();
//
//	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, bufferID.m_iFrameBufferObject);
//    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight );
//	glClear(GL_COLOR_BUFFER_BIT);
//	if (grayScale)
//	{
//		assert(!highResBuffer);
//		glBindTexture(GL_TEXTURE_2D, bufferID.m_iTextureBufferGrayscale);
//	}
//	else
//		glBindTexture(GL_TEXTURE_2D, bufferID.m_iTextureBuffer);
	Draw(highResBuffer);
	
//	glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
//	GLint viewport[4];
//    glGetIntegerv(GL_VIEWPORT, viewport);
//    assert(bufferID.m_iWidth >= viewport[2]);
//    assert(bufferID.m_iHeight >= viewport[3]);
//	
//    glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight, 
//                 GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
//
//	//QImage deformedGray = m_pFramebuffer->toImage();
//	return QImage(bufferID.m_aColorBuffer, 
//				  bufferID.m_iWidth, bufferID.m_iHeight, 
//				  QImage::Format_ARGB32 );
	return ReadBuffer(highResBuffer, grayScale);
}
#endif

void ImageWarpLinearMesh::Draw(bool highResBuffer)
{
//	if (highResBuffer)
//		std::cout<<"Rendering high res="<<highResBuffer<<std::endl;
	bool renderingTriangles = m_Faces[0].v4==-1;
	if (renderingTriangles)
		glBegin(GL_TRIANGLES);
	else
		glBegin(GL_QUADS);
	
	for (int i=0; i<(int)m_Faces.size(); i++)
	{
		if (m_Faces[i].v4==-1 && !renderingTriangles)
		{
			glEnd();
			glBegin(GL_TRIANGLES);
		}
		else if (m_Faces[i].v4>=0 && renderingTriangles)
		{
			glEnd();
			glBegin(GL_QUADS);
		}
		ImgVertex & v1 = m_Vertices[m_Faces[i].v1];
		ImgVertex & v2 = m_Vertices[m_Faces[i].v2];
		ImgVertex & v3 = m_Vertices[m_Faces[i].v3];
		int v4 = m_Faces[i].v4;
		double scaleTextX = m_Source.width();
		double scaleTextY = m_Source.height();
		double scaleVertX = 1;
		double scaleVertY = 1;
		
		if (highResBuffer)
		{
#ifdef USE_QT_FRAMEBUFFERS
			assert(m_pHighResBuffer!=NULL);
			scaleVertX = (double)m_pFramebuffer->size().width() / (double)m_pHighResBuffer->size().width();
			scaleVertY = (double)m_pFramebuffer->size().height() / (double)m_pHighResBuffer->size().height();			
#else
			assert(m_iHighResBuffer.m_aColorBuffer !=NULL);
			scaleVertX = (double)m_iFramebuffer.m_iWidth / (double)m_iHighResBuffer.m_iWidth;
			scaleVertY = (double)m_iFramebuffer.m_iHeight / (double)m_iHighResBuffer.m_iHeight;
#endif
		}
		
		if (m_bInverseMapping)
		{
			glTexCoord2d(v1.x / scaleTextX, v1.y / scaleTextY);			glVertex2d(v1.tu / scaleVertX, v1.tv / scaleVertY);
			glTexCoord2d(v2.x / scaleTextX, v2.y / scaleTextY);			glVertex2d(v2.tu / scaleVertX, v2.tv / scaleVertY);
			glTexCoord2d(v3.x / scaleTextX, v3.y / scaleTextY);			glVertex2d(v3.tu / scaleVertX, v3.tv / scaleVertY);
			if (v4>=0)
			{
				glTexCoord2d(m_Vertices[v4].x / scaleTextX, m_Vertices[v4].y / scaleTextY);		
				glVertex2d(m_Vertices[v4].tu / scaleVertX, m_Vertices[v4].tv / scaleVertY);
			}
		}
		else
		{
			glTexCoord2d(v1.tu / scaleTextX, v1.tv / scaleTextY);		glVertex2d(v1.x / scaleVertX, v1.y / scaleVertY);
			glTexCoord2d(v2.tu / scaleTextX, v2.tv / scaleTextY);		glVertex2d(v2.x / scaleVertX, v2.y / scaleVertY);
			glTexCoord2d(v3.tu / scaleTextX, v3.tv / scaleTextY);		glVertex2d(v3.x / scaleVertX, v3.y / scaleVertY);
			if (v4>=0)
			{
				glTexCoord2d(m_Vertices[v4].tu / scaleTextX, m_Vertices[v4].tv / scaleTextY);		
				glVertex2d(m_Vertices[v4].x / scaleVertX, m_Vertices[v4].y / scaleVertY);
			}
		}
	}
	
	glEnd();
}

double ImageWarpLinearMesh::CalculateDeformationPenalty()
{
	double penalty = 0;
	if (m_Penalization.m_iFields & ImagePenalization::FOLDOVER_BINARY)
	{
		for (int i = 0; i<(int)m_Faces.size() && penalty < SYM_INFINITY; i++)
		{
			ImgFace face = m_Faces[i];
			if (CheckTriangle(face.v1, face.v2, face.v3))	//triangle fold-over
				penalty = SYM_INFINITY;
			else if (face.v4>=0 && CheckTriangle(face.v1, face.v3, face.v4))	//quad fold-over
				penalty = SYM_INFINITY;
		}
	}
	return penalty;
}

bool ImageWarpLinearMesh::CheckTriangle(int v1, int v2, int v3)
{
	R3Point p1, p2, p3;
	
	p1 = R3Point(m_Vertices[v1].x, m_Vertices[v1].y, 0);
	p2 = R3Point(m_Vertices[v2].x, m_Vertices[v2].y, 0);
	p3 = R3Point(m_Vertices[v3].x, m_Vertices[v3].y, 0);		
	
	R3Vector vec1 = p3 - p1;
	R3Vector vec2 = p2 - p3;
	
	vec1.Cross(vec2);
	return vec1.Z()>0;
}

int ImageWarpLinearMesh::GetNumOutputImageTypes()
{
	return 2;
}

void ImageWarpLinearMesh::GetCenterOfMass(double * cx, double * cy)
{
//	*cx=m_pFramebuffer->size().width()/2.;
//	*cy=m_pFramebuffer->size().height()/2.;	
	if (m_iHierarchyLevel==-1)
		SwitchToHierarchy(0);
	SymMatrix2DReal & deformedMatrix = (SymMatrix2DReal & )DeformAtCurrentHierarchyLevel(GetParameterVector());
	*cx = 0;
	*cy = 0;
	double totalNorm = 0;
	for (int i=0; i<deformedMatrix.m; i++)
	{
		for (int j=0; j<deformedMatrix.n; j++)
		{
			double value = deformedMatrix(i,j);
			*cx += value * i;
			*cy += value * j;			
			totalNorm += value;
		}
	}	
	*cx /= totalNorm;
	*cy /= totalNorm;
}

void ImageWarpLinearMesh::GetOutputDimensions(int * width, int * height)
{
#ifdef USE_QT_FRAMEBUFFERS
	*width = m_pFramebuffer->size().width();
	*height = m_pFramebuffer->size().height();	
#else
	*width = m_iFramebuffer.m_iWidth;
	*height = m_iFramebuffer.m_iHeight;
#endif
}

QImage ImageWarpLinearMesh::GetDeformedImage(int imageType)
{
	bool highResExists = false;
#ifdef USE_QT_FRAMEBUFFERS
	highResExists = m_pHighResBuffer!=NULL;
#else
	highResExists = m_iHighResBuffer.m_aColorBuffer!=NULL;
#endif
	m_bRetrievingImage = true;
	if (imageType==ImageWarpLinearMesh::IMAGE_TYPE_REGULAR
		|| imageType==DeformationModel::DEFORMATION_FINAL_RESULT)
	{
		if (m_bFixBoundary && imageType==ImageWarpLinearMesh::DEFORMATION_FINAL_RESULT)
		{
			m_bFixBoundary = false;
			DeformAtCurrentHierarchyLevel(GetParameterVector());
			m_bFixBoundary = true;
		}
		else
		{
			DeformAtCurrentHierarchyLevel(GetParameterVector());
		}
		
		DrawControlPoints();
		
		//Render textured mesh (color)
		if (imageType==DeformationModel::DEFORMATION_FINAL_RESULT && highResExists)
		{
#ifdef USE_QT_FRAMEBUFFERS
			m_pHighResBuffer->makeCurrent();
			glBindTexture(GL_TEXTURE_2D, m_HighResTextureIds[0]);						
#else
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_iHighResBuffer.m_iFrameBufferObject);
			glViewport( 0, 0, m_iHighResBuffer.m_iWidth, m_iHighResBuffer.m_iHeight );
			glBindTexture(GL_TEXTURE_2D, m_iHighResBuffer.m_iTextureBuffer);
#endif
		}
		else
		{
#ifdef USE_QT_FRAMEBUFFERS
			m_pFramebuffer->makeCurrent();
			glBindTexture(GL_TEXTURE_2D, m_TextureIds[0]);			
#else
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_iFramebuffer.m_iFrameBufferObject);
			glViewport( 0, 0, m_iFramebuffer.m_iWidth, m_iFramebuffer.m_iHeight );
			glBindTexture(GL_TEXTURE_2D, m_iFramebuffer.m_iTextureBuffer);
#endif
		}
		
		glClear(GL_COLOR_BUFFER_BIT);
		Draw(imageType==DeformationModel::DEFORMATION_FINAL_RESULT && highResExists);
		if (m_bShowWireframe && imageType!=ImageWarpLinearMesh::DEFORMATION_FINAL_RESULT)
		{
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			glDisable(GL_TEXTURE_2D);
			glColor3f(1, 0, 0);
			glLineWidth(1.);
			Draw(false);
			glEnable(GL_TEXTURE_2D);
			glColor3f(1., 1., 1.);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}
		
		//glReadPixels( 0, 0, m_Source.width(), m_Source.height(), GL_RGB, GL_UNSIGNED_BYTE, m_pColorBuffer);
		if (imageType==DeformationModel::DEFORMATION_FINAL_RESULT && highResExists)
		{
#ifdef USE_QT_FRAMEBUFFERS
			m_Deformed = m_pHighResBuffer->toImage();
#else
			m_Deformed = ReadBuffer(true, false);
#endif
		}
		else
		{
#ifdef USE_QT_FRAMEBUFFERS
			m_Deformed = m_pFramebuffer->toImage();
#else
			m_Deformed = ReadBuffer(true, false);
#endif
		}
		
		if (m_bShowWireframe && m_ControlPntsImg.width()!=0 
			&& imageType!=ImageWarpLinearMesh::DEFORMATION_FINAL_RESULT)
			DrawingRoutines::DrawMaskOver(m_Deformed, m_ControlPntsImg, qRgb(0,0,0));
			
		return m_Deformed;
	}
	else if (imageType==ImageWarpLinearMesh::IMAGE_TYPE_TEXTURE_SPACE)
	{
		bool tempFix = m_bFixBoundary;
		bool tempShowWire = m_bShowWireframe;
		m_bFixBoundary = false;
		m_bInverseMapping = !m_bInverseMapping;
		m_bShowWireframe = true;		
		DeformAtCurrentHierarchyLevel(GetParameterVector());
		DrawControlPoints();
		m_bFixBoundary = tempFix;
		m_bInverseMapping = !m_bInverseMapping;
		m_bShowWireframe = tempShowWire;
		
		QImage textureImg = m_Source;
		DrawingRoutines::DrawMaskOver(textureImg, m_ControlPntsImg, qRgb(0,0,0));
		return textureImg;
	}
	m_bRetrievingImage = false;	
	return QImage();
}

void ImageWarpLinearMesh::DrawControlPoints()
{
	if (m_bShowWireframe && m_bRetrievingImage && (m_HierarchyDescriptor==NULL || m_HierarchyDescriptor->m_HierarchyType!=AS_RIGID_AS_POSSIBLE))
	{
		if (m_ControlPntsImg.width()!=m_Deformed.width() || m_ControlPntsImg.height()!=m_Deformed.height())
			m_ControlPntsImg = QImage(m_Deformed.width(), m_Deformed.height(), QImage::Format_RGB32);
			DrawingRoutines::ClearImage(m_ControlPntsImg, qRgb(0, 0, 0));
		
		for (int i=0; i<(int)m_Vertices.size(); i++)
		{
			if (m_bInverseMapping)
				DrawingRoutines::DrawCircle(m_ControlPntsImg, m_Vertices[i].tu / m_ControlPntsImg.width(), 
											m_Vertices[i].tv / m_ControlPntsImg.height(), 
											0.01, 0, qRgb(0, 0, 255), qRgb(0,0,0));
			else
				DrawingRoutines::DrawCircle(m_ControlPntsImg, m_Vertices[i].x / m_ControlPntsImg.width(), 
											m_Vertices[i].y / m_ControlPntsImg.height(), 
											0.01, 0, qRgb(0, 0, 255), qRgb(0,0,0));
		}
	}
}

void ImageWarpLinearMesh::SetPostFilter(SymFilter * filter)
{
#ifdef USE_QT_FRAMEBUFFERS
	m_pFramebuffer->makeCurrent();
	glClear(GL_COLOR_BUFFER_BIT);
	glBindTexture(GL_TEXTURE_2D, m_TextureIds[1]);
#else
	LoadBuffer(false, true);
#endif
	
	Draw(false);
	filter->Process(m_pSourceMatrix);

	float * data = m_pSourceMatrix->GetRealDataArray();
#ifdef USE_QT_FRAMEBUFFERS
	glBindTexture(GL_TEXTURE_2D, m_TextureIds[1]);	// used to be in with the qt buffer?
#endif
	glTexImage2D(GL_TEXTURE_2D, 0, 1, m_pSourceMatrix->m, m_pSourceMatrix->n, 0, GL_LUMINANCE, GL_FLOAT, data);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	
	delete [] data;	
	
	//m_pSourceMatrix->ToImage().save("AfterFlt.png");	
#ifdef USE_QT_FRAMEBUFFERS
	glBindTexture(GL_TEXTURE_2D, m_TextureIds[1]);	// used to be in with the qt buffer?
#endif
	Draw(false);
}
