#include "DeformTextureSynth.h"
#include "DrawingRoutines.h"
#include "CVLib.h"
#include <iostream>
#include <time.h>

DeformTextureSynth::DeformTextureSynth(QImage & src, int resWidth, int resHeight, 
									   int patchWidth, int patchHeight, 
									   int overlapWidth, int overlapHeight, 
									   int optimNumPatches, int optimNumBestPatches,
									   bool minCutCarving)
{
	m_bMinCutCarving = minCutCarving;
	assert(src.width()>patchHeight && src.height()>patchWidth);
	m_iCurrentPatch = -1;
	
	m_iOptimNumPatches = optimNumPatches;
	m_iOptimNumBestPatches = optimNumBestPatches;

	m_Source = src;
	m_pDeformedMatrix = new SymMatrix2DReal(resWidth, resHeight);
	m_pSourceMatrix = new SymMatrix2DReal(m_Source);
	m_Deformed = QImage(resWidth, resHeight, QImage::Format_RGB32);	
	m_iPatchWidth = patchWidth;
	m_iPatchHeight = patchHeight;
	m_iOverlapWidth = overlapWidth;
	m_iOverlapHeight = overlapHeight;	

	m_iPatchNumX = ceil((double)(resWidth-m_iOverlapWidth) / (double)(m_iPatchWidth-m_iOverlapWidth));
	m_iPatchNumY = ceil((double)(resHeight-m_iOverlapHeight) / (double)(m_iPatchHeight-m_iOverlapHeight));
	
	std::cout<<"Quilting Parameters:"<<std::endl;
	std::cout<<"\tNum patches="<<m_iPatchNumX<<" x "<<m_iPatchNumY<<" = "<<m_iPatchNumX*m_iPatchNumY<<std::endl;
	std::cout<<"\tNum BestPatches = "<<m_iOptimNumBestPatches<<std::endl;
	std::cout<<"\tUpdate every "<<m_iPatchNumX*m_iPatchNumY*m_iOptimNumBestPatches<<std::endl;
	std::cout<<"\tMinCut="<<(m_bMinCutCarving?("true"):("false"))<<std::endl;	
	m_pAllLastParams = new SymVectorReal(2 * m_iPatchNumX * m_iPatchNumY);
	m_pAllParamStep = new SymVectorReal(2 * m_iPatchNumX * m_iPatchNumY, 1);
	m_pAllMinParamStep = new SymVectorReal(2 * m_iPatchNumX * m_iPatchNumY, 1);
	
	m_TempTesterCheckIfAllPatches = QImage(m_iPatchNumX, m_iPatchNumY, QImage::Format_RGB32);
	DrawingRoutines::ClearImage(m_TempTesterCheckIfAllPatches, qRgb(0, 0, 0));

	for (int i=0; i<m_iPatchNumX * m_iPatchNumY; i++)
		m_iPatchPermutations.push_back(i);
	for (int i=0; i<m_iPatchNumX * m_iPatchNumY; i++)
	{
		int switchTo = rand()%m_iPatchPermutations.size();
		int curr=m_iPatchPermutations[i];
		int other=m_iPatchPermutations[switchTo];
		m_iPatchPermutations[i] = other;
		m_iPatchPermutations[switchTo]=curr;
	}
		
	m_ParamPrintNodes.clear();
	std::vector<int> nodes;		nodes.push_back(0);		nodes.push_back(1);
	m_ParamPrintNodes.push_back(HierarchyPrintNode("", nodes, 0));
	srand(time(NULL));
//	srand(57201);
	for (int i=0; i<m_iPatchNumX; i++)
	{
		for (int j=0; j<m_iPatchNumY; j++)
		{
			int id= 2*(j*m_iPatchNumY+i);
			(*m_pAllLastParams)(id+0) = rand() % (m_Source.width()-m_iPatchWidth-1);
			(*m_pAllLastParams)(id+1) = rand() % (m_Source.height()-m_iPatchHeight-1);
		}
	}
	
	m_pInitParams = m_pAllLastParams;
	m_pParameterSteps = m_pAllParamStep;
	m_pMinimalParamSteps = m_pAllMinParamStep;			
	
	for (int i=0; i<m_iOverlapHeight * m_iPatchWidth; i++)
		m_OverlapFieldHorizontal.push_back(DynProgCell());
	for (int i=0; i<m_iOverlapWidth * m_iPatchHeight; i++)
		m_OverlapFieldVertical.push_back(DynProgCell());
	
	for (int i=0; i<m_iPatchNumX; i++)
	{
		for (int j=0; j<m_iPatchNumY; j++)
		{
			m_HorizontalSeams.push_back(SymSeam());
			m_HorizontalSeams[m_HorizontalSeams.size()-1].m_iWidth = m_iOverlapHeight;
			m_HorizontalSeams[m_HorizontalSeams.size()-1].m_iHeight = m_iPatchWidth;			
			m_VerticalSeams.push_back(SymSeam());
			m_VerticalSeams[m_VerticalSeams.size()-1].m_iWidth = m_iOverlapWidth;
			m_VerticalSeams[m_VerticalSeams.size()-1].m_iHeight = m_iPatchHeight;
		}
	}
	
	Deform(GetParameterVector());
}

bool DeformTextureSynth::HasOutput(OutputType type)
{
	return type==DEFORM_OUT_IMG2D || type==DEFORM_OUT_SYMREAL2D;
}

void DeformTextureSynth::GetOutputDimensions(int * inWidth, int * inHeight)
{
	*inWidth = m_Deformed.width();
	*inHeight = m_Deformed.height();	
}

void DeformTextureSynth::GetMinCutSeam(std::vector<DynProgCell> & overlap, SymSeam & trgSeam)
{
	assert((int)overlap.size() == trgSeam.m_iWidth * trgSeam.m_iHeight);
	trgSeam.m_SeamBoundary.clear();

//	//TODO: remove me
//	SymMatrix2DReal cutSeamThrough(trgSeam.m_iWidth, trgSeam.m_iHeight);
	
		//fill dynaic programming table
	for (int j=0; j<trgSeam.m_iHeight; j++)
	{
		for (int i=0; i<trgSeam.m_iWidth ; i++)
		{
//			cutSeamThrough(i,j) = overlap[j * trgSeam.m_iWidth + i].m_fCost;
			if (j==0)
			{
				overlap[j * trgSeam.m_iWidth + i].m_iPrevCell = -1;
			}
			else
			{
				double bestPrevCost = SYM_INFINITY;
				for (int locI = qMax(i-1, 0); locI <= qMin(trgSeam.m_iWidth-1, i+1); locI++)
				{
					double prevCost = overlap[(j-1) * trgSeam.m_iWidth + locI].m_fCost;
					if (prevCost < bestPrevCost)
					{
						overlap[j * trgSeam.m_iWidth + i].m_iPrevCell = locI;
						bestPrevCost = prevCost;
					}
				}
				overlap[j * trgSeam.m_iWidth + i].m_fCost += bestPrevCost;
			}
		}
	}
	
	// find best final location
	double bestCost = SYM_INFINITY;
	int bestCostLoc = -1;
	for (int i=0; i<trgSeam.m_iWidth; i++)
	{
		if (overlap[(trgSeam.m_iHeight-1)* trgSeam.m_iWidth + i].m_fCost < bestCost)
		{
			bestCost = overlap[(trgSeam.m_iHeight-1)* trgSeam.m_iWidth + i].m_fCost;
			bestCostLoc = i;
		}
	}
	
//	static int temp = 0;
//	if (temp==0)
//		cutSeamThrough.ToImage().save("SeamInit.png");	
	// retrieve the min cut seam
	for (int j=trgSeam.m_iHeight-1; j>=0; j--)
	{
//		cutSeamThrough(bestCostLoc, j) = 1;
		trgSeam.m_SeamBoundary.insert(trgSeam.m_SeamBoundary.begin(), bestCostLoc);
//		trgSeam.m_SeamCost.insert(trgSeam.m_SeamCost.begin(), overlap[j*trgSeam.m_iWidth + bestCostLoc]);
		bestCostLoc = overlap[j*trgSeam.m_iWidth + bestCostLoc].m_iPrevCell;
	}
//	if (temp++==0)	
//		cutSeamThrough.ToImage().save("SeamFinal.png");
}

double DeformTextureSynth::GetColorDifference(QImage & img1, int i1x, int i1y, QImage & img2, int i2x, int i2y)
{
	if (i1x<0 || i1x>=img1.width() || i1y<0 || i1y>=img1.height() 
		|| i2x<0 || i2x>=img2.width() || i2y<0 || i2y>=img2.height())
		return 10;
	QRgb c1 = img1.pixel(i1x, i1y);
	QRgb c2 = img2.pixel(i2x, i2y);	
	double dr = ((double)(qRed(c1) - qRed(c2)))/255.;
	double dg = ((double)(qGreen(c1) - qGreen(c2)))/255.;
	double db = ((double)(qBlue(c1) - qBlue(c2)))/255.;
	return dr*dr + dg*dg + db*db;
}

bool DeformTextureSynth::ValidPixel(SymSeam * topVerticalSeam, SymSeam * leftHorizontalSeam, 
									int leftBndX, int topBndY, int imgWidth, int imgHeight, 
									int x, int y, int overlapWidth, int overlapHeight)
{
	bool satisfiesVertSeam;
	bool satisfiesHorSeam;
	bool satisfiesDims;
	satisfiesDims = x>=leftBndX && y>=topBndY && x<imgWidth && y<imgHeight;
	
	int locX = x - leftBndX;
	int locY = y - topBndY;
	if (locX > overlapWidth && locY >overlapHeight)
		return satisfiesDims;
	
	satisfiesVertSeam = topVerticalSeam==NULL || (locX > topVerticalSeam->m_SeamBoundary[locY]);
	//int maxSeam = leftHorizontalSeam==NULL ? 0 : leftHorizontalSeam->m_SeamBoundary.size()-1;
	//satisfiesHorSeam = leftHorizontalSeam==NULL || (locY > leftHorizontalSeam->m_SeamBoundary[maxSeam-locX]);	
	satisfiesHorSeam = leftHorizontalSeam==NULL || (locY > leftHorizontalSeam->m_SeamBoundary[locX]);	
	
	return satisfiesDims && satisfiesHorSeam && satisfiesVertSeam;
}

SymMatrix & DeformTextureSynth::Deform(SymVectorReal & params, double * outPenalty)
{
	ComputeAllSeams();
	if (outPenalty!=NULL)
		*outPenalty = 0;
	
	assert(m_pAllLastParams->m==params.m);
	if (m_pAllLastParams!=&params)
		*m_pAllLastParams = params;

	for (int i=0; i<m_iPatchNumX*m_iPatchNumY; i++)
	{	
		FillVerticalScanlines(i);		// vertical scanlines (use horizontal seams)
		FillHorizontalScanlines(i);		// horizontal scanlines (use vertical seams)
		FillBottomRightCorner(i);		// squares (need all seems)
	}

	SymMatrix2DReal & trgMatrix = (SymMatrix2DReal&)*m_pDeformedMatrix;		
	trgMatrix.FromImage(m_Deformed);
	return trgMatrix;
}

void DeformTextureSynth::FillVerticalScanlines(int patchID)
{	
	int patchI = patchID % m_iPatchNumX;
	int patchJ = patchID / m_iPatchNumX;	
	int x0 = patchI * (m_iPatchWidth-m_iOverlapWidth);
	int y0 = patchJ * (m_iPatchHeight-m_iOverlapHeight);
	int y1 = y0 + m_iPatchHeight;
	int srcX = GetParameterVector()(2*patchID+0);
	int srcY = GetParameterVector()(2*patchID+1);	
		
	SymSeam * upperSeam = NULL;
	SymSeam * bottomSeam = NULL;	
	if (patchJ>0 && m_iOverlapHeight!=0 && m_bMinCutCarving)
		upperSeam = &(m_HorizontalSeams[(patchJ-1)*m_iPatchNumX + patchI]);
	if (patchJ<m_iPatchNumY-1 && m_iOverlapHeight!=0 && m_bMinCutCarving)
		bottomSeam = &(m_HorizontalSeams[(patchJ)*m_iPatchNumX + patchI]);
	
	int minX = m_iOverlapWidth;
	int maxX = m_iPatchWidth-m_iOverlapWidth;

	if (!m_bMinCutCarving)
	{
		minX = 0;
		maxX = m_iPatchWidth;
	}	
	
	if (patchI==0)
		minX = 0;
	if (patchI==m_iPatchNumX-1)
		maxX = m_Deformed.width()-x0;
	
	for (int x=minX; x<maxX; x++)
	{
		int minY = upperSeam==NULL ? y0 : y0 + upperSeam->m_SeamBoundary[x];
		int maxY = bottomSeam==NULL ? y1 : y1 - m_iOverlapWidth + bottomSeam->m_SeamBoundary[x];
		for (int y=minY; y<qMin(maxY, m_Deformed.height()); y++)
			m_Deformed.setPixel(x0+x, y, m_Source.pixel(srcX+x, srcY+y-y0));
	}
}

void DeformTextureSynth::FillHorizontalScanlines(int patchID)
{
	if (m_iOverlapWidth==0 || !m_bMinCutCarving)
		return;
	
	int patchI = patchID % m_iPatchNumX;
	int patchJ = patchID / m_iPatchNumX;	
	int x0 = patchI * (m_iPatchWidth-m_iOverlapWidth);
	int y0 = patchJ * (m_iPatchHeight-m_iOverlapHeight);
	int x1 = x0 + m_iPatchWidth;	
	int srcX = GetParameterVector()(2*patchID+0);
	int srcY = GetParameterVector()(2*patchID+1);	
	
	SymSeam * leftSeam = NULL;
	SymSeam * rightSeam = NULL;	
	if (patchI>0)
		leftSeam = &(m_VerticalSeams[patchJ*m_iPatchNumX + patchI-1]);
	if (patchI<m_iPatchNumX-1)
		rightSeam = &(m_VerticalSeams[patchJ*m_iPatchNumX + patchI]);
	
	int minY = m_iOverlapHeight;
	int maxY = m_iPatchHeight-m_iOverlapHeight;
	
	if (patchJ==0)
		minY = 0;
	if (patchJ==m_iPatchNumY-1)
		maxY = m_Deformed.height()-y0;
	
		
	for (int y=minY; y<maxY; y++)
	{
		int minX, maxX;
		if (leftSeam!=NULL)
		{
			minX = x0 + leftSeam->m_SeamBoundary[y];
			maxX = x0 + m_iOverlapWidth;
			for (int x=minX; x<qMin(maxX, m_Deformed.width()); x++)
				m_Deformed.setPixel(x, y0+y, m_Source.pixel(srcX+x-x0, srcY+y));
		}
		if (rightSeam!=NULL)
		{
			minX = x1 - m_iOverlapWidth;
			maxX = x1 - m_iOverlapWidth + rightSeam->m_SeamBoundary[y];
			for (int x=minX; x<qMin(maxX, m_Deformed.width()); x++)
				m_Deformed.setPixel(x, y0+y, m_Source.pixel(srcX+x-x0, srcY+y));
		}
	}	
}

void DeformTextureSynth::FillBottomRightCorner(int patchID)
{
	if (m_iOverlapWidth==0 || m_iOverlapHeight==0 || !m_bMinCutCarving)
		return;

	int patchI = patchID % m_iPatchNumX;
	int patchJ = patchID / m_iPatchNumY;
	if (patchI==m_iPatchNumX-1 || patchJ==m_iPatchNumY-1)
		return;
		
	int x0 = patchI * (m_iPatchWidth-m_iOverlapWidth);
	int y0 = patchJ * (m_iPatchHeight-m_iOverlapHeight);
	int x1 = (patchI+1) * (m_iPatchWidth-m_iOverlapWidth); 
	int y1 = (patchJ+1) * (m_iPatchHeight-m_iOverlapHeight);
		
	// R1:11	| |	  R2:10
	// ---------------------
	// ---------------------
	// R3:01	| |   R4:00
	int src11X = GetParameterVector()(2 * (patchJ	  * m_iPatchNumX + patchI)		+ 0) - x0;
	int src11Y = GetParameterVector()(2 * (patchJ	  * m_iPatchNumX + patchI)		+ 1) - y0;
	int src10X = GetParameterVector()(2 * (patchJ  	  * m_iPatchNumX + (patchI+1))	+ 0) - x0-m_iPatchWidth+m_iOverlapWidth;
	int src10Y = GetParameterVector()(2 * (patchJ	  * m_iPatchNumX + (patchI+1))	+ 1) - y0;
	int src01X = GetParameterVector()(2 * ((patchJ+1) * m_iPatchNumX + patchI)		+ 0) - x0;
	int src01Y = GetParameterVector()(2 * ((patchJ+1) * m_iPatchNumX + patchI)		+ 1) - y0-m_iPatchHeight+m_iOverlapHeight;
	int src00X = GetParameterVector()(2 * ((patchJ+1) * m_iPatchNumX + (patchI+1))	+ 0) - x0-m_iPatchWidth+m_iOverlapWidth;
	int src00Y = GetParameterVector()(2 * ((patchJ+1) * m_iPatchNumX + (patchI+1))	+ 1) - y0-m_iPatchHeight+m_iOverlapHeight;
	
	SymSeam * leftSeam = &m_HorizontalSeams[patchJ * m_iPatchNumX + patchI];
	SymSeam * topSeam = &m_VerticalSeams[patchJ * m_iPatchNumX + patchI];	
	SymSeam * rightSeam = &m_HorizontalSeams[patchJ * m_iPatchNumX + patchI+1];
	SymSeam * bottomSeam = &m_VerticalSeams[(patchJ+1) * m_iPatchNumX + patchI];
//#define RENDER_SEAMS	
	for (int x=x0+m_iPatchWidth-m_iOverlapWidth; x<qMin(x0+m_iPatchWidth, m_Deformed.width()); x++)
	{
		for (int y=y0+m_iPatchHeight-m_iOverlapHeight; y<qMin(y0+m_iPatchHeight, m_Deformed.height()); y++)
		{
			int regionID = GetRegionID(x-x0, y-y0, leftSeam, rightSeam, topSeam, bottomSeam);
			if (regionID==1)
				m_Deformed.setPixel(x, y, m_Source.pixel(src11X+x, src11Y+y));
			else if (regionID==2)
				m_Deformed.setPixel(x, y, m_Source.pixel(src10X+x, src10Y+y));
			else if (regionID==3)
				m_Deformed.setPixel(x, y, m_Source.pixel(src01X+x, src01Y+y));
			else if (regionID==4)
				m_Deformed.setPixel(x, y, m_Source.pixel(src00X+x, src00Y+y));
			else
				m_Deformed.setPixel(x, y, qRgb(0,0,0));
#ifdef RENDER_SEAMS
			if (x==x0+m_iPatchWidth-m_iOverlapWidth || x==qMin(x0+m_iPatchWidth, m_Deformed.width())-1
				|| y==y0+m_iPatchHeight-m_iOverlapHeight || y==qMin(y0+m_iPatchHeight, m_Deformed.height())-1)
				m_Deformed.setPixel(x, y, qRgb(0, 0, 0));				
#endif
		}		
	}
	
#ifdef RENDER_SEAMS
		for (int x=m_iPatchWidth-m_iOverlapWidth; x<m_iPatchWidth; x++)
		{
			m_Deformed.setPixel(x0+x, y0+m_iPatchHeight-m_iOverlapHeight+leftSeam->m_SeamBoundary[x], qRgb(0, 128, 0));
			m_Deformed.setPixel(x0+x, y0+m_iPatchHeight-m_iOverlapHeight+rightSeam->m_SeamBoundary[x-m_iPatchWidth+m_iOverlapWidth], qRgb(0, 255, 0));
		}
		
		for (int y=m_iPatchHeight-m_iOverlapHeight; y<m_iPatchHeight; y++)
		{
			m_Deformed.setPixel(x0+m_iPatchWidth-m_iOverlapWidth+topSeam->m_SeamBoundary[y], y0+y, qRgb(0, 128, 128));
			m_Deformed.setPixel(x0+m_iPatchWidth-m_iOverlapWidth+bottomSeam->m_SeamBoundary[y-m_iPatchHeight+m_iOverlapHeight], y0+y, qRgb(0, 255, 255));
		}
#endif		
}

int DeformTextureSynth::GetRegionID(int x, int y, SymSeam * leftSeam, SymSeam * rightSeam, SymSeam * topSeam, SymSeam * bottomSeam)
{
	int aboveLeft = y < m_iPatchHeight - m_iOverlapHeight + leftSeam->m_SeamBoundary[x];
	int aboveRight = y < m_iPatchHeight - m_iOverlapHeight + rightSeam->m_SeamBoundary[x-m_iPatchWidth+m_iOverlapWidth];
	int leftTop = x < m_iPatchWidth - m_iOverlapWidth + topSeam->m_SeamBoundary[y];
	int leftBottom = x < m_iPatchWidth - m_iOverlapWidth + bottomSeam->m_SeamBoundary[y-m_iPatchHeight+m_iOverlapHeight];
	
	int val = (aboveLeft<<12) | (aboveRight<<8) | (leftTop<<4) | (leftBottom);
	switch(val)
	{
		case 0x1111:
		case 0x1110:
		case 0x1011:
		case 0x1010:	//CENTER: 1 or 4 (both seams are not valid)
		case 0x0110:	//CENTER: any
		case 0x1001:	//CENTER: any
			return 1;
		case 0x1100:
		case 0x1101:
		case 0x0100:
		case 0x0101:	//CENTER: 2 or 3 (both seams are bad)
			return 2;
		case 0x0011:
		case 0x0001:
		case 0x0111:
			return 3;
		case 0x0000:
		case 0x0010:
		case 0x1000:
			return 4;
			
	}
	return 0;
} 
		 
QImage DeformTextureSynth::GetDeformedImage(int imageType)
{
//	assert(imageType==0);
	return m_Deformed;
}


SymMatrix & DeformTextureSynth::TakeRandomStep(bool & enforceStep)
{
	enforceStep = false;
	if (m_PickedCandidatesPatches.size()==0)	//find candidates
	{
		if (m_iCurrentPatch>=0)
			UpdateOnlyOnePatch(m_iPatchPermutations[m_iCurrentPatch]);
		m_iCurrentPatch = (m_iCurrentPatch+1) % m_iPatchPermutations.size();
		int patchI = m_iPatchPermutations[m_iCurrentPatch] % m_iPatchNumX;
		int patchJ = m_iPatchPermutations[m_iCurrentPatch] / m_iPatchNumX;
		//enforceStep = true;
		if (qGray(m_TempTesterCheckIfAllPatches.pixel(patchI, patchJ))==0)
		{
			enforceStep = true;
			m_TempTesterCheckIfAllPatches.setPixel(patchI, patchJ, qRgb(255, 255, 255));
//			m_TempTesterCheckIfAllPatches.save("Visited.png");
		}
		if (m_iCurrentPatch==0)
			std::cout<<"Gone through Full Cycle!"<<std::endl;
		int leftX, leftY, rightX, rightY, topX, topY, bottomX, bottomY;
		if (patchI==0)
		{
			leftX = -1;
		}
		else
		{
			int id= 2*(patchJ*m_iPatchNumX+(patchI-1));
			leftX = GetParameterVector()(id+0);
			leftY = GetParameterVector()(id+1);			
		}
		
		if (patchI==m_iPatchNumX-1)
		{
			rightX = -1;
		}
		else
		{
			int id= 2*(patchJ*m_iPatchNumX+(patchI+1));
			rightX = GetParameterVector()(id+0);
			rightY = GetParameterVector()(id+1);			
		}
				
		if (patchJ==0)
		{
			topX = -1;
		}
		else
		{
			int id= 2*((patchJ-1)*m_iPatchNumX+patchI);
			topX = GetParameterVector()(id+0);
			topY = GetParameterVector()(id+1);			
		}
		
		if (patchJ==m_iPatchNumY-1)
		{
			bottomX = -1;
		}
		else
		{
			int id= 2*((patchJ+1)*m_iPatchNumX+patchI);
			bottomX = GetParameterVector()(id+0);
			bottomY = GetParameterVector()(id+1);			
		}
		
		FillMinOverlapCandidates(m_Source, m_iPatchWidth, m_iPatchHeight, m_iOverlapWidth, m_iOverlapHeight, 
								 leftX, leftY, topX, topY, rightX, rightY, bottomX, bottomY,
								 m_PickedCandidatesPatches, m_iOptimNumPatches, m_iOptimNumBestPatches); 
	}
	
	m_iOldPatchX = GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+0);
	m_iOldPatchY = GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+1);
		//pick next candidate
	GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+0) = m_PickedCandidatesPatches[0];
	GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+1) = m_PickedCandidatesPatches[1];
	m_PickedCandidatesPatches.erase(m_PickedCandidatesPatches.begin(), m_PickedCandidatesPatches.begin()+2);
	
	UpdateOnlyOnePatch(m_iPatchPermutations[m_iCurrentPatch]);
	
	SymMatrix2DReal & trgMatrix = (SymMatrix2DReal&)*m_pDeformedMatrix;		
	trgMatrix.FromImage(m_Deformed);
	return trgMatrix;	
}

void DeformTextureSynth::RevertLastRandomStep()
{
	GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+0) = m_iOldPatchX;
	GetParameterVector()(2*m_iPatchPermutations[m_iCurrentPatch]+1) = m_iOldPatchY;	
}

bool DeformTextureSynth::DistortsInput()
{
	return true;
}

void DeformTextureSynth::UpdateOnlyOnePatch(int patchID)
{
	RecomputeSeamsForPatch(patchID);
	
	FillVerticalScanlines(patchID);
	FillHorizontalScanlines(patchID);
	
	int patchI = patchID % m_iPatchNumX;
	int patchJ = patchID / m_iPatchNumX;
	FillBottomRightCorner(patchID);
	if (patchI>0)
	{
		FillBottomRightCorner(patchJ * m_iPatchNumX + patchI-1);
		FillHorizontalScanlines(patchJ * m_iPatchNumX + patchI-1);
	}
	if (patchI < m_iPatchNumX-1)
	{
		FillHorizontalScanlines(patchJ * m_iPatchNumX + patchI+1);
	}
	
	if (patchJ>0)
	{
		FillBottomRightCorner((patchJ-1) * m_iPatchNumX + patchI);
		FillVerticalScanlines((patchJ-1) * m_iPatchNumX + patchI);
	}
	
	if (patchJ<m_iPatchNumY-1)
	{
		FillVerticalScanlines((patchJ+1) * m_iPatchNumX + patchI);
	}
	
	if (patchJ>0 && patchI>0)
		FillBottomRightCorner((patchJ-1) * m_iPatchNumX + patchI-1);	
}


void DeformTextureSynth::FillMinOverlapCandidates(QImage & srcImg, int patchWidth, int patchHeight,
												  int overlapWidth, int overlapHeight,	
												  int leftX, int leftY,
												  int topX, int topY, int rightX, int rightY,
												  int bottomX, int bottomY, 
												  std::vector<int> & candidates, 
												  int numPatches, int numBestPatches)
{
	std::vector<SortableEntry> errors;
	std::vector<int> patchCoords;
	for (int i=0; i<numPatches; i++)
	{
		int randomPatchX = rand() % (srcImg.width() - patchWidth-1);
		int randomPatchY = rand() % (srcImg.height() - patchHeight-1);
		patchCoords.push_back(randomPatchX);
		patchCoords.push_back(randomPatchY);
		
		double err = 0;
		if (leftX>0)	// calculate error with left patch
			err+=CompareRegions(srcImg, overlapWidth, patchHeight, randomPatchX, randomPatchY, 
								leftX+patchWidth-overlapWidth, leftY);		
		
		if (topX>0)	// calculate error with top patch
			err+=CompareRegions(srcImg, patchWidth, overlapHeight, randomPatchX, randomPatchY, 
								topX, topY+patchHeight-overlapHeight);

		if (rightX>0)	// calculate error with right patch
		{
			err+=CompareRegions(srcImg, overlapWidth, patchHeight, 
								randomPatchX+patchWidth-overlapWidth, randomPatchY, rightX, rightY);
		}

		if (bottomX>0)	// calculate error with bottom patch
		{
			err+=CompareRegions(srcImg, patchWidth, overlapHeight, 
								randomPatchX, randomPatchY+patchHeight-overlapHeight, bottomX, bottomY);
		}
		errors.push_back(SortableEntry(err, i));
	}
	sort(errors.begin(), errors.end(), &SortingFunction);
	
	for (int i=0; i<numBestPatches; i++)
	{
		candidates.push_back(patchCoords[2*errors[i].m_iID+0]);
		candidates.push_back(patchCoords[2*errors[i].m_iID+1]);
	}
}

double DeformTextureSynth::CompareRegions(QImage & srcImg, int regionWidth, int regionHeight, 
										  int x1, int y1, int x2, int y2)
{
	double err = 0;
	for (int i=0; i<regionWidth; i++)
		for (int j=0; j<regionHeight; j++)
			err += GetColorDifference(srcImg, x1+i, y1+j, srcImg, x2+i, y2+j);
	return err;
}

void DeformTextureSynth::ComputeAllSeams()
{	
	for (int i=0; i<m_iPatchNumX; i++)
	{
		for (int j=0; j<m_iPatchNumY; j++)
		{
			int idCntr = j * m_iPatchNumX + i;
			if (i>0)
			{
				int idLeft = j * m_iPatchNumX + (i-1);
				ComputeSeamBetweenPatches(idCntr, idLeft);
			}
			if (j>0)
			{
				int idTop = (j-1) * m_iPatchNumX + i;
				ComputeSeamBetweenPatches(idCntr, idTop);
			}			
		}
	}
}

void DeformTextureSynth::RecomputeSeamsForPatch(int patchID)
{
	int patchI = patchID % m_iPatchNumX;
	int patchJ = patchID / m_iPatchNumX;
	if (patchI>0)
		ComputeSeamBetweenPatches(patchJ*m_iPatchNumX + (patchI-1), patchID);		// left
	if (patchJ>0)
		ComputeSeamBetweenPatches((patchJ-1)*m_iPatchNumX + patchI, patchID);		// top
	if (patchI<m_iPatchNumX-1)
		ComputeSeamBetweenPatches(patchID, patchJ*m_iPatchNumX + (patchI+1));	// right
	if (patchJ<m_iPatchNumY-1)
		ComputeSeamBetweenPatches(patchID, (patchJ+1)*m_iPatchNumX + patchI);	// bottom
}

void DeformTextureSynth::ComputeSeamBetweenPatches(int patchID1, int patchID2)
{
	int patch1I = patchID1 % m_iPatchNumX;
	int patch1J = patchID1 / m_iPatchNumX;
	int patch2I = patchID2 % m_iPatchNumX;
	int patch2J = patchID2 / m_iPatchNumX;
	
	// find seams
	if (patch1J==patch2J && qAbs(patch1I-patch2I)==1)	// vertical block
	{
		if (m_iOverlapWidth==0 || !m_bMinCutCarving)		
			return;
		
		if (patch1I > patch2I)
		{
			int temp;
			temp = patch1I;		patch1I = patch2I;		patch2I = temp;
			temp = patch1J;		patch1J = patch2J;		patch2J = temp;			
			temp = patchID1;	patchID1 = patchID2;	patchID2 = temp;
		}
		int x1 = GetParameterVector()(2*patchID1+0) + m_iPatchWidth - m_iOverlapWidth;
		int y1 = GetParameterVector()(2*patchID1+1);
		int x2 = GetParameterVector()(2*patchID2+0);
		int y2 = GetParameterVector()(2*patchID2+1);
		for (int overlapX = 0; overlapX < m_iOverlapWidth; overlapX++)
		{
			for (int overlapY = 0; overlapY < m_iPatchHeight; overlapY++)
			{
				int locX1 = x1 + overlapX;
				int locY1 = y1 + overlapY;
				int locX2 = x2 + overlapX;
				int locY2 = y2 + overlapY;	
				
				m_OverlapFieldVertical[overlapY * m_iOverlapWidth + overlapX].m_fCost = 
							GetColorDifference(m_Source, locX1, locY1, m_Source, locX2, locY2);
			}
		}
		GetMinCutSeam(m_OverlapFieldVertical, m_VerticalSeams[patchID1]);
	}
	else if (patch1I==patch2I && qAbs(patch1J-patch2J)==1)	// horizontal block
	{
		if (m_iOverlapHeight==0 || !m_bMinCutCarving)
			return;
		
		if (patch1J > patch2J)
		{
			int temp;
			temp = patch1I;		patch1I = patch2I;		patch2I = temp;
			temp = patch1J;		patch1J = patch2J;		patch2J = temp;			
			temp = patchID1;	patchID1 = patchID2;	patchID2 = temp;
		}
		
		int x1 = GetParameterVector()(2*patchID1+0);
		int y1 = GetParameterVector()(2*patchID1+1) + m_iPatchHeight - m_iOverlapHeight;
		int x2 = GetParameterVector()(2*patchID2+0);
		int y2 = GetParameterVector()(2*patchID2+1);
		
		for (int overlapX = 0; overlapX < m_iOverlapHeight; overlapX++)
		{
			for (int overlapY = 0; overlapY < m_iPatchWidth; overlapY++)
			{
				int locX1 = x1 + overlapY;
				int locY1 = y1 + overlapX;
				int locX2 = x2 + overlapY;
				int locY2 = y2 + overlapX;	
								
				m_OverlapFieldHorizontal[overlapY * m_iOverlapHeight + overlapX].m_fCost = 
						GetColorDifference(m_Source, locX1, locY1, m_Source, locX2, locY2);
			}
		}
		GetMinCutSeam(m_OverlapFieldHorizontal, m_HorizontalSeams[patchID1]);				
	}
	else
	{
		std::cout<<"PatchIDs = "<<patchID1<<" and "<<patchID2<<std::endl;
		assert(false);
	}
}


