#include "SymOptimizer.h"
#include "SymPipeline.h"

double optimF(const gsl_vector * x, void * params)
{
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("OptimizationFullIteration");
#endif
	
	gsl_vector * nonConstVec = (gsl_vector * )x;	//hack to use my SymVectorReal
	SymVectorReal theta(nonConstVec);
	DeformationModel * deform = ((SymOptimizer*)params)->m_pDeformationModel;
	
	SymEnergy * energy = ((SymOptimizer*)params)->m_pEnergyFn;
	double penalty = 0;
	
#ifdef CALCULATE_TIMING	
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("OptimizationDeforming");
#endif
	SymMatrix & f = deform->DeformAtCurrentHierarchyLevel(theta, &penalty);
#ifdef CALCULATE_TIMING	
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("OptimizationDeforming", false);
#endif
	
#ifdef CALCULATE_TIMING	
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("OptimizationEnergyCalc");
#endif	
	double e = energy->GetEnergy(f);
#ifdef CALCULATE_TIMING	
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("OptimizationEnergyCalc", false);
#endif
	
	if (penalty<SYM_INFINITY && e<SYM_INFINITY)
		e += penalty;
	else
		e = SYM_INFINITY;
	((SymOptimizer*)params)->m_iFunctionEvals++;
	
#ifdef CALCULATE_TIMING
		TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("OptimizationFullIteration", false);	
#endif
	return e;
}

void optimDF(const gsl_vector * x, void * params, gsl_vector * g)
{	
	double tempF;
	optimFDF(x, params, &tempF, g);
}

void optimFDF(const gsl_vector * x, void * params, double * f, gsl_vector * g)
{
	((SymOptimizer*)params)->m_iDerivativeEvals++;	
	gsl_vector * trg = ((SymOptimizer*)params)->m_pTempVec;
	gsl_vector_memcpy(trg, x);
	*f = optimF(trg, params);
	
	double dx = ((SymOptimizer*)params)->m_fStep;
	for (int i=0; i<(int)x->size; i++)	//simple forward difference for derivative 
		gsl_vector_set(g, i, ((SymOptimizer*)params)->CentralDifference(trg, i, dx));
}
									
double SymOptimizer::ForwardDifference(gsl_vector * params, int dim, double dx, double currVal)
{
	gsl_vector_set(params, dim, gsl_vector_get(params, dim)+dx);
	double nextVal = optimF(params, this);
	gsl_vector_set(params, dim, gsl_vector_get(params, dim)-dx);	
	
	if (nextVal==SYM_INFINITY || currVal==SYM_INFINITY)
		return 0;
	else
		return (nextVal-currVal)/dx;
}

double SymOptimizer::CentralDifference(gsl_vector * params, int dim, double dx)
{
	gsl_vector_set(params, dim, gsl_vector_get(params, dim)-dx/2.);
	double prevVal = optimF(params, this);
	gsl_vector_set(params, dim, gsl_vector_get(params, dim)+dx);
	double nextVal = optimF(params, this);	
	gsl_vector_set(params, dim, gsl_vector_get(params, dim)-dx/2.);	
	if (nextVal==SYM_INFINITY || prevVal==SYM_INFINITY)
		return 0;
	else		
		return (nextVal-prevVal)/dx;
}

SymOptimizer::SymOptimizer(SymEnergy * energyFn, DeformationModel * deformationModel, int maxIterations, 
						   int maxFunctionEvals, int updateAfterIterations, double rescaleAtUpdate,
						   ExampleWithAnswer * perfectSolution)
{
	assert(updateAfterIterations>0);
		
	m_fStep = 0;
	m_pTempVec = NULL;
	m_iOutputBias = 0;
	m_fParamChangeLastIter = 0;	
	m_pEnergyFn = energyFn;
	if (perfectSolution==NULL)
		m_pDeformationModel = deformationModel;
	else
		m_pDeformationModel = perfectSolution->m_pValidSolutionSpace;
	m_pAnswer = perfectSolution;
	m_iFunctionEvals = 0;
	m_iNumIterations = 0;
	m_IterationType = ITERATE_BY_FN_EVALS;
	m_iMaxIterations = maxIterations;
	m_iMaxFunctionEvals = maxFunctionEvals;
	m_bVerbous = updateAfterIterations>0;
	m_iUpdateEvery = updateAfterIterations;
	m_fRescaleAtUpdate = rescaleAtUpdate;
	m_bStepIsTooSmall = false;
}

SymVectorReal * SymOptimizer::GetBestSolution()
{
	return m_pBestSolution;
}

bool SymOptimizer::KeepIterating(int maxIters)
{
	m_iNumIterations++;		
	switch(m_IterationType)
	{
		case ITERATE_BY_FN_EVALS:
			return m_iFunctionEvals-m_iTimeStampFnEvals < maxIters;
		case ITERATE_BY_ALGORITHM_ITERS:
			return m_iNumIterations-m_iTimeStampIterations-1 < maxIters;			
		case ITERATE_BY_TIME:
		default:
			std::cout<<"SymOptimizer::KeepIterating - ITERATE_BY_TIME: Not implemented yet"<<std::endl;
			exit(0);
			return false;
	}
}

void SymOptimizer::StartIterating()
{
	m_iTimeStampFnEvals = m_iFunctionEvals;
	m_iTimeStampIterations = m_iNumIterations;
}

bool SymOptimizer::CheckSuccess()
{
	return m_iFunctionEvals >= m_iMaxFunctionEvals || m_iNumIterations >= m_iMaxIterations || m_bStepIsTooSmall;
}

bool SymOptimizer::RequireUpdate(double bestEnergy, double currentStepChange, bool saveHistoryLandmarkAndRescale)
{
	bool update = false;
	m_EnergyHistory.push_back(bestEnergy);
	m_DXStepHistory.push_back(currentStepChange);
	m_iFnEvalsHistory.push_back(m_iFunctionEvals);
	
	if (saveHistoryLandmarkAndRescale)
	{
		if (m_HistoryLandmark.size()>0
			&& m_EnergyHistory[m_HistoryLandmark[m_HistoryLandmark.size()-1]]==bestEnergy
			&& m_fRescaleAtUpdate<1)	//rescale
		{
			std::cout<<"Rescaling by factor: "<<m_fRescaleAtUpdate<<std::endl;
			m_pDeformationModel->ReduceSuggestedStepByFactor(m_fRescaleAtUpdate);
			if (m_bVerbous)
				m_pDeformationModel->PrintParams(true);
			m_bStepIsTooSmall = true;
			for (int i=0; i<m_pDeformationModel->GetMinimalParamStep().m && m_bStepIsTooSmall; i++)
				m_bStepIsTooSmall = m_pDeformationModel->GetMinimalParamStep()(i) > m_pDeformationModel->GetSuggestedParamStep()(i);
			update = true;
		}
			
		m_HistoryLandmark.push_back(m_iFnEvalsHistory.size()-1);
	}
	return update;
}

void SymOptimizer::PrintUpdate()
{
	//print text for curr iteration	
	m_Profiler.finishedProcess("iterations", false);
	std::cout<<"Iteration "<<m_iNumIterations<<std::flush;
	std::cout<<" E="<<m_fBestValue<<" dX="<<m_fParamChangeLastIter<<" Fn="<<m_iFunctionEvals<<std::flush;
	std::cout<<" Time="<<m_Profiler.getTotalProcessLengthInSeconds("iterations").toStdString().c_str()<<std::flush;
	m_Profiler.startedProcess("iterations");
	
	//output images
	double penalty;
	SymMatrix & f = m_pDeformationModel->DeformAtCurrentHierarchyLevel(*GetBestSolution(), &penalty);
	QString numStr = SymPipeline::AddTrailingZeros(m_iOutputBias+m_HistoryLandmark.size(), 6);
	if (m_pDeformationModel->HasOutput(DeformationModel::DEFORM_OUT_SYMREAL2D))
	{
		SymMatrix2DReal & symmetry = (SymMatrix2DReal &)m_pEnergyFn->GetUnderlyingSymmetrySpace()->CalculateSymmetryAndProcess(f);
		QImage img = m_pDeformationModel->GetDeformedImage();
		QImage imgSVis = ((SymSpace2D*)m_pEnergyFn->GetUnderlyingSymmetrySpace())->GetVisualizationOfSymSpace(symmetry, &img);
		imgSVis.save(QString("progress_SfVis_")+numStr+".png");						
	}

	if (m_pDeformationModel->HasOutput(DeformationModel::DEFORM_OUT_IMG2D))
	{
		((SymMatrix2D&)f).ToImage().save("progress_f1D_"+numStr+".png");
		QImage img = m_pDeformationModel->GetDeformedImage();
		img.save(QString("progress_f_")+numStr+".png");
		for (int i=1; i<m_pDeformationModel->GetNumOutputImageTypes(); i++)
			m_pDeformationModel->GetDeformedImage(i).save(QString("progress_")+QString::number(i)+"_f_"+ numStr + ".png");
		
		if (m_pAnswer!=NULL)
		{
			//QImage tempImg = ((SymMatrix2DReal&)f).ToImage();
			double imgError = m_pAnswer->GetImageSpaceError(img);
			std::cout<<"\n\tErrorImg="<<imgError<<std::flush;
			m_ImageErrors.push_back(imgError);
		}
	}	

	if (m_pAnswer!=NULL)
	{
		std::cout<<" ErrorParam="<<m_pAnswer->GetParameterSpaceError(m_pBestSolution)<<std::flush;		
		m_ParamErrors.push_back(m_pAnswer->GetParameterSpaceError(m_pBestSolution));
	}
	std::cout<<std::endl;
}

/*********************** GSL SMOOTH OPTIMIZER *******************/
SymOptimizerGSLSmooth::SymOptimizerGSLSmooth(int numVariables, double step, double minLineTol, double minStepNorm,  
											 SymEnergy * energyFn, DeformationModel * deformationModel, 
											 OptimizationAlgirithm algorithm, int maxIterations, int maxFunctionEvals, 
											 int updateAfterIterations, double rescaleAtUpdate,
											 ExampleWithAnswer * perfectSolution)
: SymOptimizer(energyFn, deformationModel, maxIterations, maxFunctionEvals, updateAfterIterations, 
			   rescaleAtUpdate, perfectSolution)
{
	m_fMinStep = minStepNorm;
	m_fStep = step;
	m_fMinLineTol = minLineTol;
	m_iDerivativeEvals = 0;
	m_iNumVars = numVariables;
	m_pTempVec = gsl_vector_alloc(m_iNumVars);
	m_pBestSolution = new SymVectorReal(m_iNumVars);
	m_fBestValue = SYM_INFINITY;
	
	m_Algorithm = algorithm;
	if (m_Algorithm==FDF_CONJ_FLETCHER_REEVES)	//allocate solver
		m_pFDFMinimizer = gsl_multimin_fdfminimizer_alloc(gsl_multimin_fdfminimizer_conjugate_fr, m_iNumVars);
	else if (m_Algorithm==FDF_CONJ_POLAK_RIBIERE)
		m_pFDFMinimizer = gsl_multimin_fdfminimizer_alloc(gsl_multimin_fdfminimizer_conjugate_pr, m_iNumVars);
	else if (m_Algorithm==FDF_BFGS)
		m_pFDFMinimizer = gsl_multimin_fdfminimizer_alloc(gsl_multimin_fdfminimizer_vector_bfgs2, m_iNumVars);
	else if (m_Algorithm==FDF_GRAD_DESC)
		m_pFDFMinimizer = gsl_multimin_fdfminimizer_alloc(gsl_multimin_fdfminimizer_steepest_descent, m_iNumVars);
	else
	{
		std::cout<<"SymOptimizerGSLSmooth::SymOptimizerGSLSmooth - Wrong optimizer for given algorithm"<<std::endl;
		exit(0);
	}
}

void SymOptimizerGSLSmooth::Iterate()
{	
	StartIterating();
	gsl_vector * currentBest = NULL;
	double currentBestValue = SYM_INFINITY;
	int terminateChange = GSL_CONTINUE;
	bool update = true;
	m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
	double penalty = 0;
	SymMatrix2DReal &tempMatrix = (SymMatrix2DReal &)m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector(), &penalty);	
	m_fBestValue = m_pEnergyFn->GetEnergy(tempMatrix);

	if (penalty<SYM_INFINITY && m_fBestValue<SYM_INFINITY)
		m_fBestValue += penalty;
	else
		m_fBestValue = SYM_INFINITY;

	if (m_bVerbous)
	{
		m_Profiler.startedProcess("iterations");
		PrintUpdate();	
		RequireUpdate(m_fBestValue, -1, true);
	}
	
	while (terminateChange!=GSL_SUCCESS && !CheckSuccess())
	{
		if (update)		//update optimizer
		{
			if (m_bVerbous)
				std::cout<<"Initializing optimizer..."<<std::flush;
			
			gsl_multimin_function_fdf fOpt;
			fOpt.n = m_iNumVars;
			fOpt.f = &optimF;
			fOpt.df = &optimDF;
			fOpt.fdf = &optimFDF;
			fOpt.params = (void *)this;
//			std::cout<<"step="<<m_fStep<<" minLineTol="<<m_fMinLineTol<<std::endl;
			gsl_multimin_fdfminimizer_set(m_pFDFMinimizer, &fOpt, m_pDeformationModel->GetParameterVector().m_pData, 
										  m_fStep, m_fMinLineTol);
			update = false;
			if (m_bVerbous)
				std::cout<<" done!"<<std::endl;			
		}
		StartIterating();
		while (KeepIterating(m_iUpdateEvery) && terminateChange!=GSL_SUCCESS && !CheckSuccess())
		{
			gsl_multimin_fdfminimizer_iterate(m_pFDFMinimizer);
			currentBest = gsl_multimin_fdfminimizer_x(m_pFDFMinimizer);
			currentBestValue = gsl_multimin_fdfminimizer_minimum(m_pFDFMinimizer);
			gsl_vector * gradientVec = gsl_multimin_fdfminimizer_gradient(m_pFDFMinimizer);
			terminateChange = gsl_multimin_test_gradient(gradientVec, m_fMinStep);
			RequireUpdate(currentBestValue, -1, false);
		}

		if (currentBestValue <= m_fBestValue && currentBest!=NULL)
		{
			m_fBestValue = currentBestValue;
			m_pBestSolution->CopyValuesFrom(currentBest);
		}
		
		if (m_bVerbous)
			PrintUpdate();
		update = RequireUpdate(currentBestValue, -1, true);		
	}
}

/*********************** GSL DISCRETE OPTIMIZER *******************/
SymOptimizerGSLDiscrete::SymOptimizerGSLDiscrete(int numVariables, SymEnergy * energyFn, DeformationModel * deformationModel,
												 int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
												 double rescaleAtUpdate, ExampleWithAnswer * perfectSolution)
: SymOptimizer(energyFn, deformationModel, maxIterations, maxFunctionEvals, updateAfterIterations, 
			   rescaleAtUpdate, perfectSolution)
{
	m_iNumVars = numVariables;
	m_pBestSolution = new SymVectorReal(m_iNumVars);
	m_fBestValue = SYM_INFINITY;	
	m_pFMinimizer =  gsl_multimin_fminimizer_alloc (gsl_multimin_fminimizer_nmsimplex2, m_iNumVars);
	m_fMinStep = 0;	
	std::cout<<"\tInitializig Discrete optimizer with "<<m_iNumVars<<" variables."<<std::endl;		
}

void SymOptimizerGSLDiscrete::Iterate()
{
	gsl_vector * currentBest = NULL;
	double currentBestValue = SYM_INFINITY;
	int terminateChange = GSL_CONTINUE;
	bool update = true;
	m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
	double penalty = 0;
	SymMatrix2DReal &tempMatrix = (SymMatrix2DReal &)m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector(), &penalty);	
	m_fBestValue = m_pEnergyFn->GetEnergy(tempMatrix);

	if (penalty<SYM_INFINITY && m_fBestValue<SYM_INFINITY)
		m_fBestValue += penalty;
	else
		m_fBestValue = SYM_INFINITY;
	
	if (m_bVerbous)
	{
		m_Profiler.startedProcess("iterations");
		PrintUpdate();	
		RequireUpdate(m_fBestValue, -1, true);
	}
	
	while (terminateChange!=GSL_SUCCESS && !CheckSuccess())
	{
		if (update)		//update optimizer
		{
			if (m_bVerbous)
				std::cout<<"Initializing GSL Discrete optimizer..."<<std::endl;
			
			gsl_multimin_function fOpt;
			fOpt.n = m_iNumVars;		
			fOpt.f = &optimF;
			fOpt.params = (void *)this;

			gsl_multimin_fminimizer_set(m_pFMinimizer, &fOpt, m_pDeformationModel->GetParameterVector().m_pData, 
										m_pDeformationModel->GetSuggestedParamStep().m_pData);			
			update=false;			
			if (m_bVerbous)
				std::cout<<" done!"<<std::endl;			
		}
		StartIterating();				
		while (KeepIterating(m_iUpdateEvery) && terminateChange!=GSL_SUCCESS && !CheckSuccess())
		{
			gsl_multimin_fminimizer_iterate(m_pFMinimizer);			
#ifdef CALCULATE_TIMING
			std::cout<<"Num Function Evals: "<<m_iFunctionEvals<<std::endl;
			std::cout<<TimeProfiler::m_pGlobalSuperProfiler->profile().toStdString().c_str()<<std::endl;
#endif			
			currentBest = gsl_multimin_fminimizer_x(m_pFMinimizer);
			currentBestValue = gsl_multimin_fminimizer_minimum(m_pFMinimizer);
			double varChange = gsl_multimin_fminimizer_size(m_pFMinimizer);
			terminateChange = gsl_multimin_test_size(varChange, m_fMinStep);	
		}
		if (currentBestValue <= m_fBestValue && currentBest!=NULL)
		{
			m_fBestValue = currentBestValue;
			m_pBestSolution->CopyValuesFrom(currentBest);
		}
		
		if (m_bVerbous)
			PrintUpdate();
		update = RequireUpdate(currentBestValue, -1, true);		
	}
}



/*********************** Hierarchical Optimizer *******************/
MultiLevelOptimizer::MultiLevelOptimizer(SymEnergy * energyFn, DeformationModel * deformationModel, 
										 OptimizationAlgirithm algorithm, int maxIterations, int maxFunctionEvals, 
										 int updateAfterIterations, double rescaleAtUpdate, bool refeedImages,
										 ExampleWithAnswer * perfectSolution, int numDims)
: SymOptimizer(energyFn, deformationModel, maxIterations, maxFunctionEvals, updateAfterIterations, 
			   rescaleAtUpdate, perfectSolution) 
{
	m_bRefeedImageAtEachIteration = refeedImages;
	m_Algorithm = algorithm;
	m_iNumLevels = deformationModel->GetNumHierarchyLevels();
	
	for (int i=0; i<m_iNumLevels; i++)
	{
		//for each level initialize new optimizer
		deformationModel->SwitchToHierarchy(i);
		SymOptimizer * optimizer=NULL;
		
		int newMaxIterations = maxIterations/m_iNumLevels;
		int newMaxFunctionEvals = maxFunctionEvals/m_iNumLevels;
		
		if (m_Algorithm==SymOptimizer::FDF_CONJ_FLETCHER_REEVES || m_Algorithm==SymOptimizer::FDF_CONJ_POLAK_RIBIERE
				 || m_Algorithm==SymOptimizer::FDF_GRAD_DESC || m_Algorithm==SymOptimizer::FDF_BFGS)
		{
			optimizer = new SymOptimizerGSLSmooth(deformationModel->GetParameterVector().m, 
												  deformationModel->GetSuggestedParamStep()(0), 0, 0, energyFn, 
												  deformationModel, m_Algorithm, newMaxIterations, 
												  newMaxFunctionEvals, updateAfterIterations, rescaleAtUpdate,
												  m_pAnswer);
		}
		else if (m_Algorithm==SymOptimizer::F_SIMPLEX_NELDER_MEAD)
		{
			optimizer = new SymOptimizerGSLDiscrete(deformationModel->GetParameterVector().m, energyFn, 
													deformationModel, newMaxIterations, newMaxFunctionEvals,
													updateAfterIterations, rescaleAtUpdate, m_pAnswer);
		}
		else if (m_Algorithm==SymOptimizer::GREEDY_WALK)
		{
			optimizer = new SymOptimizerGreedyWalk(energyFn, deformationModel, newMaxIterations, 
												   newMaxFunctionEvals, updateAfterIterations, 
												   rescaleAtUpdate, m_pAnswer);
		}
		else if (m_Algorithm==SymOptimizer::MY_GRADIENT_DESCENT)
		{
//			assert(false);	//set num dims
			optimizer = new SymOptimizerNDGradientDescent(numDims, energyFn, deformationModel, 
														  newMaxIterations, newMaxFunctionEvals, updateAfterIterations, 
														  rescaleAtUpdate, m_pAnswer);
		}
	
		assert(optimizer!=NULL);
		m_OptimizersPerLevel.push_back(optimizer);
	}
}

void MultiLevelOptimizer::Iterate()
{
//	if (m_pAnswer!=NULL)
//	{
//		m_pBestSolution = m_pAnswer->m_pPerfectSolution;		
//		return;
//	}
	
	if (m_bVerbous)
	{
		std::cout<<"Multi-level optimizer. Iterating: "<<std::endl;
		std::cout<<"Initial Steps:";		m_pDeformationModel->PrintParams(0.00001);		std::cout<<std::endl;
	}
	
	for (int i=0; i<m_iNumLevels; i++)
	{
		m_pDeformationModel->SwitchToHierarchy(i);		
		if (m_bVerbous)
			std::cout<<"+++++++++++++++++++ Optimizing at level "<<i<<" vars="<<m_pDeformationModel->GetParameterVector().m<<" ++++++++++++++++++"<<std::endl;
		
		m_OptimizersPerLevel[i]->m_pDeformationModel = m_pDeformationModel;
		
		m_OptimizersPerLevel[i]->Iterate();

		for (int j=0; j<(int)m_OptimizersPerLevel[i]->m_HistoryLandmark.size(); j++)
			m_HistoryLandmark.push_back(m_OptimizersPerLevel[i]->m_HistoryLandmark[j]);
		
		for (int j=0; j<(int)m_OptimizersPerLevel[i]->m_EnergyHistory.size(); j++)
		{
			m_EnergyHistory.push_back(m_OptimizersPerLevel[i]->m_EnergyHistory[j]);
			m_DXStepHistory.push_back(m_OptimizersPerLevel[i]->m_DXStepHistory[j]);
			m_iFnEvalsHistory.push_back(m_OptimizersPerLevel[i]->m_iFnEvalsHistory[j]);
		}
		
		for (int j=0; j<(int)m_OptimizersPerLevel[i]->m_ParamErrors.size(); j++)
		{
			m_ParamErrors.push_back(m_OptimizersPerLevel[i]->m_ParamErrors[j]);
			m_ImageErrors.push_back(m_OptimizersPerLevel[i]->m_ImageErrors[j]);
		}
		
		if (i<m_iNumLevels-1)
		{
			m_OptimizersPerLevel[i+1]->m_iOutputBias = m_HistoryLandmark.size();
		}
		if (m_bRefeedImageAtEachIteration && i!=m_iNumLevels-1)
		{
			DeformationModel * oldModel = m_pDeformationModel;
			m_pDeformationModel = m_pDeformationModel->CloneSelfUsingBestParams(*(m_OptimizersPerLevel[i]->GetBestSolution()));
			delete oldModel;
		}
	}
	
	if (m_bVerbous)
		std::cout<<"Finished Multi-level optimization"<<std::endl;
	
	m_pBestSolution = new SymVectorReal(*(m_OptimizersPerLevel[m_iNumLevels-1]->GetBestSolution()));
}

SymOptimizerGreedyWalk::SymOptimizerGreedyWalk(SymEnergy * energyFn, DeformationModel * deformationModel, 
											   int maxIterations, int maxFunctionEvals, int updateAfterIterations, 
											   double rescaleAtUpdate, ExampleWithAnswer * perfectSolution)
: SymOptimizer(energyFn, deformationModel, maxIterations, maxFunctionEvals, updateAfterIterations, 
			   rescaleAtUpdate, perfectSolution) 

{
	m_pBestSolution = new SymVectorReal(deformationModel->GetParameterVector().m);
	m_fBestValue = SYM_INFINITY;
	m_bNoOpimization = false;
}

void SymOptimizerGreedyWalk::Iterate()
{
	int terminateChange = GSL_CONTINUE;
	bool update = true;
	m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
	SymMatrix2DReal &tempMatrix = (SymMatrix2DReal &)m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector());
	m_fBestValue = m_pEnergyFn->GetEnergy(tempMatrix);
	if (m_bVerbous)
	{
		m_Profiler.startedProcess("iterations");
		PrintUpdate();	
		RequireUpdate(m_fBestValue, -1, true);		
	}
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("xx=======TotalTime");
#endif
	
	while (terminateChange!=GSL_SUCCESS && !CheckSuccess())
	{
		if (update)		//update optimizer ? nothing
			update = false;
		
		StartIterating();				
		while (KeepIterating(m_iUpdateEvery) && terminateChange!=GSL_SUCCESS && !CheckSuccess())
		{
			m_iFunctionEvals++;
#ifdef CALCULATE_TIMING
			if (m_iFunctionEvals%200==0)
			{
				std::cout<<"Function Evals="<<m_iFunctionEvals<<std::endl;
				TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("xx=======TotalTime", false);
				std::cout<<TimeProfiler::m_pGlobalSuperProfiler->profile().toStdString().c_str()<<std::endl;
				TimeProfiler::m_pGlobalSuperProfiler->startedProcess("xx=======TotalTime");			
			}
			TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymDeform:RandomStep");			
#endif
			bool enforceStep = false;
			SymMatrix & deformMatrix = m_pDeformationModel->TakeRandomStep(enforceStep);
#ifdef CALCULATE_TIMING			
			TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymDeform:RandomStep", false);
#endif
			double energy = m_pEnergyFn->GetEnergy(deformMatrix);
			if (energy < m_fBestValue || m_bNoOpimization || enforceStep)
			{
				m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
				m_fBestValue = energy;				
			}
			else
			{
#ifdef CALCULATE_TIMING							
				TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymDeform:RandomStep");
#endif
				m_pDeformationModel->RevertLastRandomStep();
#ifdef CALCULATE_TIMING							
				TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymDeform:RandomStep", false);				
#endif
				m_pDeformationModel->GetParameterVector().CopyValuesFrom(m_pBestSolution->m_pData);
			}
			if (m_pDeformationModel->TerminalState())
				terminateChange = GSL_SUCCESS;
		}

		if (m_bVerbous)
			PrintUpdate();
		update = RequireUpdate(m_fBestValue, -1, true);		
	}	
}

SymOptimizerNDGradientDescent::SymOptimizerNDGradientDescent(int numDims, SymEnergy * energyFn, DeformationModel * deformationModel, 
							  int maxIterations, int maxFunctionEvals, int update, 
							  double rescaleAtUpdate, ExampleWithAnswer * perfectSolution)
: SymOptimizer(energyFn, deformationModel, maxIterations, maxFunctionEvals, 
			   update < 0 ? 3*deformationModel->GetParameterVector().m : update, 
			   rescaleAtUpdate, perfectSolution)
{
	m_pTempVec = gsl_vector_alloc(deformationModel->GetParameterVector().m);	
	m_iNumDimensions = numDims;
	m_iGroupedDims = 0;
	if (m_iNumDimensions==-1)
		m_iNumDimensions = deformationModel->GetParameterVector().m;
	else if (m_iNumDimensions<-1) // variables grouped by dimension rather than by coord ID
	{
		m_iGroupedDims = -m_iNumDimensions;
		m_iNumDimensions = deformationModel->GetParameterVector().m/m_iGroupedDims;
	}
	assert(deformationModel->GetParameterVector().m%m_iNumDimensions==0);
	std::cout<<"Num Dimensions = "<<m_iNumDimensions<<std::endl;
	m_pBestSolution = new SymVectorReal(deformationModel->GetParameterVector().m);
	m_fBestValue = SYM_INFINITY;
	m_pGradDirection = new SymVectorReal(m_iNumDimensions);
	m_fStep = 0;
}

void SymOptimizerNDGradientDescent::Iterate()
{	
	int terminateChange = GSL_CONTINUE;
	bool update = true;
	m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
	SymMatrix2DReal &tempMatrix = (SymMatrix2DReal &)m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector());

	m_fBestValue = m_pEnergyFn->GetEnergy(tempMatrix);
//	std::cout<<"MYGRADIENTDESCEND: best="<<m_fBestValue<<std::endl;
//	int param = 1;	
//	std::cout<<"Initial Parameter["<<param<<"] = "<<m_pDeformationModel->GetParameterVector()(param)<<std::endl;	
//	std::cout<<"TEMPORARY TEST\nE(55):\n"<<std::flush;
//	double penalty;
//	
//	double scale;
//	FindGradientDirection(m_iNumDimensions, m_iGroupedDims, 0, m_pEnergyFn, m_pDeformationModel, 
//						  *m_pGradDirection, &scale, m_fBestValue);
//	std::cout<<"Gradient Direction = "<<std::flush;
//	m_pGradDirection->PrintVector(0); 
//	std::cout<<std::endl;
//	
//	std::cout<<"plot(["<<m_pEnergyFn->GetEnergy(((SymMatrix2DReal&)m_pDeformationModel->TakeStepInDimension(param, 0, penalty)))<<" "<<std::flush;
//	m_pDeformationModel->GetDeformedImage().save("img"+QString::number(0)+".png");
//	for (int i=0; i<20; i++)
//	{		
//		std::cout<<m_pEnergyFn->GetEnergy(((SymMatrix2DReal&)m_pDeformationModel->TakeStepInDimension(param, -4.*2*3.14/360, penalty)))<<" "<<std::flush;
//		m_pDeformationModel->GetDeformedImage().save("img"+QString::number(i+1)+".png");
//	}
//	std::cout<<"]);"<<std::endl;
//	std::cout<<"DONE TEMPORARY TEST!"<<std::endl;
//	exit(0);	
	
	if (m_bVerbous)
	{
		m_Profiler.startedProcess("iterations");
		PrintUpdate();	
		RequireUpdate(m_fBestValue, -1, true);		
	}
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("xx=======TotalTime");
#endif
	int numParams = m_pDeformationModel->GetParameterVector().m/m_iNumDimensions;
	assert(m_pDeformationModel->GetParameterVector().m%m_iNumDimensions==0);
	int parameterID = 0;	
	while (terminateChange!=GSL_SUCCESS && !CheckSuccess())
	{
		if (update)		//update optimizer ? nothing
			update = false;
		
		StartIterating();
		bool tookStep=false;
		while (KeepIterating(m_iUpdateEvery) && terminateChange!=GSL_SUCCESS && !CheckSuccess())
		{
			//std::cout<<"Iterations: "<<m_iFunctionEvals<<std::endl;
			m_iFunctionEvals+=1+m_iNumDimensions;
#ifdef CALCULATE_TIMING
			std::cout<<"Function Evals="<<m_iFunctionEvals<<std::endl;
			TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("xx=======TotalTime", false);
			std::cout<<TimeProfiler::m_pGlobalSuperProfiler->profile().toStdString().c_str()<<std::endl;
			TimeProfiler::m_pGlobalSuperProfiler->startedProcess("xx=======TotalTime");
#endif
			double optionalScale=1;	// relative to min step

//			std::cout<<"===== "<<parameterID<<" Finding Gradient Direction ======"<<std::endl;
			FindGradientDirection(m_iNumDimensions, m_iGroupedDims, parameterID, m_pEnergyFn, m_pDeformationModel, 
								  *m_pGradDirection, &optionalScale, m_fBestValue);
			double penalty=0;
			double energy = m_fBestValue+100;
//			std::cout<<"\tGrad["<<parameterID<<"] = ["<<(*m_pGradDirection)(0)<<", "<<(*m_pGradDirection)(1);
//			std::cout<<"] * "<<optionalScale<<"=["<<(*m_pGradDirection)(0)*optionalScale<<", "<<(*m_pGradDirection)(1)*optionalScale<<"] "<<std::endl;
			//std::cout<<"NumEvals="<<m_iFunctionEvals<<std::endl;
//			std::cout<<"Grad Direction ["<<parameterID<<"] "<<std::flush;
//			m_pGradDirection->PrintVector(0);			std::cout<<" * "<<optionalScale<<std::endl;
			
			int currParamID;
			double fullGradStep=0;
			if (m_iNumDimensions==m_pDeformationModel->GetParameterVector().m)
			{
				for (int i=0; i<m_iNumDimensions; i++)
				{
					int currParamID = GetParameterID(m_iNumDimensions, m_iGroupedDims, parameterID, i);
					m_pDeformationModel->GetParameterVector()(currParamID) -= (*m_pGradDirection)(i)*optionalScale;
					fullGradStep += qAbs(pow((*m_pGradDirection)(i)*optionalScale, 2));
				}
				
				energy = m_pEnergyFn->GetEnergy(m_pDeformationModel->DeformAtCurrentHierarchyLevel(m_pDeformationModel->GetParameterVector(), &penalty));
			}
			else
			{
//				std::cout<<"numDims="<<m_iNumDimensions<<std::endl;
				for (int i=0; i<m_iNumDimensions; i++)
				{
					if ((*m_pGradDirection)(i)*optionalScale==0)
						continue;
					int currParamID = GetParameterID(m_iNumDimensions, m_iGroupedDims, parameterID, i);
					if (i==m_iNumDimensions-1)
						energy = m_pEnergyFn->GetEnergy(m_pDeformationModel->TakeStepInDimension(currParamID, -(*m_pGradDirection)(i)*optionalScale, penalty));
					else
						m_pDeformationModel->TakeStepInDimension(currParamID, -(*m_pGradDirection)(i)*optionalScale, penalty);
					fullGradStep += qAbs(pow((*m_pGradDirection)(i)*optionalScale, 2));					
//					std::cout<<"\tenergy["<<currParamID<<"] = "<<m_pDeformationModel->GetParameterVector()(currParamID)<<": "<<energy;
//					std::cout<<"\t\tGrad="<<-(*m_pGradDirection)(i)*optionalScale<<std::endl;
				}
//				if (fullGradStep!=0)
//				{
//					std::cout<<"Saving... GRADSTEP.png"<<std::endl;
//					static int gradStepID=0;
//					m_pDeformationModel->GetDeformedImage().save("Gradstep"+QString::number(gradStepID++)+".png");
//				}
			}
//			assert(energy>=0);
			
			if (penalty<SYM_INFINITY)
				energy += penalty;
			else
				energy = SYM_INFINITY;
				
//			std::cout<<"value["<<m_pDeformationModel->GetParameterVector()(0)<<"]="<<energy<<std::endl;
			if (energy < m_fBestValue)
			{
				tookStep = true;
//				std::cout<<"\t\tStep Taken"<<std::endl;
				m_pBestSolution->CopyValuesFrom(m_pDeformationModel->GetParameterVector().m_pData);
				m_fBestValue = energy;				
			}
//			else	//would not work for point distributions
//			{
//				if (initScale>1)
//				{
//					initScale = qMax(initScale/2, 1.);
//					std::cout<<"New optional scale="<<initScale<<std::endl;				
//				}
//				m_pDeformationModel->GetParameterVector().CopyValuesFrom(m_pBestSolution->m_pData);
//			}
			parameterID = (parameterID+1) % numParams;
			if (parameterID==0)
			{
				if (!tookStep)
					break;
				else
					tookStep=false;					
			}
		}
		
		if (m_bVerbous)
			PrintUpdate();
		update = RequireUpdate(m_fBestValue, -1, true);		
	}		
}

int SymOptimizerNDGradientDescent::GetParameterID(int numDims, int grouped, int parameterID, int dim)
{
	if (grouped==0)	// no grouping
		return parameterID*numDims+dim;
	else
		return dim*grouped+parameterID;	
}

void SymOptimizerNDGradientDescent::FindGradientDirection(int numDims, int grouped, int parameterID, SymEnergy * energyFn, 
				DeformationModel * deformationModel, SymVectorReal & gradDirection, double * optionalScale,
				double currVal)
{
	SymVectorReal & params = deformationModel->GetParameterVector();
	double penalty=0;
	double norm = 0;
	double maxStep = 0;
	for (int dim=0; dim<numDims; dim++)
	{
		int currParamID = GetParameterID(numDims, grouped, parameterID, dim);
		double dx=2*deformationModel->GetMinimalParamStep()(currParamID);
		
		double penalty = 0;
		double newValue = energyFn->GetEnergy(deformationModel->TakeStepInDimension(currParamID, dx/2., penalty));			//forward difference
//		std::cout<<"value["<<deformationModel->GetParameterVector()(0)<<"]="<<newValue<<std::endl;
		if (penalty<SYM_INFINITY)
			newValue += penalty;
		else
			newValue = SYM_INFINITY;

		double prevValue = energyFn->GetEnergy(deformationModel->TakeStepInDimension(currParamID, -dx, penalty));			//forward difference

		if (newValue==SYM_INFINITY || prevValue==SYM_INFINITY || (newValue>currVal && prevValue>currVal))
		{
//			std::cout<<"Derivative["<<parameterID*numDims+dim<<"] = 0. Vals=["<<prevValue<<", "<<currVal<<", "<<newValue<<"]"<<std::endl;
			newValue = 0;
			prevValue = 0;
		}
		
//		std::cout<<"\t\tdf/dx_"<<dim<<"=("<<newValue<<"-"<<prevValue<<")/2 = "<<(newValue - prevValue)/dx<<std::endl;
		gradDirection(dim) = (newValue - prevValue)/dx;
		//std::cout<<"GradDim["<<dim<<"]="<<gradDirection(dim)<<std::endl;
		maxStep = qMax(maxStep, qAbs(gradDirection(dim)));
		
		deformationModel->TakeStepInDimension(currParamID, dx/2., penalty);
	}
	
	int stepID = GetParameterID(numDims, grouped, parameterID, 0);
	
	*optionalScale = (maxStep==0) ? (1.) : (deformationModel->GetSuggestedParamStep()(stepID) / maxStep);
	
//	if (maxStep>0 && maxStep < deformationModel->GetMinimalParamStep()(parameterID*numDims))
//		*optionalScale = deformationModel->GetMinimalParamStep()(parameterID*numDims) / maxStep;
//	else if (maxStep>0 && maxStep > deformationModel->GetSuggestedParamStep()(parameterID*numDims))
//		*optionalScale = deformationModel->GetSuggestedParamStep()(parameterID*numDims) / maxStep;		
//	else
//		*optionalScale = 1;
	
//	std::cout<<"Gradient["<<parameterID<<"] = ["<<gradDirection(0)<<", "<<gradDirection(1)<<"] scale="<<*optionalScale<<std::endl;	
//	std::cout<<"RescaledGradient["<<parameterID<<" = "<<gradDirection(0)**optionalScale<<", "<<gradDirection(1)**optionalScale<<std::endl;	
}


