#include "SymSpace.h"
FFTManager SymSpace::m_FFTManager;

/****************** Symetry spaces base class ***************************/
SymSpace::SymSpace(double p, NormalizationType norm, double constNormalizationScalar, 
				   int iterativeApplication, bool cropToDisk)
{
	m_bCropToDisk = cropToDisk;
	m_pSymmetryValues = NULL;
	this->m_Norm = norm;
	this->p = p;
	this->m_fConstNormalizationScalar = constNormalizationScalar;
	m_iIterativeSymmetry = iterativeApplication;
	assert(m_iIterativeSymmetry>=1);	
	
	m_iVisR = 0; 
	m_iVisG = 0;
	m_iVisB = 96;
	m_fVisCropMin = -1;
	m_fVisCropMax = -1;
	m_fVisAttenuateGamma = -1;
	
}

void SymSpace::SetCenterOfMass()
{
	assert(false);
}

void SymSpace::Normalize(SymMatrix & SfGen, double L2f)
{
	if (SfGen.GetNumDims()==2 && SfGen.GetEntryType()==SYM_ENTRY_REAL)
	{
		SymMatrix2DReal & Sf = (SymMatrix2DReal&)SfGen;
		//SymMatrix2DReal & f = (SymMatrix2DReal&)fGen;		
		
		if (m_Norm==SymSpace::NORM_SYMMETRY_SPACE)
		{
			double normVal = Sf.SumElements(1)/m_fConstNormalizationScalar;
			assert(normVal!=0);
			Sf.DivideBy(normVal);
		}			
		else if (m_Norm==SymSpace::NORM_IMAGE_SPACE)
		{
			//double normVal = ((SymMatrix2DReal &)f).SumElements(2);
			//assert(normVal!=0);			
			assert(L2f!=0);
			Sf.DivideBy(L2f);
		}
		else if (m_Norm==SymSpace::NORM_GAUSS_MEAN0_STD1)
		{
			double mean = Sf.Mean();
			Sf.Subtract(mean);
			double std_dev = sqrt(Sf.SumElements(2)) ;	///(Sf.m * Sf.n)
			assert(std_dev!=0);
			Sf.DivideBy(std_dev/sqrt(Sf.m*Sf.n));
		}
		else if (m_Norm==SymSpace::NORM_SYMMETRY_SPACE_MAX_TO_1)
		{
			double maxVal = Sf.GetMax(true);
			assert(maxVal!=0);			
			Sf.DivideBy(maxVal);			
		}
		
		if (p!=1)
			Sf.PointwisePower(p, true);
	}
	else
	{
		std::cout<<"Error: SymSpace::Normalize: Only real 2D normalization is implemented!"<<std::endl;
		exit(0);
	}	
//	std::cout<<"Done"<<std::endl;
}

SymMatrix & SymSpace::CalculateSymmetryAndProcess(SymMatrix & f, bool normalize, double * L2fout)
{
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymSpace:FullCalculation");
#endif
	if (m_iIterativeSymmetry==1)
	{
		if (m_bCropToDisk)
		{
			assert(f.GetNumDims()==2);			
			CropToDisk((SymMatrix2DReal&)f);
		}
		
//		std::cout<<"Calculating symmetry transform"<<std::endl;
		double L2f = 0;
		CalculateSymmetry(f, &L2f);
		if (L2fout!=NULL)
			*L2fout = L2f;
		if (normalize)
		{
//			std::cout<<"\t1. normalizing: max="<<m_pSymmetryValues->GetMax(false)<<" L2f="<<L2f<<std::endl;
			Normalize(*m_pSymmetryValues, L2f);
//			std::cout<<"\t2. normalized: max="<<m_pSymmetryValues->GetMax(false)<<std::endl;
		}
	}
	else
	{
		CalculateSymmetryMultiIter(f);
	}
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymSpace:FullCalculation", false);
#endif	
	return *m_pSymmetryValues;
}

void SymSpace::CropToDisk(SymMatrix2DReal & f)
{
	SetCenterOfMass();
	double dist1 = qMin(m_Center.X(), ((double)f.m)-m_Center.X());
	double dist2 = qMin(m_Center.Y(), ((double)f.n)-m_Center.Y());			
	double minDist = qMin(dist1, dist2);
//	std::cout<<"center=["<<cx<<", "<<cy<<"] dims="<<f.m<<" x "<<f.n<<" dist="<<minDist<<std::endl;
	minDist *= minDist;
	for (int i=0; i<f.m; i++)
	{
		for (int j=0; j<f.n; j++)
		{
			double dx = m_Center.X()-(((double)i)+0.5);
			double dy = m_Center.Y()-(((double)j)+0.5);
			if (dx*dx + dy*dy > minDist)
				f(i, j) = 0;
		}
	}
	static int croppedID = 0;
	if (croppedID<2)
	{
		std::cout<<"[WARNING] Printing cropped image."<<std::endl;;
		f.ToImage().save("Cropped"+QString::number(croppedID++)+".png");
	}	
}

void SymSpace::CalculateSymmetryMultiIter(SymMatrix & f)
{
	CalculateSymmetry(f, NULL);
	
	for (int i=1; i<m_iIterativeSymmetry; i++)
		CalculateSymmetry(*m_pSymmetryValues, NULL);
}

void SymSpace::SetVisualizationParams(std::vector<int> &colors, 
									  std::vector<double> & crop, double gamma)
{
	if (colors.size()==3)
	{
		m_iVisR = colors[0];
		m_iVisG = colors[1];
		m_iVisB = colors[2];
	}
	
	if (crop.size()==2)
	{
		m_fVisCropMin = crop[0];
		m_fVisCropMax = crop[1];
	}
	
	m_fVisAttenuateGamma = gamma;	
}

SymMatrix * SymSpace::GetFilteredSymSpace(SymMatrix * f, SymFilter * symSpaceFilter)
{
	bool normalizeBeforeFilter = true;
	bool normalizeAfterFilter = true;
	// no filter:				normalize -> no filter
	// filter + ImgSpace:		normalize -> filter	
	// filter + SymSpace:		filter -> normalize
	// filter + Std1Mean0:		filter -> normalize
	// filter + SymSpaceMax1:	filter -> normalize
	if (symSpaceFilter!=NULL)
	{
		std::cout<<"norm="<<m_Norm<<std::endl;
		switch(m_Norm)
		{
			case NORM_IMAGE_SPACE:
			case NORM_NONE:
				normalizeBeforeFilter = true;
				normalizeAfterFilter = false;
				break;
			case NORM_SYMMETRY_SPACE:
			case NORM_GAUSS_MEAN0_STD1:
			case NORM_SYMMETRY_SPACE_MAX_TO_1:
				normalizeBeforeFilter = false;
				normalizeAfterFilter = true;
				break;
			default:
				std::cout<<"[ERROR]: Unknown Symmetry Space"<<std::endl;
				assert(false);
		}
	}
	else
	{
		normalizeBeforeFilter = true;
		normalizeAfterFilter = false;
	}
	
	std::cout<<"Calculating Sg:"<<std::endl;
	double L2f=0;
	SymMatrix * Sg = &(CalculateSymmetryAndProcess(*f, normalizeBeforeFilter, &L2f));	//normalize if no filter

	if (symSpaceFilter!=NULL)
	{
		if (symSpaceFilter->m_sSpecialDirections=="InputIsSymSpace")
		{
			return f;
		}
		else
		{
			std::cout<<"Filtering Sg"<<std::endl;	
			//assert(m_pSymSpaceTargetFilter!=NULL);
			symSpaceFilter->Process(Sg, &Sg);
			if (normalizeAfterFilter)
				Normalize(*Sg, L2f);
		}
	}
	return Sg;
}

void SymSpace::Renormalize(SymMatrix * Sf)
{
	switch(m_Norm)
	{
		case NORM_IMAGE_SPACE:
		case NORM_NONE:
			break;
		case NORM_SYMMETRY_SPACE:
		case NORM_GAUSS_MEAN0_STD1:
		case NORM_SYMMETRY_SPACE_MAX_TO_1:
			Normalize(*Sf, 0);
			break;
		default:
			std::cout<<"[ERROR]: Unknown Symmetry Space"<<std::endl;
			assert(false);
	}
}

SymMatrix * SymSpace::RescaleAndNormalize(SymMatrix * Sf)
{
	// suppose there is no need to re-scale. 
	assert(Sf->GetNumDims()==2 && Sf->GetEntryType()==SYM_ENTRY_REAL);
	SymMatrix2DReal * SfReal = (SymMatrix2DReal *) Sf;
	assert(SfReal->m==((SymSpace2D*)this)->symm_m && SfReal->n==((SymSpace2D*)this)->symm_n);
	return SfReal;
}


SymSpace::~SymSpace()
{
}

/****************** SymSpace 2D ***************************/
SymSpace2D::SymSpace2D(int im_m, int im_n, int sym_m, int sym_n, double p, NormalizationType norm, 
					   double constNormalizationScalar, int iterativeApplication, bool cropToDisk)
: SymSpace(p, norm, constNormalizationScalar, iterativeApplication, cropToDisk)
{
	this->image_m = im_m;
	this->image_n = im_n;
	m_pSymmetryValues = new SymMatrix2DReal(sym_m, sym_n);
	this->symm_m = sym_m;
	this->symm_n = sym_n;
}

SymSpace2D::~SymSpace2D()
{
}

QImage SymSpace2D::GetVisualizationOfSymSpace(SymMatrix2DReal & Sf, QImage * )
{
	double minVal, maxVal;
	if (m_fVisCropMin<m_fVisCropMax)
	{
		minVal=m_fVisCropMin;
		maxVal=m_fVisCropMax;
	}
	else
	{
		minVal = Sf.GetMin(false);
		maxVal = Sf.GetMax(false);
	}
//	std::cout<<"min="<<m_fVisCropMin<<" max="<<m_fVisCropMax<<" gamma="<<m_fVisAttenuateGamma<<std::endl;
	if (m_fVisCropMin<m_fVisCropMax || m_fVisAttenuateGamma>0)
	{
		for (int i=0; i<Sf.m; i++)
		{
			for (int j=0; j<Sf.n; j++)
			{
				double val = Sf(i,j);
				if (m_fVisCropMin<m_fVisCropMax)
				{
					if (val<m_fVisCropMin)
						val = m_fVisCropMin;
					else if (val>m_fVisCropMax)
						val = m_fVisCropMax;
				}
				
				val = (val-minVal) / (maxVal-minVal);
				if (m_fVisAttenuateGamma>0)
					val = pow(val, m_fVisAttenuateGamma);
				Sf(i,j) = val;
			}
		}
	}
	return Sf.ToImage();
}

void SymSpace2D::CropCenter(SymMatrix2DReal & source)
{
	assert(source.m >= symm_m && source.n >=symm_n);
	
	int minX = (int)(((double)source.m) / 2. - ((double)symm_m)/2. + 0.5);
	int minY = (int)(((double)source.n) / 2. - ((double)symm_n)/2. + 0.5);
	source.CopySquare((SymMatrix2D *) m_pSymmetryValues, minX, minY, 0, 0, symm_m, symm_n);
}

void SymSpace2D::CropCorners(SymMatrix2DReal & source)
{
//	std::cout<<"src: "<<source.m<<" x "<<source.n<<" symm: "<<symm_m<<" x "<<symm_n<<std::endl;
	assert(source.m >= symm_m && source.n >=symm_n);
	int upperSymX = ceil(((double)symm_m)/2.);
	int upperSymY = ceil(((double)symm_n)/2.);
	int lowerSymX = floor(((double)symm_m)/2.);
	int lowerSymY = floor(((double)symm_n)/2.);
	
	assert(upperSymX+lowerSymX==symm_m && upperSymY+lowerSymY==symm_n);
						//Src X					SrcY					Trg X			TrgY	DimX		DimY	
	source.CopySquare((SymMatrix2D*)m_pSymmetryValues, 
					  0,						0,						lowerSymX,	lowerSymY,	upperSymX, upperSymY);
	source.CopySquare((SymMatrix2D*)m_pSymmetryValues, 
					  source.m - lowerSymX,		0,						0,			lowerSymY,	lowerSymX, upperSymY);
	source.CopySquare((SymMatrix2D*)m_pSymmetryValues, 
					  0,						source.n - lowerSymY,	lowerSymX,	0,			upperSymX, lowerSymY);
	source.CopySquare((SymMatrix2D*)m_pSymmetryValues, 
					  source.m - lowerSymX,		source.n - lowerSymY,	0,			0,			lowerSymX, lowerSymY);
	
}

/****************** Toroidal translations ***************************/
SymSpaceTransToroidal2D::SymSpaceTransToroidal2D(int m, int n, int s_m, int s_n, NormalizationType norm, double p, 
												 double constNormalizationScalar, int iterativeApplication)
: SymSpace2D(m, n, s_m<0 ? m : s_m, s_n<0 ? n : s_n, p, norm, constNormalizationScalar, iterativeApplication)
{
//	std::cout<<"Dims="<<m<<" x "<<n<<" Symms="<<symm_m<<" x "<<symm_n<<std::endl;
	m_pTmpFFTConj = new SymMatrix2DComplex(m, n);
	m_pTmpFFT = new SymMatrix2DComplex(m, n);
	m_pTmpConvFFT = new SymMatrix2DComplex(m,n);
	m_pTmpFullAnswer = new SymMatrix2DReal(m,n);
}

SymSpaceTransToroidal2D::~SymSpaceTransToroidal2D()
{
}

SymMatrix * SymSpaceTransToroidal2D::RescaleAndNormalize(SymMatrix * Sf)
{
	SymMatrix2DReal * SfReal = ((SymMatrix2DReal*)Sf);
	if (SfReal->m == symm_m && SfReal->n ==  symm_n)
		return Sf;
	else
	{
		CropCenter(*SfReal);
		Renormalize(m_pSymmetryValues);
		return m_pSymmetryValues;
	}
}

void SymSpaceTransToroidal2D::CalculateSymmetry(SymMatrix & f, double *L2f)
{
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymSpaceToroidal:FFT2D");
#endif
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	m_FFTManager.FFT2D((SymMatrix2DReal &)f, *m_pTmpFFT);
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymSpaceToroidal:FFT2D", false);
#endif	
	m_pTmpFFT->Conj(*m_pTmpFFTConj);
	m_pTmpFFT->PointwiseMultiply(*m_pTmpFFTConj, *m_pTmpConvFFT);
	
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->startedProcess("SymSpaceToroidal:IFFT2D");
#endif	
	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
#ifdef CALCULATE_TIMING
	TimeProfiler::m_pGlobalSuperProfiler->finishedProcess("SymSpaceToroidal:IFFT2D", false);
#endif	
	
//	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
//	{
//		Sreal = *m_pTmpFFT;
//	}
//	else
//	{
		*m_pTmpFullAnswer = *m_pTmpFFT;
		CropCorners(*m_pTmpFullAnswer);
//	}
}

void SymSpaceTransToroidal2D::CalculateSymmetryMultiIter(SymMatrix & f)
{
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	SymMatrix2DReal * findS = (SymMatrix2DReal*)(&f);
	
	m_FFTManager.FFT2D(*findS, *m_pTmpFFT);
	for (int i=0; i<m_iIterativeSymmetry; i++)
	{
		m_pTmpFFT->Conj(*m_pTmpFFTConj);
		m_pTmpFFT->PointwiseMultiply(*m_pTmpFFTConj, *m_pTmpConvFFT);
		*m_pTmpFFT = *m_pTmpConvFFT;
		
		if (i==0 && m_Norm==NORM_IMAGE_SPACE)
		{
			m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
			m_pTmpFFT->DivideBy(((SymMatrix2DReal &)f).SumElements(2));
			m_FFTManager.FFT2D(*m_pTmpFFT, *m_pTmpConvFFT);
			*m_pTmpFFT = *m_pTmpConvFFT;
		}
		else if (m_Norm==NORM_SYMMETRY_SPACE_MAX_TO_1)
		{
			double maxVal = m_pTmpFFT->GetMax(true);
			m_pTmpFFT->DivideBy(maxVal);
		}
	}
	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
	
	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
	{
		Sreal = *m_pTmpFFT;
	}
	else
	{
		*m_pTmpFullAnswer = *m_pTmpFFT;
		CropCorners(*m_pTmpFullAnswer);
	}
}

void SymSpaceTransToroidal2D::TestSymmetryTransform()
{
	SymMatrix2DReal f(4, 4);
    int baseInd = 0;
   	f(baseInd+0, baseInd+0) = 1;	f(baseInd+0, baseInd+1) = 0;	f(baseInd+0, baseInd+2) = 0;	f(baseInd+0, baseInd+3) = 0;			
	f(baseInd+1, baseInd+0) = 0;    f(baseInd+1, baseInd+1) = 1;	f(baseInd+1, baseInd+2) = 0;	f(baseInd+1, baseInd+3) = 0;			
	f(baseInd+2, baseInd+0) = 0;	f(baseInd+2, baseInd+1) = 0;	f(baseInd+2, baseInd+2) = 1;	f(baseInd+2, baseInd+3) = 0;			
	f(baseInd+3, baseInd+0) = 0;	f(baseInd+3, baseInd+1) = 0;	f(baseInd+3, baseInd+2) = 0;	f(baseInd+3, baseInd+3) = 1;			
	
	SymSpaceTransToroidal2D S(f.m, f.n);
	//SymSpaceTrans2D S(f.m, f.n);
	
	SymMatrix2D & Sf = (SymMatrix2D &)S.CalculateSymmetryAndProcess(f);
	std::cout<<"f="<<std::endl;
	f.PrintMatrix();
	std::cout<<"S(f)="<<std::endl;
	Sf.PrintMatrix(0);
}

/****************** Translations ***************************/
SymSpaceTrans2D::SymSpaceTrans2D(int im_m, int im_n, int s_m, int s_n, NormalizationType norm, double p, 
								 double constNormalizationScalar)
: SymSpace2D(im_m, im_n, s_m<0 ? 2*im_m : s_m, s_n<0 ? 2*im_n : s_n, p, norm, constNormalizationScalar)
{
	m_pTmpFFTConj = new SymMatrix2DComplex(2*im_m, 2*im_n);
	m_pTmpFFT = new SymMatrix2DComplex(2*im_m, 2*im_n);
	m_pTmpConvFFT = new SymMatrix2DComplex(2*im_m, 2*im_n);	
	m_pTmpFullAnswer = new SymMatrix2DReal(2*im_m, 2*im_n);
	m_pPaddedF1 = new SymMatrix2DReal(2*im_m, 2*im_n);
	m_pPaddedF2 = new SymMatrix2DReal(2*im_m, 2*im_n);
	
}

SymSpaceTrans2D::~SymSpaceTrans2D()
{
}

void SymSpaceTrans2D::CalculateSymmetry(SymMatrix & f, double * L2f)
{
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
	
	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
	SymMatrix2DReal & fReal = (SymMatrix2DReal &)f;
	for (int i=0; i<m_pPaddedF1->m; i++)
	{
		for (int j=0; j<m_pPaddedF1->n; j++)
		{
			(*m_pPaddedF1)(i,j) = 0;
			(*m_pPaddedF2)(i,j) = 0;
		}
	}
	
	for (int i=0; i<fReal.m; i++)
	{
		for (int j=0; j<fReal.n; j++)
		{
			(*m_pPaddedF1)(i,j) = fReal(i,j);
			(*m_pPaddedF2)(fReal.m+i, fReal.n+j) = fReal(i,j);
		}
	}
	
	m_FFTManager.FFT2D(*m_pPaddedF2, *m_pTmpFFT);	
	m_pTmpFFT->Conj(*m_pTmpFFTConj);	
	m_FFTManager.FFT2D(*m_pPaddedF1, *m_pTmpFFT);
	m_pTmpFFT->PointwiseMultiply(*m_pTmpFFTConj, *m_pTmpConvFFT);
	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);

	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
	{
		Sreal = *m_pTmpFFT;
	}
	else
	{
		*m_pTmpFullAnswer = *m_pTmpFFT;
		CropCenter(*m_pTmpFullAnswer);
	}
}

/****************** Identity ***************************/
SymSpaceIdentity::SymSpaceIdentity(int dimW, int dimH, NormalizationType normType) 
: SymSpace2D(dimW, dimH, dimW, dimH, 1, normType, 1, 1)
{
}

void SymSpaceIdentity::CalculateSymmetry(SymMatrix & f, double * L2f)
{
	if (L2f!=NULL)
		*L2f = f.SumElements(2);
	
	((SymMatrix2DReal &)*m_pSymmetryValues) =(SymMatrix2DReal &) f;
}

/****************** Translational Cross Correlation ***************************/
//SymSpaceTransCrossCorr2D::SymSpaceTransCrossCorr2D(int im_m, int im_n, int S_m, int S_n,  NormalizationType norm)
//: SymSpace2D(im_m, im_n, s_m<0 ? 2*im_m : s_m, s_n<0 ? 2*im_n : s_n, 1, norm, 1, 1)
//{
//	m_pTmpFFTConj = new SymMatrix2DComplex(s_m, s_n);
//	m_pTmpFFT = new SymMatrix2DComplex(s_m, s_n);
//	m_pTmpConvFFT = new SymMatrix2DComplex(s_m, s_n);	
//	m_pTmpFullAnswer = new SymMatrix2DReal(s_m, s_n);
//	m_pPaddedF1 = new SymMatrix2DReal(s_m, s_n);
//	m_pPaddedF2 = new SymMatrix2DReal(s_m, s_n);
//	
//}
//
//void SymSpaceTransCrossCorr2D::CalculateSymmetry(SymMatrix & f)
//{
//	SymMatrix2DReal & Sreal = (SymMatrix2DReal &)*m_pSymmetryValues;
//	SymMatrixCollection & matrixCollection = (SymMatrixCollection &)f;
//	assert(matrixCollection.GetMatrixInCollection(0).GetNumDims()==2 
//		   && matrixCollection.GetMatrixInCollection(0).GetEntryType()==SYM_ENTRY_REAL
//		   && matrixCollection.GetMatrixInCollection(1).GetNumDims()==2 
//		   && matrixCollection.GetMatrixInCollection(1).GetEntryType()==SYM_ENTRY_REAL);
//	SymMatrix2DReal & f = (SymMatrix2DReal &)matrixCollection.GetMatrixInCollection(0);
//	SymMatrix2DReal & g;	
//	
//	for (int i=0; i<m_pPaddedF1->m; i++)
//	{
//		for (int j=0; j<m_pPaddedF1->n; j++)
//		{
//			(*m_pPaddedF1)(i,j) = 0;
//			(*m_pPaddedF2)(i,j) = 0;
//		}
//	}
//	
//	for (int i=0; i<fReal.m; i++)
//	{
//		for (int j=0; j<fReal.n; j++)
//		{
//			(*m_pPaddedF1)(i,j) = fReal(i,j);
//			(*m_pPaddedF2)(fReal.m+i, fReal.n+j) = fReal(i,j);
//		}
//	}
//	
//	m_FFTManager.FFT2D(*m_pPaddedF2, *m_pTmpFFT);	
//	m_pTmpFFT->Conj(*m_pTmpFFTConj);	
//	m_FFTManager.FFT2D(*m_pPaddedF1, *m_pTmpFFT);
//	m_pTmpFFT->PointwiseMultiply(*m_pTmpFFTConj, *m_pTmpConvFFT);
//	m_FFTManager.IFFT2D(*m_pTmpConvFFT, *m_pTmpFFT);
//	
//	if (m_pTmpFFT->m==Sreal.m && m_pTmpFFT->n==Sreal.n)
//	{
//		Sreal = *m_pTmpFFT;
//	}
//	else
//	{
//		*m_pTmpFullAnswer = *m_pTmpFFT;
//		CropCenter(*m_pTmpFullAnswer);
//	}	
//}




