
#include "Numerical.h"
#include "SymSpace.h"
#include "TimeProfiler.h"
#include "R2Shapes/R2Shapes.h"

#ifndef __SYM_SPACE_REFLECTIVE_H
#define __SYM_SPACE_REFLECTIVE_H

class DeformationModel;

/**
 * Kazhdan's symmetry (i.e. reflective for planes around one point, uses FFT for faster transform)
 */
class SymSpaceSinglePointReflection2D : public SymSpace2D
{
	public:
		enum CenterOfMassFlags
		{
			COM_AREA_MASS = 0,			// just center of an image
			COM_INTENSITY_MASS = 1,		// center weigthed by intensities
			COM_SYMMETRY_MASS = 2, 
			COM_UPDATE = 4
		};
	
		SymSpaceSinglePointReflection2D(int img_m, int img_n, int numAngles, int numR, 
										DeformationModel * deformationModel, double p = 1, 
										NormalizationType norm = SymSpace::NORM_IMAGE_SPACE, 
										double constNormScalar=1, bool cropToDisk=true,
										int centerOfMass = SymSpaceSinglePointReflection2D::COM_AREA_MASS);
		virtual QImage GetVisualizationOfSymSpace(SymMatrix2DReal & Sf, QImage * img);
		static void ConvertToPolar(SymMatrix2DReal & src_f, SymMatrix2DReal & trg_f_polar, double scaleR, double cx, double cy);
		static void FillColSymmetryPerAngle(SymMatrix2DReal & fillMe, int col, 
											SymMatrix2DReal & f_polar, SymMatrix2DComplex * tmpFFT, 
											SymMatrix2DComplex * tmpConvFFT);
		static void TestPointReflection();
		virtual void SetCenterOfMass();
		virtual void GetSymmetryCenterOfMass(double * cx, double *cy);
	
		int m_iCenterOfMassFlags;
		bool m_bCenterInitialized;
	
	protected:
		virtual void CalculateSymmetry(SymMatrix & f, double *L2f);	
		QImage * m_Image;
		
		int m_fNumR;
		double m_fScaleR;
		SymMatrix2DReal * m_FPolar;
		SymMatrix2DComplex * m_pTmpFFT;
		SymMatrix2DComplex * m_pTmpConvFFT;
		DeformationModel * m_pDeformationModel;
		int m_iVisualizationAlphaScaleFactor;
	
};

/**
 * Podolak's PRST transform to symmetry space of all planar reflections.
 */
class SymSpacePRST2D : public SymSpace2D
{
	public:
		enum PRSTCalculationMethod
		{
			PRST_CALC_SAMPLE_ORIENTATIONS,
			PRST_CALC_SAMPLE_POINTS,
			PRST_CALC_SAMPLE_N4
		};
		//SymSpacePRST2D(int img_m, int img_n, int numAngles, int numR, double cx, double cy, double p = 1, NormalizationType norm = SymSpace::NORM_IMAGE_SPACE, double constNormScalar=1);
		SymSpacePRST2D(int img_m, int img_n, int numAngles, int numR=-1, 
					   PRSTCalculationMethod method=PRST_CALC_SAMPLE_ORIENTATIONS, 
					   double p = 1, NormalizationType norm = SymSpace::NORM_IMAGE_SPACE, 
					   double constNormScalar=1, bool cropToDisk=true);
		virtual QImage GetVisualizationOfSymSpace(SymMatrix2DReal & Sf, QImage * img);

		void PRSTBruteForce(SymMatrix2DReal & f);
		void PRSTSampleOrientations(SymMatrix2DReal & f);
		inline void ResampleAt(double ox, double oy, double dx, double dy, double nx, double ny, SymMatrix2DReal & trg, SymMatrix2DReal & src);
		inline void ConvertToLine(double radius, double theta, double * x, double * y, double * dx, double * dy, double * nx, double * ny);
		inline void ConvertXYGammaToRTheta(double x, double y, double gamma, double * r, double * theta);
		
		static void ResampleFAtArbitraryBasis(SymMatrix2DReal & f, double tX, double tY, SymMatrix2DReal & f_resampled);
		static void TestPRSTReflections();
		
	protected:
		virtual void CalculateSymmetry(SymMatrix & f, double *L2f);
		int m_iNumRadii;
		int m_iNumAngles;
		int m_iMaxDiagonal;
		QImage * m_Image;
		SymMatrix2DReal * m_VisualizationMatrix;
		double m_fScaleR;
		double m_fCenterX;
		double m_fCenterY;	
		SymMatrix2DReal * m_FResampled;
		SymMatrix2DComplex * m_pTmpFFT;
		SymMatrix2DComplex * m_pTmpConvFFT;
		PRSTCalculationMethod m_CalcMethod;
		
		double m_bExtraVerbous;
		TimeProfiler m_TimeProfiler;
};

#endif



